/*
 * partition.h: flash paritioning
 *
 * Copyright (C) 2001 2002  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 *
 * $Id: partition.h,v 1.7 2003/01/28 03:43:06 choover Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: partition.h,v 1.7 2003/01/28 03:43:06 choover Exp $"

#ifndef BLOB_PARTITION_H
#define BLOB_PARTITION_H


#include <blob/types.h>

/* A partition table can contain three types of entries:
 * - The last entry
 * - A valid entry
 * - An invalid entry
 *
 * A last entry can be changed into a valid entry by changing the
 * second last bit (with mask 0x00000002) from 1 to 0, or into an
 * invalid entry by changing the last two bits (with mask 0x00000003)
 * from 1 to 0. By doing it like this we can avoid erasing a flash
 * block.
 *
 * Flash chips wear from erase operations (that's why flash lifetime
 * is specified in erase cycles), so we try to limit the number of
 * erase operations. Luckily the flash helps us a little bit with
 * this, because it only allows you to change a '1' bit into a '0'
 * during a write operation. This means that 0xffff can be changed
 * into 0x1010, and 0x1010 into 0x0000, but 0x0000 can't be changed
 * into anything else anymore because there are no '1' bits left.
 *
 * The magic values have been carefully choosen to allow changing a
 * partition entry without having to erase the flash. For a "last"
 * entry, it is best to keep all other fields except the magic
 * unitialised: chances are high that the other fields are still at
 * 0xffffffff so we can reuse them when we change the partition type.
 */

#define BLOB_PART_LAST_MAGIC	(0x42504533) /* BPE3, Blob Partition Entry 3 */
#define BLOB_PART_VALID_MAGIC	(0x42504531) /* BPE1, Blob Partition Entry 1 */
#define BLOB_PART_INVALID_MAGIC	(0x42504530) /* BPE0, Blob Partition Entry 0 */
#define BLOB_PART_NAMELEN	(32)

#define BLOB_PART_OFS_APPEND	((u32) -1)   /* append to previous  */
#define BLOB_PART_SIZ_FULL	((u32) -1)   /* use remaining flash */

typedef struct {
	/* generic (i.e.: blob independent) data */
	u32 magic;		/* magic value */
	u32 next;		/* offset (in bytes) to next partition entry */
	u32 offset;		/* offset (in bytes) wrt start of flash */
	u32 size;		/* partition size (in bytes) */
	char name[BLOB_PART_NAMELEN];

	/* blob specific items, can be changed for other bootloaders */
	u32 flags;		/* blob specific flags */
	u32 mem_base;		/* load address in RAM */
	u32 entry_point;	/* entry point in RAM */
} blob_partition_t;

#define BLOB_PART_FLAG_PTABLE	(1<<0) /* contains a partition table */
#define BLOB_PART_FLAG_LOAD	(1<<1) /* load partition into RAM */
#define BLOB_PART_FLAG_EXEC	(1<<2) /* partition/target is executable */
#define BLOB_PART_FLAG_JFFS2	(1<<3) /* contains a JFFS2 */
#define BLOB_PART_FLAG_CRAMFS	(1<<4) /* contains a cramfs */


/* A partition table is marked with a special entry that marks the
 * start of the table.
 *
 * There are two kinds of blob partition tables: the first one is the
 * regular partition table that can be changed by the bootloader. The
 * second one is a default partition table that can be put into the
 * bootloader binary so it can avoid to waste a flash block for the
 * partition table. The partition table parser first tries to find a
 * regular partition table and will fall back on the default table if
 * it can't find one.
 *
 * A partition table entry point has to specify the total size of the
 * flash (in the size member), and the offset has to specify the
 * absolute base address.
 */

#define BLOB_FLASH_PART_TABLE_MAGIC	(0x42465054) /* BFPT, Blob Flash
						      * Partition Table */
#define BLOB_DEFAULT_PART_TABLE_MAGIC	(0x42445054) /* BDPT, Blob Default
						      * Partition Table */
#define BLOB_COPY_PART_TABLE_MAGIC	(0x42435054) /* BCPT, Blob Copy
						      * Partition Table */




/* flash functions */

extern const blob_partition_t *default_partition_table;
extern blob_partition_t *flash_partition_table;

const blob_partition_t *pt_first(void);
const blob_partition_t *pt_next(const blob_partition_t *p);
const blob_partition_t *pt_find_by_name(const char *s);
const blob_partition_t *pt_find_by_address(u32 addr);

void *pt_flash_start(const blob_partition_t *p);

#endif
