/*
 * badge4.c: Badge 4 specific stuff
 *
 * Copyright (C) 2001-2002  Hewlett-Packard Company
 * Written by Christopher Hoover <ch@hpl.hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: badge4.c,v 1.18 2003/09/04 17:39:00 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch/badge4.h>
#include <blob/flash.h>
#include <blob/init.h>
#include <blob/arch.h>
#include <blob/reboot.h>
#include <blob/serial.h>
#include <blob/util.h>
#include <blob/i2c.h>
#include <blob/i2c-gpio.h>
#include <blob/spd.h>
#include <blob/partition.h>
#include <blob/led.h>
#include <blob/time.h>

#define DEBUG

static struct i2c_bus_gpio_private badge4_i2c_bus_gpio_private = {
	.sda_gpio= BADGE4_GPIO_SDSDA,
	.scl_gpio= BADGE4_GPIO_SDSCL,
	.delay= 100
};

static struct i2c_bus badge4_i2c_bus = {
	.init_bus= init_bus_gpio,
	.set_sda= set_sda_gpio,
	.get_sda= get_sda_gpio,
	.set_scl= set_scl_gpio,
	.get_scl= get_scl_gpio,
	.private= &badge4_i2c_bus_gpio_private
};

/* flash descriptors for BadgePAD 4 flash */
static const flash_descriptor_t badge4_flash_descriptors[] =
	QUERY_FLASH_FOR_DESCRIPTORS(0);


/* default partition table for BadgePAD 4 */
static blob_partition_t badge4_default_partition_table[] = {
	{
		/* start of table */
		.magic=		BLOB_DEFAULT_PART_TABLE_MAGIC,
		.next=		sizeof(blob_partition_t),
		.offset=	0x00000000,	/* absolute base address */
		.size=		BLOB_PART_SIZ_FULL
	},
	{
		/* blob itself */
		.magic=		BLOB_PART_VALID_MAGIC,
		.next=		sizeof(blob_partition_t),
		.offset=	BLOB_PART_OFS_APPEND,
		.size=		BLOB_FLASH_LEN,
		.name=		"blob",
		.mem_base=	BLOB_RAM_BASE
	},
	{
		/* parameter block */
		.magic=		BLOB_PART_VALID_MAGIC,
		.next=		sizeof(blob_partition_t),
		.offset=	BLOB_PART_OFS_APPEND,
		.size=		PARAM_FLASH_LEN,
		.name=		"param",
		.flags=		BLOB_PART_FLAG_PTABLE
	},
#ifdef CONFIG_ZIMAGE_SUPPORT
	{
		/* parameter block */
		.magic=		BLOB_PART_VALID_MAGIC,
		.next=		sizeof(blob_partition_t),
		.offset=	BLOB_PART_OFS_APPEND,
		.size=		KERNEL_FLASH_LEN,
		.name=		"kernel",
		.flags=		BLOB_PART_FLAG_EXEC,
		.mem_base=	KERNEL_RAM_BASE,
		.entry_point=	KERNEL_RAM_BASE
	},
#endif
	{
		/* root */
		.magic=		BLOB_PART_VALID_MAGIC,
		.next=		sizeof(blob_partition_t),
		.offset=	BLOB_PART_OFS_APPEND,
		.size=		BLOB_PART_SIZ_FULL,
		.name=		"ramdisk", /* not really */
		.flags=		BLOB_PART_FLAG_JFFS2
	},
	{
		/* last entry */
		.magic=		BLOB_PART_LAST_MAGIC
	}
};


static void badge4_set_partition_table(void)
{
	/* set up the default partition table */
	default_partition_table = badge4_default_partition_table;

	/* we don't know where the flash partition table will live, so
	 * point to the flash and let the partition table
	 * parser figure out.
	 */
	flash_partition_table = (blob_partition_t*)0x00000000;
}
__initlist(badge4_set_partition_table, INIT_LEVEL_OTHER_STUFF + 1);


static void badge4_select_drivers(void)
{
	/* select drivers */
	reboot_driver = &sa11x0_reboot_driver;
	serial_driver = &sa11x0_serial_driver;
	led_driver = &sa11x0_gpio_led_driver;
	timer_driver = &intelarm_timer_driver;

	flash_driver = &intel16_flash_driver;
	flash_descriptors = badge4_flash_descriptors;
}
__initlist(badge4_select_drivers, INIT_LEVEL_DRIVER_SELECTION);


static void badge4_banner(void)
{
	printf("\n\n\n" BOARD_NAME "\n"
	       "Blob port by Christopher Hoover <ch@hpl.hp.com>.\n"
	       "  (revision: $Revision: 1.18 $, built: " __DATE__ ")\n"
	       "\n");
}
__initlist(badge4_banner, INIT_LEVEL_INITIAL_HARDWARE + 1);


static void badge4_init_sa1111(void)
{
	/*
	 * We have to do part of the SA-1111 initialization here.
	 * Whenever we pull down GPIO_SA1111_NRST (GPIO 25, 1111_RST~
	 * on the schematic), the SDRAM goes away as that gpio is also
	 * wired into the CPLD that controls the SO-DIMM SDRAM
	 * interface.  (Argh.)
	 *
	 * The only reason this code runs at all is that we have set
	 * up blob to run out of SRAM.  Fortunately all BadgePAD 4's
	 * have SRAM.
	 */

	/* Asert nRESET. */
	GPCR = BADGE4_GPIO_SA1111_NRST;
	GPDR |= BADGE4_GPIO_SA1111_NRST;

	/*
	 * Set up the 3.6864MHz clock on GPIO 27 for the SA-1111:
	 * (SA-1110 Developer's Manual, section 9.1.2.1)
	 */
	GAFR |= GPIO_32_768kHz;
	GPDR |= GPIO_32_768kHz;
	TUCR = TUCR_3_6864MHz;

	/* De-assert nRESET */
	GPSR = BADGE4_GPIO_SA1111_NRST;

	/*
	 * The SA-1111 initialization is not yet done, but Linux
	 * can do the rest.
	 */
}

static void badge4_setup_sdram(void)
{
	int i, size;
	int typ0, typ1;
	int memory_type, row_bits, col_bits, module_rows, row_density;
	u8 spd[128];

	/* Disable memory bus/grant signals to avoid any suprious
           memory requests */
	PGSR &= ~GPIO_MBGNT;
	GPCR = GPIO_MBGNT;
	GPDR = (GPDR & ~GPIO_MBREQ) | GPIO_MBGNT;
	GAFR &= ~(GPIO_MBGNT | GPIO_MBREQ);

	GPDR |= (BADGE4_GPIO_SDTYP1 | BADGE4_GPIO_SDTYP0);

	/* Intialize the I2C bus and driver. */
	i2c_init(&badge4_i2c_bus);

	/* Probe for the SPD EEPROM */
	if (i2c_probe(&badge4_i2c_bus, SPD_I2C_ADDR(0)) != 0) {
		printf("SDRAM: SDRAM SO-DIMM absent.\n");
		return;
	}

	/* Read the SPD to determine the SO-DIMM SDRAM parameters. */
	if (i2c_read_device_bytes(&badge4_i2c_bus, SPD_I2C_ADDR(0), 
				  0, spd, sizeof(spd)) != 0) {
		printf("SDRAM: failed to read SPD EEPROM.\n");
		goto fail;
	}

	/* quick sanity check */
	memory_type = spd[SPD_MEMORY_TYPE];
	if (memory_type != SPD_MEMORY_TYPE_SDRAM) {
		printf("SDRAM: SDRAM SO-DIMM wrong type (%d)\n", memory_type);
		goto fail;
	}

	row_bits = spd[SPD_ROW_BITS];
	col_bits = spd[SPD_COL_BITS];
	module_rows = spd[SPD_MODULE_ROWS];
	row_density = spd[SPD_ROW_DENSITY];

	/* calculate the size: bit 0 is 4M, bit 1 is 8M, etc. */
	size = 0;
	for (i = 0; i < 8; i++)
		if (row_density & (1<<i))
			size += (4<<i);
	size *= module_rows;

	printf("SDRAM: %d MiB (rows=%d, row bits=%d, col bits=%d)\n", 
	       size, module_rows, row_bits, col_bits);

	/*
	 * Here's what the CPLD expects
	 *
	 * typ1 typ0	flavor				typical geometry
	 * 0	0	SO-DIMM with 64 Mbit parts	xx/8
	 * 0	1	SO-DIMM with 128 Mbit parts	12/9
	 * 1	0	SO-DIMM with 256 Mbit parts	13/9
	 */

	/* ### should we use the other SPD info or is col/row bits enough? */
	if (col_bits == 8) {
		typ1 = 0; typ0 = 0;
	} else if (row_bits == 12 && col_bits == 9) {
		typ1 = 0; typ0 = 1;
	} else if (row_bits == 13 && col_bits == 9) {
		typ1 = 1; typ0 = 0;
	} else {
		printf("SDRAM: unexpected geometry\n");
		goto fail;
	}

 set:
	printf("SDRAM: setting CPLD typ1=%d, typ0=%d\n", typ1, typ0);
	if (typ1)
		GPSR = BADGE4_GPIO_SDTYP1;
	else
		GPCR = BADGE4_GPIO_SDTYP1;
	if (typ0)
		GPSR = BADGE4_GPIO_SDTYP0;
	else
		GPCR = BADGE4_GPIO_SDTYP0;

	return;

 fail:
	printf("SDRAM: making unwarranted assumption that SDRAM is 12x9\n");
	typ1 = 0; typ0 = 1;
	goto set;
}

static void badge4_init_hardware2(void)
{
	badge4_init_sa1111();
	badge4_setup_sdram();
}
__initlist(badge4_init_hardware2, INIT_LEVEL_INITIAL_HARDWARE + 2);

