/*-------------------------------------------------------------------------
 * Filename:      flash.c
 * Version:       $Id: flash.c,v 1.19 2003/01/24 09:26:00 choover Exp $
 * Copyright:     Copyright (C) 1999, Jan-Derk Bakker
 * Author:        Jan-Derk Bakker <J.D.Bakker@its.tudelft.nl>
 * Description:   Flash I/O functions for blob
 * Created at:    Mon Aug 23 20:00:00 1999
 * Modified by:   Erik Mouw <J.A.K.Mouw@its.tudelft.nl>
 * Modified at:   Sat Jan 15 19:16:34 2000
 *-----------------------------------------------------------------------*/
/*
 * flash.c: Flash I/O functions for blob
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 1999  Jan-Derk Bakker (J.D.Bakker@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: flash.c,v 1.19 2003/01/24 09:26:00 choover Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/debug.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/led.h>
#include <blob/util.h>
#include <blob/flash.h>

#include <blob/init.h>
#include <blob/command.h>


/* this is enough for a 32MB flash with 128kB blocks */
#define NUM_FLASH_BLOCKS	(256)


typedef struct {
	u32 start;
	u32 size;
	int lockable;
} flash_block_t;


static flash_block_t flash_blocks[NUM_FLASH_BLOCKS];
static int num_flash_blocks;

const flash_descriptor_t *flash_descriptors;
flash_driver_t *flash_driver;




/* dummy function for enable_vpp and disable_vpp */
int flash_dummy_ok(void)
{
	return 0;
}




/* debug function: just print out the flash map and if the blocks are
 * locked or not
 */
static void print_flash_map(void)
{
#ifdef BLOB_DEBUG
	int i;

	printf("Flash map:\n");
	for(i = 0; i < num_flash_blocks; i++) {
		u32 start = flash_blocks[i].start;

		printf("%3i: 0x%08x @ 0x%08x (%i kb), ", 
		       i, flash_blocks[i].size, start,
		       flash_blocks[i].size / 1024);

		if(flash_blocks[i].lockable && 
		   flash_driver->query_block_lock) {
			if ((flash_driver->query_block_lock)((u32 *) start))
				printf("locked\n");
			else
				printf("unlocked\n");
		} else {
			printf("not lockable\n");
		}
	}
#endif
}




/* initialise the flash blocks table */
static void init_flash(void)
{
	int i = 0;
	unsigned int j;
	u32 start = 0;
	
	if(flash_descriptors == NULL) {
		deprintf("undefined flash_descriptors\n");
		return;
	}

	if(flash_driver == NULL) {
		deprintf("undefined flash_driver\n");
		return;
	}

	/* fill out missing flash driver functions */
	if(flash_driver->enable_vpp == NULL)
		flash_driver->enable_vpp = flash_dummy_ok;

	if(flash_driver->disable_vpp == NULL)
		flash_driver->disable_vpp = flash_dummy_ok;

	if (flash_descriptors[0].num == (unsigned int)-1) {
		u32 *flashAddr = (u32 *) flash_descriptors[0].size;
		flash_descriptor_t *d;
		int rc;

		if (flash_driver->query_descriptors == 0) {
			deprintf("driver does not support query_descriptors\n");
			flash_descriptors = 0;
			return;
		}

		rc = flash_driver->query_descriptors(flashAddr, &d);
		if (rc) {
			deprintf("driver failed to query flash for descriptors\n");
			flash_descriptors = 0;
			return;
		}

		flash_descriptors = d;
	}

	/* initialise flash blocks table */
	num_flash_blocks = 0;

	while(flash_descriptors[i].size != 0) {
		dprintf("%i x 0x%08x, ",
			flash_descriptors[i].num,
			flash_descriptors[i].size);

		if(!flash_descriptors[i].lockable){
			dprintf("not ");
		}

		dprintf("lockable\n");

		for(j = 0; j < flash_descriptors[i].num; j++) {
			flash_blocks[num_flash_blocks].start = start;
			flash_blocks[num_flash_blocks].size =
				flash_descriptors[i].size;
			flash_blocks[num_flash_blocks].lockable =
				flash_descriptors[i].lockable;

			start += flash_descriptors[i].size;

			num_flash_blocks++;

			if(num_flash_blocks >= NUM_FLASH_BLOCKS) {
				eprintf("not enough flash_blocks\n");
				break;
			}
		}

		i++;
	}

//	print_flash_map();
}


__initlist(init_flash, INIT_LEVEL_OTHER_STUFF + 1);




int flash_erase_region(u32 *start, u32 nwords)
{
	u32 *cur;
	u32 *end;
	int rv;

	cur = start;
	end = start + nwords;

	dprintf("erasing 0x%08x (%d) words at 0x%08x\n",
		nwords, nwords, (unsigned int)start);

	flash_driver->enable_vpp();

	while(cur < end) {
		if(*cur != 0xffffffff) {
			printf("erasing dirty block at 0x%08x\n", (u32)cur);

			/* dirty block */
			rv = flash_driver->erase(cur);

			if(rv < 0) {
				eprintf("flash erase error at 0x%08x\n",
					(u32)cur);
				flash_driver->disable_vpp();
				return rv;
			}

			continue;
		}
		cur ++;
	}
	
	flash_driver->disable_vpp();

	return 0;
}


/* get the flash size
 *
 */
u32 flash_get_size( void )
{
	int block;
	u32 size = 0L;

	for ( block=0; block<num_flash_blocks; block++ ) {
		size += flash_blocks[block].size;
	}

	return size;
}


/* Write a flash region with a minimum number of erase operations.
 *
 * Flash chips wear from erase operations (that's why flash lifetime
 * is specified in erase cycles), so we try to limit the number of
 * erase operations in this function. Luckily the flash helps us a
 * little bit with this, because it only allows you to change a '1'
 * bit into a '0' during a write operation. This means that 0xffff can
 * be changed into 0x1010, and 0x1010 into 0x0000, but 0x0000 can't be
 * changed into anything else anymore because there are no '1' bits
 * left.
 */
int flash_write_region(u32 *dst, const u32 *src, u32 nwords)
{
	int rv;
	u32 nerrors = 0;
	u32 i = 0;

	u32 nerase = 0;
	u32 nwrite = 0;
	u32 nscandown = 0;
	u32 nskip = 0;

	dprintf("flashing 0x%08x (%d) words from 0x%08x to 0x%08x\n",
		nwords, nwords, (u32)src, (u32)dst);

	flash_driver->enable_vpp();

	while(i < nwords) {
		/* spinner */
		{
			char spinner[4] = { '/','-','\\','|' };
			if ((i & 0x3ff) == 0) {
				printf("%c\r", spinner[(i >> 10) & 0x3]);
			}
		}
		/* nothing to write */
		if(dst[i] == src[i]) {
			i++;
			nskip++;
			continue;
		}

		/* different, so write to this location */
		rv = flash_driver->write(&dst[i], &src[i]);
		nwrite++;
		
		if(rv == 0) {
			i++;
		} else {
			nerrors++;
			
			printf("erasing at 0x%08x...", (u32)&dst[i]);

			/* erase block at current location */
			rv = flash_driver->erase(&dst[i]);
			nerase++;
			if(rv < 0) {
				/* something is obviously wrong */
				printf(" error\n");
				flash_driver->disable_vpp();

				return rv;
			}
				
			printf(" scanning down...");

			/* scan down until we find the first
                           non-erased location and restart writing
                           again from that location */
			while((i > 0) &&
			      ((dst[i] != src[i]) || (dst[i] == 0xffffffff))) {
				i--;
				nscandown++;
			}

			printf(" resume writing at 0x%08x\n", (u32)&dst[i]);
		}
		
		/* there is something seriously wrong if this is true */
		if(nerrors > 2 * nwords) {
			eprintf("too many flash errors, probably hardware error\n");
			flash_driver->disable_vpp();

			return -EFLASHPGM;
		}
	}


	dprintf("%d words source image\n", nwords);
	dprintf("%d words written to flash\n", nwrite);
	dprintf("%d words skipped\n", nskip);
	dprintf("%d erase operations\n", nerase);
	dprintf("%d words scanned down\n", nscandown);

	flash_driver->disable_vpp();

	return 0;
}


/* given an address, return the flash block index number (or negative
 * error number otherwise
 */
static int find_block(u32 address)
{
	int i;

	for (i = 0; i < num_flash_blocks; i++) {
		u32 start = flash_blocks[i].start;
		int length = flash_blocks[i].size;
		u32 endIncl = start + length - 1;

		if (address >= start && address <= endIncl)
			return i;
	}

	return -ERANGE;
}


/* given an address, return the size of that flash block */
int flash_get_block_size(u32 address)
{
	int block = find_block(address);
	if (block == -ERANGE) 
		return 0;
	return flash_blocks[block].size;
}


/* convert address range to range of flash blocks. returns 0 on
 * success or negative error number on failure.
 */
static int address_range_to_block_range(u32 startAddress, int length, 
					int *startBlock, int *endInclBlock)
{
	int sb, eib;

	dprintf("startAddress = 0x%08x, length = 0x%08x\n",
		startAddress, length);

	sb = find_block(startAddress);
	eib = find_block(startAddress + length - 1);

	if(sb < 0)
		return sb;

	if(eib < 0)
		return eib;

	dprintf("sb = %d, eib = %d\n", sb, eib);

	/* FIXME: would be nice to warn if sb isn't at the beginning
	 * of its block or eib isn't at the very end of its block.
	 */

	*startBlock = sb;
	*endInclBlock = eib;

	return 0;
}




static int do_flash_lock(u32 *start, u32 nwords, int lock)
{
	int sb, eib;
	int rv;
	int i;
	u32 *addr;

	dprintf("%slock at 0x%08x, nwords = 0x%08x\n",
		lock ? "" : "un",
		(u32)start, nwords);
		
	rv = address_range_to_block_range((u32) start, nwords * sizeof(u32),
					  &sb, &eib);

	if(rv < 0)
		return rv;

	/* check if it is lockable at all */
	for(i = sb; i <= eib; i++) {
		if(!flash_blocks[i].lockable) {
			deprintf("can't (un)lock unlockable blocks\n");
			return -EFLASHPGM;
		}
	}
	
	flash_driver->enable_vpp();

	for(i = sb; i <= eib; i++) {
		addr = (u32 *)flash_blocks[i].start;

		if(lock)
			rv = flash_driver->lock_block(addr);
		else
			rv = flash_driver->unlock_block(addr);

		if(rv < 0) {
			flash_driver->disable_vpp();
			deprintf("can't (un)lock block at 0x%08x",
				 (u32)addr);
			return rv;
		}
	}
	
	flash_driver->disable_vpp();

	return 0;
}




int flash_lock_region(u32 *start, u32 nwords)
{
	return do_flash_lock(start, nwords, 1);
}




int flash_unlock_region(u32 *start, u32 nwords)
{
	return do_flash_lock(start, nwords, 0);
}




/* return number of blocks in the region locked, 0 if everything is
 * unlocked, or negative error number otherwise */
int flash_query_region(u32 *start, u32 nwords)
{
	int sb, eib, rv, i;
	int cnt = 0;
	u32 *addr;

	rv = address_range_to_block_range((u32) start, nwords * sizeof(u32),
					  &sb, &eib);

	if(rv < 0)
		return rv;

	for(i = sb; i <= eib; i++) {
		addr = (u32 *)flash_blocks[i].start;
		
		if(flash_blocks[i].lockable) {
			rv = flash_driver->query_block_lock(addr);

			if(rv < 0) 
				return rv;

			if(rv > 0)
				cnt++;
		}
	}

	return cnt;
}




int flash_get_first_block_address(u32 *addr)
{
	if(num_flash_blocks != 0) {
		*addr = flash_blocks[0].start;
		return 0;
	} else {
		return -ERANGE;
	}
}




int flash_get_next_block_address(u32* addr, u32 current)
{
	int i;
	u32 new_addr;

	/* first find the current flash block */
	i = find_block(current);

	/* invalid address? */
	if(i < 0)
		return i;

	/* calculate new address and check if it is valid */
	new_addr = flash_blocks[i].start + flash_blocks[i].size;
	i = find_block(new_addr);

	if(i < 0)
		return i;

	/* it's valid, return new address */
	*addr = new_addr;

	return 0;
}
