/*
 * intel16: Intel 16 bit flash driver
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 1999  Jan-Derk Bakker (J.D.Bakker@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: intel16.c,v 1.9 2003/09/04 17:39:00 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/errno.h>
#include <blob/flash.h>
#include <blob/util.h>
#include <blob/serial.h>
#include <blob/time.h>
#include <blob/debug.h>

/* flash commands for a single 16 bit intel flash chip */
#define READ_ARRAY		0x000000FF
#define ERASE_SETUP		0x00000020
#define ERASE_CONFIRM		0x000000D0
#define	PGM_SETUP		0x00000040
#define	STATUS_READ		0x00000070
#define	STATUS_CLEAR		0x00000050
#define READ_QUERY		0x00000098

#define STATUS_READY		0x00000080
#define STATUS_ERASE_SUSP	0x00000040
#define STATUS_ERASE_ERR	0x00000020
#define STATUS_PGM_ERR		0x00000010
#define STATUS_VPPLOW_ERR	0x00000008
#define STATUS_PGM_SUSP		0x00000004
#define STATUS_LOCKED_ERR	0x00000002

#define CONFIG_SETUP		0x00000060
#define LOCK_SECTOR		0x00000001
#define UNLOCK_SECTOR		0x000000D0
#define READ_CONFIG		0x00000090
#define LOCK_STATUS_LOCKED	0x00000001
#define LOCK_STATUS_LOCKEDDOWN	0x00000002

#define WRITE_ERROR_MASK \
  (STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define ERASE_ERROR_MASK \
  (STATUS_ERASE_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define LOCK_ERROR_MASK \
  (STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define CONFIG_ERROR_MASK \
  (STATUS_ERASE_ERR|STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)

#define FLASH_TIMEOUT (20)

static int get_status(u16 *addr)
{
	u32 start, end;

	start = timer_read();
	end = start + (FLASH_TIMEOUT * timer_tps());

	for (;;) {
		u16 status;

		status = *addr;
		barrier();

		if ((status & STATUS_READY) == STATUS_READY)
			return status;

		if (timer_read() >= end)
			return -1;
	}
}

static void clear_status(u16 *addr)
{
	*addr = STATUS_CLEAR;
	barrier();
}

static void print_flash_error(const char *what, u16 *addr, int status)
{
	serial_write('\n');
	SerialOutputString(what);
	SerialOutputString(" failed for address 0x");
	SerialOutputHex((u32) addr);
	serial_write('\n');

	SerialOutputString("  status = 0x");
	SerialOutputHex(status);
	serial_write('\n');

	if (status < 0) {
		SerialOutputString("  { timeout fetching status }");
	} else {
		SerialOutputString("  { ");

#define PRINT(x) \
  if (status & x) SerialOutputString(#x " ")

		PRINT(STATUS_READY);
		PRINT(STATUS_ERASE_SUSP);
		PRINT(STATUS_ERASE_ERR);
		PRINT(STATUS_PGM_ERR);
		PRINT(STATUS_VPPLOW_ERR);
		PRINT(STATUS_PGM_SUSP);
		PRINT(STATUS_LOCKED_ERR);

#undef PRINT

		serial_write('}');
	}
	
	SerialOutputString("\n\n");
}

/* erases a flash block at the given address */
static int do_erase(u16 *addr)
{
	int status;

	clear_status(addr);

	/* prepare for erase */
	*addr = ERASE_SETUP;
	barrier();

	/* erase block */
	*addr = ERASE_CONFIRM;
	barrier();

	status = get_status(addr);

	/* put flash back into Read Array mode */
	*addr = READ_ARRAY;
	barrier();

	if ((status < 0) ||
	    ((status & ERASE_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, addr, status);
		return -EFLASHERASE;
	}

	return 0;
}

/* write a flash block at a given location */
static int do_write(u16 *dst, const u16* src)
{
	int status;

	clear_status(dst);

	/* setup flash for writing */
	*dst = PGM_SETUP;
	barrier();

	/* write data */
	*dst = *src;
	barrier();

	/* status check */
	status = get_status(dst);

	/* put flash back into Read Array mode */
	*dst = READ_ARRAY;
	barrier();

	if ((status < 0) || 
	    ((status & WRITE_ERROR_MASK) != 0) ||
	    (*dst != *src)) {
#ifdef FLASH_DEBUG
		/* this often fails because we need to erase, so
		   don't be noisy about it */
		print_flash_error(__FUNCTION__, dst, status);
#endif
		return -EFLASHPGM;
	}
		
	return 0;
}


/* erases a flash block at the given address */
/* we have to break this up in two erases at 16 bit aligned addresses
 * (if necessary)
 */
static int flash_erase_intel16(u32 *addr)
{
	int result;
	u16 *addr16 = (u16*)addr;

	/* erase first block */
	result = do_erase(addr16);
	if(result != 0)
		return result;

	addr16++;
	/* if the second address is not erased, also erase it */
	if(*addr16 != 0xffff)
		result = do_erase(addr16);

	if(result != 0)
		return result;
	
	return 0;
}


/* write a flash block at a given location */
/* this has to be broken up into two consectutive 16 bit writes */
static int flash_write_intel16(u32 *dst, const u32* src)
{
	u16 *dst16 = (u16 *)dst;
	u16 *src16 = (u16 *)src;
	int result;
	
	result = do_write(dst16, src16);
	if(result != 0)
		return result;

	dst16++;
	src16++;
	
	result = do_write(dst16, src16);
	if(result != 0)
		return result;

	return 0;
}

static int flash_lock_block_intel16(u32 *blockStart)
{
	u16 *p = (u16 *) blockStart;
	int status;

	clear_status(p);

	*p = CONFIG_SETUP;
	barrier();
	*p = LOCK_SECTOR;
	barrier();

	status = get_status(p);

	*p = READ_ARRAY;
	barrier();

	if ((status < 0) ||
	    ((status & LOCK_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, p, status);
		return status;
	}

	return 0;
}

static int flash_unlock_block_intel16(u32 *blockStart)
{
	u16 *p = (u16 *) blockStart;
	int status;

	clear_status(p);

	*p = CONFIG_SETUP;
	barrier();
	*p = UNLOCK_SECTOR;
	barrier();

	status = get_status(p);

	*p = READ_ARRAY;
	barrier();


	if ((status < 0) ||
	    ((status & CONFIG_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, p, status);
		return status;
	}

	return 0;
}

static int flash_query_block_lock_intel16(u32 *blockStart)
{
	u16 *p = (u16 *) blockStart;
	u16 result;

	clear_status(p);

	*p = READ_CONFIG;
	barrier();
	result = *(p + 2);
	barrier();

	*p = READ_ARRAY;
	barrier();

#ifdef FLASH_DEBUG
	SerialOutputString(__FUNCTION__ ": lock status of block starting at 0x");
	SerialOutputHex((u32) blockStart);
	SerialOutputString(": 0x");
	SerialOutputHex((u32) result);
	SerialOutputString("\n");
#endif

	return ((result & (LOCK_STATUS_LOCKED|LOCK_STATUS_LOCKEDDOWN)) != 0);
}


/* automatic descriptor support */

#define MAX_AUTO_DESCRIPTORS 4
static flash_descriptor_t auto_descriptors[MAX_AUTO_DESCRIPTORS + 1];

static int flash_query_descriptors_intel16(u32 *flashAddr, 
					   flash_descriptor_t **out)
{
	u16 *addr = (u16 *) flashAddr;
	int result, nregions, n;

	addr[0x55] = READ_QUERY;
	barrier();
	
	if (addr[0x10] != 'Q' && addr[0x11] != 'R' && addr[0x12] != 'Y') {
		result = -EINVAL;
		goto out;
	}

	dprintf("intel16: manuf_code = %02X\n", addr[CFI_MANUF_CODE]);
	dprintf("intel16: device code = %02X\n", addr[CFI_DEVICE_CODE]);
	dprintf("intel16: device size = 2^%d bytes\n", addr[CFI_DEVICE_SIZE_LG2]);

	nregions = addr[CFI_ERASE_REGION_COUNT];

#define READ_U16LE(off)   ((addr[(off)]) | (addr[(off) + 1]<<8))

	for (n = 0; n < nregions && n < MAX_AUTO_DESCRIPTORS; n++) {
		u32 xcount = READ_U16LE(CFI_ERASE_REGION_XBLKS(n));
		u32 count = xcount + 1;
		u32 xsize = READ_U16LE(CFI_ERASE_REGION_XSIZE(n));
		u32 size = (xsize == 0) ? 128 : xsize * 256;
		
		dprintf("intel16: erase region #%d\n", n);
		dprintf("intel16:   count = %d\n", count);
		dprintf("intel16:   size  = %d\n", size);

		auto_descriptors[n].size = size;
		auto_descriptors[n].num = count;
		auto_descriptors[n].lockable = 1; /* fix me */
	}

#undef READ_U16LE

	auto_descriptors[n].size = 0;
	auto_descriptors[n].num = 0;

	*out = auto_descriptors;
	result = 0;

out:
	/* put flash back into Read Array mode */
	barrier();

	*addr = READ_ARRAY;
	barrier();
	
	return result;
}

/* flash driver structure */
flash_driver_t intel16_flash_driver = {
	.erase =		flash_erase_intel16,
	.write =		flash_write_intel16,
	.lock_block = 		flash_lock_block_intel16,
	.unlock_block = 	flash_unlock_block_intel16,
	.query_block_lock = 	flash_query_block_lock_intel16,
	.query_descriptors =	flash_query_descriptors_intel16
};
