/*
 * intel32.c: Intel 32 bit (2x 16 bit) flash driver
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 1999  Jan-Derk Bakker (J.D.Bakker@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: intel32.c,v 1.7 2003/09/04 17:39:00 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/errno.h>
#include <blob/flash.h>
#include <blob/util.h>
#include <blob/serial.h>
#include <blob/time.h>

/* flash commands for two 16 bit intel flash chips */
#define READ_ARRAY		0x00FF00FF
#define ERASE_SETUP		0x00200020
#define ERASE_CONFIRM		0x00D000D0
#define	PGM_SETUP		0x00400040
#define	STATUS_READ		0x00700070
#define	STATUS_CLEAR		0x00500050

#define STATUS_READY		0x00800080
#define STATUS_ERASE_SUSP	0x00400040
#define STATUS_ERASE_ERR	0x00200020
#define STATUS_PGM_ERR		0x00100010
#define STATUS_VPPLOW_ERR	0x00080008
#define STATUS_PGM_SUSP		0x00040004
#define STATUS_LOCKED_ERR	0x00020002

#define CONFIG_SETUP		0x00600060
#define LOCK_SECTOR		0x00010001
#define UNLOCK_SECTOR		0x00D000D0
#define READ_CONFIG		0x00900090
#define LOCK_STATUS_LOCKED	0x00010001
#define LOCK_STATUS_LOCKEDDOWN	0x00020002

#define WRITE_ERROR_MASK \
  (STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define ERASE_ERROR_MASK \
  (STATUS_ERASE_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define LOCK_ERROR_MASK \
  (STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)
#define CONFIG_ERROR_MASK \
  (STATUS_ERASE_ERR|STATUS_PGM_ERR|STATUS_VPPLOW_ERR|STATUS_LOCKED_ERR)

#define FLASH_TIMEOUT (20)

static int get_status(u32 *addr)
{
	u32 start, end;

	start = timer_read();
	end = start + (FLASH_TIMEOUT * timer_tps());

	for (;;) {
		u32 status;

		status = data_from_flash(*addr);
		barrier();

		if ((status & STATUS_READY) == STATUS_READY)
			return status;

		if (timer_read() >= end)
			return -1;
	}
}

static void clear_status(u32 *addr)
{
	*addr = data_to_flash(STATUS_CLEAR);
	barrier();
}

static void print_flash_error(const char *what, u32 *addr, int status)
{
	serial_write('\n');
	SerialOutputString(what);
	SerialOutputString(" failed for address 0x");
	SerialOutputHex((u32) addr);
	serial_write('\n');

	SerialOutputString("  status = 0x");
	SerialOutputHex(status);
	serial_write('\n');

	if (status < 0) {
		SerialOutputString("  { timeout fetching status }");
	} else {
		SerialOutputString("  { ");

#define PRINT(x) \
  if (status & x) SerialOutputString(#x " ")

		PRINT(STATUS_READY);
		PRINT(STATUS_ERASE_SUSP);
		PRINT(STATUS_ERASE_ERR);
		PRINT(STATUS_PGM_ERR);
		PRINT(STATUS_VPPLOW_ERR);
		PRINT(STATUS_PGM_SUSP);
		PRINT(STATUS_LOCKED_ERR);

#undef PRINT

		serial_write('}');
	}
	
	SerialOutputString("\n\n");
}

/* erases a flash block at the given address */
static int flash_erase_intel32(u32 *addr)
{
	u32 status;

	clear_status(addr);

	/* prepare for erase */
	*addr = data_to_flash(ERASE_SETUP);
	barrier();

	/* erase block */
	*addr = data_to_flash(ERASE_CONFIRM);
	barrier();

	status = get_status(addr);

	/* put flash back into Read Array mode */
	*addr = data_to_flash(READ_ARRAY);
	barrier();

	if ((status < 0) ||
	    ((status & ERASE_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, addr, status);
		return -EFLASHERASE;
	}

	return 0;
}

/* write a flash block at a given location */
static int flash_write_intel32(u32 *dst, const u32* src)
{
	u32 status;

	clear_status(dst);

	/* setup flash for writing */
	*dst = data_to_flash(PGM_SETUP);
	barrier();

	/* write data */
	*dst = *src;
	barrier();

	/* status check */
	status = get_status(dst);

	/* put flash back into Read Array mode */
	*dst = data_to_flash(READ_ARRAY);
	barrier();

	if ((status < 0) || 
	    ((status & WRITE_ERROR_MASK) != 0) ||
	    (*dst != *src)) {
#ifdef FLASH_DEBUG
		/* this often fails because we need to erase, so
		   don't be noisy about it */
		print_flash_error(__FUNCTION__, dst, status);
#endif
		return -EFLASHPGM;
	}
		
	return 0;
}

static int flash_lock_block_intel32(u32 *blockStart)
{
	u32 status;

	clear_status(blockStart);

	*blockStart = data_to_flash(CONFIG_SETUP);
	barrier();
	*blockStart = data_to_flash(LOCK_SECTOR);
	barrier();

	status = get_status(blockStart);

	*blockStart = data_to_flash(READ_ARRAY);
	barrier();

	if ((status < 0) ||
	    ((status & LOCK_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, blockStart, status);
		return status;
	}

	return 0;
}

static int flash_unlock_block_intel32(u32 *blockStart)
{
	u32 status;

	clear_status(blockStart);

	*blockStart = data_to_flash(CONFIG_SETUP);
	barrier();
	*blockStart = data_to_flash(UNLOCK_SECTOR);
	barrier();

	status = get_status(blockStart);

	*blockStart = data_to_flash(READ_ARRAY);
	barrier();

	if ((status < 0) ||
	    ((status & CONFIG_ERROR_MASK) != 0)) {
		print_flash_error(__FUNCTION__, blockStart, status);
		return status;
	}

	return 0;
}

static int flash_query_block_lock_intel32(u32 *blockStart)
{
	u32 result;

	clear_status(blockStart);

	*blockStart = data_to_flash(READ_CONFIG);
	barrier();
	result = data_from_flash(*(blockStart + 2));
	barrier();

	*blockStart = data_to_flash(READ_ARRAY);
	barrier();

#ifdef FLASH_DEBUG
	SerialOutputString(__FUNCTION__ ": lock status of block starting at 0x");
	SerialOutputHex((u32) blockStart);
	SerialOutputString(": 0x");
	SerialOutputHex( result);
	serial_write('\n');
#endif

	return ((result & (LOCK_STATUS_LOCKED|LOCK_STATUS_LOCKEDDOWN)) != 0);
}



/* flash driver structure */
flash_driver_t intel32_flash_driver = {
	erase:			flash_erase_intel32,
	write:			flash_write_intel32,
	lock_block:		flash_lock_block_intel32,
	unlock_block:		flash_unlock_block_intel32,
	query_block_lock:	flash_query_block_lock_intel32
};
