/*
 * lubbock.c: Lubbock specific stuff
 *
 * Copyright (C) 2002 Intel Corporation
 * by :
 *   Rusty Geldmacher (russell.r.geldmacher@intel.com)
 *   Yu Tang (yu.tang@intel.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/flash.h>
#include <blob/init.h>
#include <blob/reboot.h>
#include <blob/serial.h>
#include <blob/led.h>
#include <blob/time.h>
#include <net/ether.h>
#include <net/smc91x.h>




/* flash descriptor for Assabet flash */
/* 2x Intel 28F128J3A strataflash (16MB) */
static const flash_descriptor_t lubbock_flash_descriptors[] =
{
	{
		size: 2 * 128 * 1024,
		num: 64,
		lockable: 1
	},
	{
		/* NULL block */
	},
};




static void init_lubbock_flash_driver(void)
{
	flash_descriptors = lubbock_flash_descriptors;
	flash_driver = &intel32_flash_driver;
}

__initlist(init_lubbock_flash_driver, INIT_LEVEL_DRIVER_SELECTION);



/* PXA250 clock base */
#define BASE_CLK	3686400

/* PXA250 CCCR bits */
#define CCCR_N30	(0x300)
#define CCCR_N20	(0x200)
#define CCCR_N15	(0x180)
#define CCCR_N10	(0x100)
#define CCCR_M4		(0x060)
#define CCCR_M2		(0x040)
#define CCCR_M1		(0x020)
#define CCCR_L45	(0x05)
#define CCCR_L40	(0x04)
#define CCCR_L36	(0x03)
#define CCCR_L32	(0x02)
#define CCCR_L27	(0x01)

#define CLKCFG_T	(0x1)		/* Turbo mode */
#define CLKCFG_F	(0x2)		/* Frequnce change */

/* CPU Frequence Table */
static u32 cpufreqs[] = {
        (CCCR_L27 | CCCR_M1 | CCCR_N10), /* {100/100/100/100} */ 
        (CCCR_L27 | CCCR_M1 | CCCR_N20), /* {200/100/100/100} */
        (CCCR_L27 | CCCR_M1 | CCCR_N30), /* {300/100/100/100} */
        (CCCR_L32 | CCCR_M1 | CCCR_N10), /* {118/118/118/59} */
        (CCCR_L32 | CCCR_M1 | CCCR_N20), /* {236/118/118/59} */
        (CCCR_L32 | CCCR_M1 | CCCR_N30), /* {354/118/118/59} */
        (CCCR_L36 | CCCR_M1 | CCCR_N10), /* {133/133/133/66} */
        (CCCR_L36 | CCCR_M1 | CCCR_N20), /* {266/133/133/66} */
        (CCCR_L36 | CCCR_M1 | CCCR_N30), /* {399/133/133/66} */
        (CCCR_L40 | CCCR_M1 | CCCR_N10), /* {148/148/148/74} */
        (CCCR_L40 | CCCR_M1 | CCCR_N20), /* {296/148/148/74} */
        (CCCR_L45 | CCCR_M1 | CCCR_N10), /* {166/166/166/83} */
        (CCCR_L45 | CCCR_M1 | CCCR_N20), /* {332/166/166/83} */
        (CCCR_L27 | CCCR_M2 | CCCR_N10), /* {200/200/100/100} */
        (CCCR_L27 | CCCR_M2 | CCCR_N15), /* {300/200/100/100} */
        (CCCR_L27 | CCCR_M2 | CCCR_N20), /* {400/200/100/100} */
        (CCCR_L32 | CCCR_M2 | CCCR_N10), /* {236/236/118/59} */
        (CCCR_L36 | CCCR_M2 | CCCR_N10), /* {266/266/133/66} */
        (CCCR_L40 | CCCR_M2 | CCCR_N10), /* {295/295/148/74} */
        (CCCR_L45 | CCCR_M2 | CCCR_N10), /* {332/332/166/83} */ 
	(CCCR_L27 | CCCR_M4 | CCCR_N10)  /* {398/398/100/100} C0/Dalhart permissable */
};

/* 
 * adjust cpu frequence according to switches
 *
 * S25, S26 used to provide all 400 MHz BIN values for Cotulla (0,0 - 1,3)
 *
 * S23: Force the halving of MemClk when deriving SDCLK[1]
 *      DOT: no override  !DOT(1): halve (if not already forced half)
 *      For certain MemClks, SDCLK's derivation is forced to be halved
 *
 * S24: Run/Turbo.  
 * 	DOT: Run mode   !DOT(1): Turbo mode
 */
static void adjust_cpufreq(void)
{
	int sw23, sw24, index;
	int turbo, half, memclk;
	volatile u32 cccr, mdrefr, clkcfg, reg;
	
	/* sw25,sw26(HEX) */
	index = LUB_USERSW & 0x00ff;

	sw23 = LUB_USERSW & 0x0200;
	sw24 = LUB_USERSW & 0x0100;

	turbo = half = 0;
	if (sw23) turbo = 1;
	if (sw24) half = 1;

	clkcfg = CLKCFG_F;
	/* default {200/100/100} */
	cccr = CCCR_L27 | CCCR_M2 | CCCR_N10;

	if ( index < (sizeof(cpufreqs)/sizeof(u32)) )
		cccr = cpufreqs[index];

	memclk = BASE_CLK * (cccr & CCCR_L_MASK) / 1000000;
	/* when memclk > 100, then
	 * 1. force sdclk[1] = memclk/2
	 * 2. adjust msc0, msc1.LH
	 */
	if ( memclk > 100) {
		reg = MSC0;
		reg &= ~0x7F007F00;
		reg |= 0x46004600;
		MSC0 = reg;
		reg = MSC0;

		reg = MSC1;
		reg &= ~0x00007FF0;
		reg |= 0x00004880;
		MSC1 = reg;
		reg = MSC1;

		half = 1;
	}

	mdrefr = MDREFR;
	if (half) 
		mdrefr |= MDREFR_K1DB2;
	else
		mdrefr &= ~MDREFR_K1DB2;

	MDREFR = mdrefr; 

	CCCR = cccr;

	if (turbo) 
		clkcfg |= CLKCFG_T;

	asm volatile("mcr p14, 0, %0, c6, c0, 0"::"r"(clkcfg));

	/* re-write MDREFR after frequence change */
	mdrefr = MDREFR;
	MDREFR = mdrefr;
}

static void lubbock_init_hardware(void)
{
	/* Light HEX Led */
	LUB_LEDHEX = 0xb10bb10b;

	/* Adjuect cpu freq according to switching setting */
	adjust_cpufreq();

	/* select drivers */
	reboot_driver = &pxa_reboot_driver;
	serial_driver = &pxa_serial_driver;
	led_driver = &mmap_led_driver;
	timer_driver = &intelarm_timer_driver;
#ifdef CONFIG_USBDNET_SUPPORT
	ether_driver = &bvdusb_ether_driver;
#else
	ether_driver = &smc91x_ether_driver;
#endif
}


__initlist(lubbock_init_hardware, INIT_LEVEL_DRIVER_SELECTION);
