/* monahans_df.c
*
* Created by stanley cai (stanley.cai@intel.com)

 * (C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
*
* Note:
*   This is a private file which will NOT be release without any modification. The file
* may include some source codes from other open source. There is no warranty to the code.
*   -stanley
*/
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/led.h>
#include <blob/util.h>
#include <blob/serial.h>
#include <blob/flash.h>

#include <blob/init.h>
#include <blob/command.h>

#include <blob/nand.h>
#include <blob/nand_ids.h>

#include "xllp_defs.h"
#include "xllp_clkmgr.h"
#include "xllp_dfc.h"

#define BLOCK_STATUS_UNKNOWN    0x01
#define BLOCK_STATUS_BAD        0x02
#define BLOCK_STATUS_READONLY   0x04
#define BLOCK_STATUS_RESERVED   0x08
#define BLOCK_STATUS_XIP        0x10

// FMD OEM reserved area bitfield.
#define OEM_BLOCK_RESERVED      0x01
#define OEM_BLOCK_READONLY      0x02

#define DFC_READ_PG_SIZE_512            544
#define DFC_READ_PG_SIZE_DATA_SPARE     520
#define NAND_READ_TIMEOUT               100

typedef enum {
	NONDMA_MODE = 0,
	DMA_MODE = 1
} TRANSFER_MODE;

typedef enum {
	ECC_DISABLE = 0,
	ECC_ENABLE = 1
} ECC_MODE;

typedef enum {
	SPARE_DISABLE = 0,
	SPARE_ENABLE = 1
} SPARE_MODE;

typedef enum {
	NAND_FLASH,
	CARBONDALE,
	PIXLEY
} FLASH_MODE;

typedef enum {
	SAMSUNG_CODE = 0xEC
} MAKER_CODE;

/* Relocation Table */
#define NAND_RELOC_MAX 127

typedef struct {
	XLLP_UINT16_T  From;
	XLLP_UINT16_T  To;
} Rel_T;

typedef struct {
	XLLP_UINT16_T  Header;
	XLLP_UINT16_T  NumReloc;
	Rel_T   Relo[NAND_RELOC_MAX];
} Reloc_T;

typedef struct {
	XLLP_UINT8_T  manufacturerId;
	XLLP_UINT8_T  deviceId;
	XLLP_UINT32_T blocks;
	XLLP_UINT32_T pagePerBlock;
	XLLP_UINT32_T pageSize;
} NAND_INFO;

// Flash timing structure for Zylonite board (Samsung K9K1216Q0C)
XLLP_DFC_FLASH_TIMING_T g_FlashTimingZylonite = {
	10,  // tCH, Enable signal hold time.
	0,   // tCS, Enable signal setup time.
	20,  // tWH, ND_nWE high duration.
	40,  // tWP, ND_nWE pulse time.
	20,  // tRH, ND_nRE high duration.
	40,  // tRP, ND_nRE pulse width.
	11123,// tR = tR + tRR + tWB + 1, ND_nWE high to ND_nRE low for read.
	110,  // tWHR, ND_nWE high to ND_nRE low delay for status read.
	10    // tAR, ND_ALE low to ND_nRE low delay.
};

const XLLP_DFC_FLASH_SPECS_T DefaultFlashInfoZylonite = {
	10,         // tCH, Enable signal hold time.
	0,          // tCS, Enable signal setup time.
	20,         // tWH, ND_nWE high duration.
	40,         // tWP, ND_nWE pulse time.
	20,         // tRH, ND_nRE high duration.
	40,         // tRP, ND_nRE pulse width.
	11123,      // tR = tR + tRR + tWB + 1, ND_nWE high to ND_nRE low for read.
	110,        // tWHR, ND_nWE high to ND_nRE low delay for status read.
	10,         // tAR, ND_ALE low to ND_nRE low delay.
	1,          // Data flash bus arbiter enable (ND_ARB_EN).
	32,         // 32, Pages per block (PG_PER_BLK).
	0,          // Second cycle start, Row address start position (RA_START).
	7,          // 2 bytes, returned ID bytes(RD_ID_CNT).
	NAND_FLASH, // NAND, (ND_MODE)
	0,          // Chip select don't care bit (NCSX).
	512,        // 512, Page size in bytes (PAGE_SZ).
	16,         // 16, Width of Flash memory (DWIDTH_M).
	16,         // 16, Width of flash controller(DWIDTH_C).
	4096,       // Number of physical blocks in Flash
	// command codes
	0x0000,     // Read
	0x0000,     // Read1 unused, current DFC don't support
	0x1080,     // Write, two cycle command
	0x0070,     // Read status
	0x0090,     // Read ID
	0xD060,     // Erase, two cycle command
	0x00FF,     // Reset
	0x002A,     // Lock whole flash
	0x2423,     // Unlock, two cycle command, supporting partial unlock
	0x007A      // Read block lock status
};

typedef enum  _FLASH_TYPE { NAND, NOR } FLASH_TYPE;

// NAND geometry info
#define Zylonite_Nand_Info    \
	{SAMSUNG_CODE, 0x56, 4096, 32, 512 }

static NAND_INFO nand_info[] = {
	Zylonite_Nand_Info
};

static struct {
	P_XLLP_DFC_CONTEXT_T pDfc;
	const NAND_INFO *pNANDInfo;
} Monahans_NAND_Info;

typedef struct _FlashInfo
{
	FLASH_TYPE	flashType;
	XLLP_UINT32_T	NumBlocks;
	XLLP_UINT32_T	BytesPerBlock;
	XLLP_UINT32_T   PagePerBlock;
	XLLP_UINT32_T	DataBytesPerPage;
} FlashInfo, *PFlashInfo;

/* Global variables */
XLLP_DFC_CONTEXT_T Monahans_DFC;
XLLP_DFC_MODE_T Monahans_DFC_mode;
Reloc_T BadBlockTable;

XLLP_BOOL_T GetFlashInfo(FlashInfo *pFlashInfo)
{
	XLLP_BOOL_T rc = XLLP_FALSE;

	if (Monahans_NAND_Info.pNANDInfo == NULL) 
		return rc;

        // Memory type is NAND
	pFlashInfo->flashType = NAND;
	pFlashInfo->NumBlocks = Monahans_NAND_Info.pNANDInfo->blocks;
	pFlashInfo->PagePerBlock = Monahans_NAND_Info.pNANDInfo->pagePerBlock;
	pFlashInfo->DataBytesPerPage = Monahans_NAND_Info.pNANDInfo->pageSize;
	pFlashInfo->BytesPerBlock =
		Monahans_NAND_Info.pNANDInfo->pageSize * Monahans_NAND_Info.pNANDInfo->pagePerBlock;

        // Done
	rc = XLLP_TRUE;

	return rc;
}

static int nand_read_page( u32 page, u8* buffer )
{
	XLLP_BOOL_T rc = XLLP_FALSE;
        XLLP_UINT32_T timeout = NAND_READ_TIMEOUT;
        XLLP_UINT8_T bufferTemp[DFC_READ_PG_SIZE_512];
        XLLP_UINT32_T eccErrorPageNum = 0;
        P_XLLP_DFC_CONTEXT_T pDfc = (P_XLLP_DFC_CONTEXT_T)Monahans_NAND_Info.pDfc;
        XLLP_STATUS_T status;

	XllpFlashDfcSetECC(pDfc);
	XllpFlashDfcSetSpare(pDfc);

	memset(bufferTemp, 0x1F, DFC_READ_PG_SIZE_512);
        status = XllpFlashDfcRead(pDfc, bufferTemp, DFC_READ_PG_SIZE_DATA_SPARE,
                                      page, 1, timeout, &eccErrorPageNum);
        if (status != XLLP_STATUS_SUCCESS)
        {
		printf("XllpFlashDfcRead() failed.\n");
                return rc;
    	}

	memcpy(buffer, bufferTemp, Monahans_NAND_Info.pNANDInfo->pageSize);

	rc = XLLP_TRUE;
	return rc;
}

static int nand_load_data( u32 start_block, u8* buffer, u32 size, FlashInfo *pInfo )
{
	u32 pageSize = pInfo->DataBytesPerPage;
	u32 pagePerBlock = pInfo->PagePerBlock;
	u32 pageCount = (size + pageSize - 1) / pageSize;
	u32 i;
	u32 startPage = pagePerBlock * start_block;

	if ( buffer ) {
		for ( i = 0 ; i < pageCount; i++ ) {
			if ( XLLP_TRUE != nand_read_page(startPage + i, buffer) ) {
				printf("nand_read_page( %d, xxx) failed.\n", startPage + i );
				return -1;
			}
			printf(".");
			buffer += pageSize;	
		} 
		printf("Done!\n");
	}
	return 0;
}

/*********************************************************************
 * cmd_nand_load
 *
 * AUTHOR:
 * REVISED:
 *
 * Command wrapper for low-level NAND flash loading
 *
 */
static int cmd_nand_load( int argc, char *argv[] )
{
        int ret = 0;
        u32 start_block = 0L;
	u32 dest_address = 0L;
        u32 size = 0L;
        FlashInfo info;

        if (GetFlashInfo(&info)==XLLP_FALSE) {
                printf("GetFlashInfo() failed!\n");
                ret = -1;
                goto DONE;
        }

        if ( argc < 4 ) {
                ret = -1;
                goto DONE;
        }

        ret = strtou32( argv[1], &start_block );
        ret = strtou32( argv[2], &dest_address );
	ret = strtou32( argv[3], &size );

	printf(" start load data (%d bytes) from NAND block #%d to address 0x%08x\n",
		size, start_block, dest_address );

	nand_load_data( start_block, (u8*)dest_address, size, &info );	
DONE:
        return ret;
}

static char nandloadhelp[] = "nload start_block dest size\n"
"load a NAND block to memory\n";
__commandlist( cmd_nand_load, "nload", nandloadhelp );

int nand_save_data( u32 address, u32 size, u32 start_block, FlashInfo *pInfo )
{
#if 0
        u32 pageSize = pInfo->DataBytesPerPage;
        u32 pagePerBlock = pInfo->PagePerBlock;
        u32 pageCount = (size + pageSize - 1) / pageSize;
        u32 i;
        u32 startPage = pagePerBlock * start_block;

        if ( !buffer ) {
                for ( i = 0 ; i < pageCount; i++ ) {
                        if ( XLLP_TRUE != nand_read_page(startPage + i, buffer) ) {
                                printf("nand_read_page( %d, xxx) failed.\n", startPage + i );
                                return -1;
                        }
                        buffer += pageSize;
                }
	}
#endif
	printf(" Just an empty function!\n");
	return 0;
}

/**********************************************************************
 * cmd_nand_save
 *
 * AUTHOR:
 * REVISED:
 *
 * Command wrapper for low-level flash nand saving
 *
 */
static int cmd_nand_save( int argc, char* argv[] )
{
        int ret = 0;
        u32 source_address = 0L;
        u32 start_block = 0L;
        u32 size = 0L;
        FlashInfo info;

        if (GetFlashInfo(&info)==XLLP_FALSE) {
                printf("GetFlashInfo() failed!\n");
                ret = -1;
                goto DONE;
        }

        if ( argc < 4 ) {
                ret = -1;
                goto DONE;
        }

        ret = strtou32( argv[1], &source_address );
	ret = strtou32( argv[2], &size );
        ret = strtou32( argv[3], &start_block );

	printf("Save data (%d bytes) from address 0x%08x to nand block %d.\n",
		size, source_address, start_block);

	nand_save_data( source_address, size, start_block, &info);	
DONE:
        return ret;
}

static char nandsavehelp[] = "nsave source_address size start_block\n"
"save a NAND flash region\n";
__commandlist( cmd_nand_save, "nsave", nandsavehelp );

/***********************************************************************
 * cmd_nand_erase
 *
 * AUTHOR:
 * REVISED:
 *
 * Command wrapper for low-level flash nand erasing
 *
 */
static int cmd_nand_erase( int argc, char* argv[] )
{
	int ret = 0;
#if 0
        u32 start_block = 0L;
        u32 block_number = 0L;
        u32 len = 0L;
        FlashInfo info;

        if (GetFlashInfo(&info)==XLLP_FALSE) {
        	printf("GetFlashInfo() failed!\n");
        	ret = -1;
        	goto DONE;
        }

        if ( argc < 3 ) {
                ret = -1;
                goto DONE;
        }

        ret = strtou32( argv[1], &start_block );
        ret = strtou32( argv[2], &block_number );

DONE:
#endif
	printf("Just an empty function!\n");
	return ret;
}

static char nanderasehelp[] = "nerase start_block block_number(bytes)\n"
"erase a NAND flash region\n";
__commandlist( cmd_nand_erase, "nerase", nanderasehelp );

/* Sample function to initialize nand flash driver in BLOB */
int init_monahans_nand_driver(void)
{
	XLLP_STATUS_T status;
	XLLP_UINT8_T manufacturer, device;
	P_XLLP_DFC_CONTEXT_T pDfc;
        P_XLLP_CLKMGR_T pCLKMGR_REGS = (P_XLLP_CLKMGR_T)(0x41340000);

	memset(&Monahans_DFC, 0, sizeof(Monahans_DFC));
	memset(&Monahans_DFC_mode, 0, sizeof(Monahans_DFC_mode));

	Monahans_NAND_Info.pDfc = &Monahans_DFC;
	Monahans_NAND_Info.pNANDInfo = &nand_info[0];
	pDfc = &Monahans_DFC;

	Monahans_DFC_mode.enableDMA = NONDMA_MODE;
        Monahans_DFC_mode.enableECC = ECC_ENABLE;
        Monahans_DFC_mode.enableSpareArea = SPARE_ENABLE;
        Monahans_DFC_mode.chipSelect = 0;

	pDfc->pFlashInfo = (P_XLLP_DFC_FLASH_SPECS_T)&DefaultFlashInfoZylonite;
        pDfc->pDFCMode = &Monahans_DFC_mode;
        pDfc->pDFCReg = (P_XLLP_DFC_REG_T)(0x43100000);
        pDfc->pOSTReg = (P_XLLP_OST_T)(0x40A00000);

	/* enable NAND clock, default is on */
	CKENA |= CKENA_4_DFC;

	status = XllpFlashDfcInitNOGPIO(pDfc);
	if (status != XLLP_STATUS_SUCCESS) {
		printf("XllpFlashDfcInitNoGPIO() failed.\n");
		return 0;
	}
		
	XllpFlashDfcReadID(pDfc, &manufacturer, &device);
	printf("\nNAND flash(Manu=0x%x Device=0x%x) detected!\n", manufacturer, device);
	
	return 0;
}

__initlist(init_monahans_nand_driver, INIT_LEVEL_OTHER_HARDWARE);
