/*
 * partition.c: flash partitioning
 *
 * Copyright (C) 2001 2002  Erik Mouw <J.A.K.Mouw@its.tudelft.nl>
 *
 * $Id: partition.c,v 1.10 2003/08/07 19:31:01 abz Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: partition.c,v 1.10 2003/08/07 19:31:01 abz Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/command.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/flash.h>
#include <blob/init.h>
#include <blob/partition.h>
#include <blob/serial.h>
#include <blob/util.h>

#include <blob/debug.h>

#define next_ptable_entry(s) \
	(blob_partition_t *)((u32)s + s->next)


/* pointers to the partition tables */
const blob_partition_t *default_partition_table;
blob_partition_t *flash_partition_table;

/* copy of the flash partition table in RAM */
#define MAX_PARTITIONS	(64)
static blob_partition_t ptable[MAX_PARTITIONS];




static int check_ptable_magic(const blob_partition_t *entry, u32 magic)
{
	if(entry->magic != magic) {
		deprintf("magic failed at 0x%08x (0x%08x != 0x%08x)\n",
			(unsigned int)&entry, 
			entry->magic,
			magic);
		return -EMAGIC;
	} else {
		return 0;
	}
}
	



/* find the flash partition table */
static void find_flash_ptable(void)
{
	u32 old_a, a;
	int rv;

	/* assume that the architecture dependent code has set 
	 * the flash_partition_table pointer to something valid
	 */
	if(flash_partition_table->magic == BLOB_FLASH_PART_TABLE_MAGIC) {
		dprintf("Immediately found flash partition table at 0x%08x\n",
		       (unsigned int)flash_partition_table);
		return;
	}

	/* apparently not. we now assume that the architecture
	 * dependent code has put the flash_partition_table pointer at
	 * the start of the flash, so we start scanning the start of
	 * all flash blocks until we find a partition table.
	 */
	dprintf("Searching flash partition table... \n");
	rv = flash_get_first_block_address(&a);
	if(rv < 0) {
		dprintf("first block not found\n");
		return;
	}

	dprintf("  block at 0x%08x...\n", a);

	while(((blob_partition_t *)a)->magic != BLOB_FLASH_PART_TABLE_MAGIC) {
		old_a = a;
		rv = flash_get_next_block_address(&a, old_a);

		if(rv < 0){
			dprintf("not found\n");
			return;
		}

		dprintf("  block at 0x%08x...\n", a);
	}

	dprintf("found at 0x%08x\n",
	       (unsigned int)a);
	flash_partition_table = (blob_partition_t *)a;
}




/* copy partition table from src to dst. dst can only contain max_dst
 * entries. src must be of type magic. also copy invalidated entries
 * if copy_invalid != 0 */
static int copy_ptable(const blob_partition_t *src, blob_partition_t *dst,
		       u32 magic, int max_dst, int copy_invalid)
{
	int i = 1, rv = 0;

	dprintf("copy from 0x%08x to 0x%08x,\n",
		(u32)src, (u32)dst);
	dprintf("magic = 0x%08x, max_dst = %d, copy_invalid = %d\n",
		magic, max_dst, copy_invalid);

	/* sanity check */
	if(max_dst < 2) {
		deprintf("max_dst too small (%i < 2)\n", max_dst);
		return -EINVAL;
	}

	if(src->magic != magic) {
		deprintf("src not of requested type (0x%08x != 0x%08x)\n",
			 src->magic, magic);
		return -EINVAL;
	}
	
	/* copy start of partition table and go to next entry*/
	memcpy(dst, src, src->next);
	src = next_ptable_entry(src);
	dst = next_ptable_entry(dst);
	i++;

	for(;;) {
		/* we're at the last one */
		if(i == max_dst) {
			rv = -ERANGE;
			break;
		}

		/* check for a valid entry and copy it */
		rv = check_ptable_magic(src, BLOB_PART_VALID_MAGIC);
		if(rv == 0)
			goto do_copy;

		/* check for an invalidated entry */
		rv = check_ptable_magic(src, BLOB_PART_INVALID_MAGIC);
		if(rv == 0) {
			/* only copy when asked */
			if(copy_invalid)
				goto do_copy;
			else
				continue;
		}

		/* check for a last partition entry and stop if found */
		rv = check_ptable_magic(src, BLOB_PART_LAST_MAGIC);
		if(rv == 0) {
			rv = 0;
		} else {
			/* we should NEVER reach this point */
			dprintf("unknown entry at 0x%08x (08%08x)\n",
				(unsigned int)src, src->magic);
			dprintf("this should not happen, bailing out\n");
			rv = -EMAGIC;
		}
		break;

	do_copy:
		memcpy(dst, src, src->next);
		src = next_ptable_entry(src);
		dst = next_ptable_entry(dst);
		i++;
	}

	/* fill out last entry */
	dst->magic = BLOB_PART_LAST_MAGIC;

	dprintf("%d entries copied, returning %d\n", i, rv);

	return rv;
}




static void ptable_print(const blob_partition_t *t)
{
	printf("Partition table for %d kB flash @ 0x%08x:\n",
	       t->size / 1024, t->offset);
	
	for(;;) {
		t = next_ptable_entry(t);
		
		switch(t->magic) {
		case BLOB_PART_LAST_MAGIC:
			return;

		case BLOB_PART_VALID_MAGIC:
			printf("  0x%08x @ 0x%08x (%d kB): '%s'\n",
			       t->size, t->offset, t->size, t->name);

			if(t->flags & BLOB_PART_FLAG_PTABLE)
				printf("    contains parameter blocks\n");

			if(t->mem_base)
				printf("    %sload at 0x%08x\n",
				       (t->flags & BLOB_PART_FLAG_LOAD) ?
				         "automatically " : "",
				       t->mem_base);
			
			if(t->entry_point) {
				printf("    entry point at 0x%08x",
				       t->entry_point);
				if(t->flags & BLOB_PART_FLAG_EXEC)
					printf(", automatically execute\n");
				else
					printf("\n");
			}

			if(t->flags & BLOB_PART_FLAG_JFFS2)
				printf("    contains JFFS2\n");

			if(t->flags & BLOB_PART_FLAG_CRAMFS)
				printf("    contains cramfs\n");
			break;

		case BLOB_PART_INVALID_MAGIC:
			dprintf("invalidated entry\n");
			break;

		default:
			eprintf("ptable magic failed at 0x%08x\n", (u32)t);
			return;
		}
	}
}

static int fixup_ptable(blob_partition_t *ptable)
{
	/* walk through ptable and fix up BLOB_PART_OFS_APPEND and
	   BLOB_PART_SIZ_FULL entries */

	u32 offset = ptable->offset;
	u32 size = ptable->size;
	int no_more_parts_allowed = 0;
	blob_partition_t *t = ptable;
	
	dprintf("ptable->offset = 0x%08X\n", ptable->offset);
	dprintf("ptable->size = 0x%08X\n", ptable->size);
	if (ptable->size == BLOB_PART_SIZ_FULL) {
		ptable->size = size = flash_get_size();
		dprintf("ptable->size => 0x%08X\n", ptable->size);
	}

	for(;;) {
		t = next_ptable_entry(t);

		switch(t->magic) {
		case BLOB_PART_LAST_MAGIC:
			return 0;

		case BLOB_PART_INVALID_MAGIC:
			break;

		case BLOB_PART_VALID_MAGIC:
			dprintf("part @ 0x%p\n", t);

			if (no_more_parts_allowed) {
				eprintf("ptable fixup found additional parts"
					" after BLOB_PART_SIZ_FULL size\n");
				return -EINVAL;
			}

			dprintf("  offset = 0x%08X\n", t->offset);
			if (t->offset == BLOB_PART_OFS_APPEND) {
				t->offset = offset;
				dprintf("  offset => 0x%08X\n", t->offset);
			}

			dprintf("  size   = 0x%08X\n", t->size);
			if (t->size == BLOB_PART_SIZ_FULL) {
				t->size = size;
				no_more_parts_allowed = 1;
				dprintf("  size => 0x%08X\n", t->size);
			}
			offset += t->size;
			size -= t->size;
			break;

		default:
			eprintf("ptable magic failed at 0x%08x\n", (u32)t);
			return -EINVAL;
		}
	}
}


static void ptable_init(void)
{
	int rv;

	/* try to copy the default partition table */
	rv = copy_ptable(default_partition_table, ptable,
			 BLOB_DEFAULT_PART_TABLE_MAGIC, MAX_PARTITIONS, 0);

	if(rv == 0) {
		ptable->magic = BLOB_COPY_PART_TABLE_MAGIC;
		dprintf("found default partition table at 0x%08x\n",
			(unsigned int)default_partition_table);
	}

	/* try to find the flash partition table and try to copy it.
	 * flash partition table overrides a default partition table.
	 */
	find_flash_ptable();
	rv = copy_ptable(flash_partition_table, ptable,
			 BLOB_FLASH_PART_TABLE_MAGIC, MAX_PARTITIONS, 1);
	
	if(rv == 0) {
		ptable->magic = BLOB_COPY_PART_TABLE_MAGIC;
		dprintf("found flash partition table at 0x%08x\n",
			(unsigned int)flash_partition_table);
	}

	/* FIXME: if there is still no flash partition table found we
	 * could check for a bootldr partition table over here */
	
	if (check_ptable_magic(ptable, BLOB_COPY_PART_TABLE_MAGIC) != 0)  {
		dprintf("invalid partition table magic\n");
		goto invalid_part;
	}

	if (fixup_ptable(ptable) != 0) {
		dprintf("could not fix up partition table\n");
		goto invalid_part;
	}

done:
#ifdef BLOB_DEBUG
//	ptable_print(ptable);
#endif
	return;

invalid_part:
	/* construct a minimal partition table */
	ptable[0].magic = BLOB_COPY_PART_TABLE_MAGIC;
	ptable[0].next = sizeof(blob_partition_t);
	ptable[0].offset = 0;
	ptable[0].size = 0;
	ptable[1].magic = BLOB_PART_LAST_MAGIC;
	
	goto done;
}

__initlist(ptable_init, INIT_LEVEL_OTHER_STUFF + 2);



const blob_partition_t *pt_first(void)
{
	return pt_next(ptable);
}

const blob_partition_t *pt_next(const blob_partition_t *p)
{
	for (;;) {
		p = next_ptable_entry(p);

		switch (p->magic) {
		case BLOB_PART_LAST_MAGIC:
			return 0;
		case BLOB_PART_VALID_MAGIC:
			return p;
		case BLOB_PART_INVALID_MAGIC:
			break;
		default:
			return 0;
		}
	}
}


const blob_partition_t *pt_find_by_name(const char *s)
{
	blob_partition_t *t = ptable;

	/* first entry is always the descriptor so it's safe to skip
	 * it */

	for(;;) {
		t = next_ptable_entry(t);

		switch(t->magic) {
		case BLOB_PART_LAST_MAGIC:
			/* this is the last one, we didn't find it :( */
			return NULL;

		case BLOB_PART_VALID_MAGIC:
			if(strncmp(t->name, s, BLOB_PART_NAMELEN)==0)
				return t;
			break;

		case BLOB_PART_INVALID_MAGIC:
			/* skip */
			break;

		default:
			eprintf("ptable magic failed at 0x%08x\n", (u32)t);
			return NULL;
		}
	}
}




const blob_partition_t *pt_find_by_address(u32 addr)
{
	blob_partition_t *t = ptable;

	for(;;) {
		t = next_ptable_entry(t);

		switch(t->magic) {
		case BLOB_PART_LAST_MAGIC:
			return NULL;

		case BLOB_PART_VALID_MAGIC:
			if((addr >= t->offset) && 
			   (addr < (t->offset + t->size)))
				return t;
			break;

		case BLOB_PART_INVALID_MAGIC:
			break;

		default:
			eprintf("ptable magic failed at 0x%08x\n", (u32)t);
			return NULL;
		}
	}
}

void *pt_flash_start(const blob_partition_t *p)
{
	/* absolute location within flash of the given partition */
	return (void *) (ptable->offset + p->offset);
}



static int ptprint(int __attribute__((unused)) argc,
		char __attribute__((unused)) *argv[])
{
	ptable_print(ptable);

	return 0;
}

static char ptprint_help[] = "ptprint\n"
"print flash partition table\n";

__commandlist(ptprint, "ptprint", ptprint_help);
