/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/***************************************************************************
**   xllp_keypad_proc.c
**
**   Valid for Zylonite Keypad
**
*****************************************************************************/

#include	"xllp_keypad.h"

/*---------------------------------------------------------------------------
Function: read_scan_code_automatically
Purpose:This functions reads the scan code from the KeyPad controller triggered by 
	setting of the ASACT bit in the KeyPad Control register. If there is a valid key 
	detected then it returns the scan code. 
Returns:  success/failure.
type no:  I/F proc 3
-----------------------------------------------------------------------------*/

XLLP_STATUS_T read_scan_code_automatically(
			XLLP_KEYPAD_REGS *v_pkeypad_regs,
			XLLP_UINT8_T 	*key_index)
{
	XLLP_BOOL_T	retval = XLLP_STATUS_FAILURE;  /* Default to failure indication*/
	XLLP_UINT32_T	num_of_keys_pressed=0;
	int		scan_code_index;
	/* minimize slow hardware accesses*/
	XLLP_UINT32_T	kp_asr_tmp;

	*key_index = NO_KEY;  /* Default to failure indication*/

	num_of_keys_pressed = 
		((v_pkeypad_regs->kpAutomaticScanReg & MULTI_KEYS_PRESS)>>26);

	if(num_of_keys_pressed == 1)
	{
		kp_asr_tmp = v_pkeypad_regs->kpAutomaticScanReg;
		/* test for valid data*/
		if ((ROW_SELECTED_MASK  == (kp_asr_tmp & ROW_SELECTED_MASK)) ||
		(COL_SELECTED_MASK  == (kp_asr_tmp & COL_SELECTED_MASK))){
			/* If either row or column field == 0xFF, 
			data are invalid*/
			*key_index = NO_KEY;
		}else{
			/* if not invalid, 3 bits of data for each field.
			Compress to 6 bits for index into 
			the 64-entry LUT.*/
			scan_code_index = 
				((kp_asr_tmp & ROW_SELECTED_MASK) >> 1) |
				(kp_asr_tmp &  COL_SELECTED_MASK);

			*key_index= scan_code_index;

			retval = XLLP_STATUS_SUCCESS;
		}
	}else {
		/*current do not support multi key press*/
		*key_index= NO_KEY;
	}/*end of numofkeypressed ==1*/
	return retval;
}
	
/*-------------------------------------------------------------------------
Function: read_direct_keys
Purpose:  This function looks for any Thumbwheel movement or
	button press and returns a scan code. 
Returns:  success/failure.
type no:  I/F proc 2
---------------------------------------------------------------------------*/
XLLP_STATUS_T  read_direct_keys(
	XLLP_KEYPAD_REGS *v_pkeypad_regs,
	XLLP_UINT8_T *key)
{
	XLLP_UINT32_T curr_count,save_kp_rotary_encoder_count_reg;
	static XLLP_UINT32_T prev_count = START_VALUE;

	save_kp_rotary_encoder_count_reg = 
		v_pkeypad_regs->kpRotaryEncoderCountReg; 
	curr_count = save_kp_rotary_encoder_count_reg & COUNT_MASK;

	if(save_kp_rotary_encoder_count_reg & OVERFLOW_ROTARY_ENC_0)
	{
		v_pkeypad_regs->kpRotaryEncoderCountReg = START_VALUE;
		prev_count   = START_VALUE;
		*key    = SCAN_CODE_SCROLL_UP;/* Scroll Up*/
	}
	else if(save_kp_rotary_encoder_count_reg & UNDERFLOW_ROTARY_ENC_0)
	{
		v_pkeypad_regs->kpRotaryEncoderCountReg = START_VALUE;
		prev_count   = START_VALUE;
		*key    = SCAN_CODE_SCROLL_DOWN; /* Scroll Down*/
	}
	else if(curr_count > prev_count)
	{
		*key    = SCAN_CODE_SCROLL_UP;
		prev_count   = curr_count;/* Scroll Up*/
	}
	else if(curr_count < prev_count)
	{
		*key    = SCAN_CODE_SCROLL_DOWN;
		prev_count   = curr_count; /* Scroll Down*/
	}

	return XLLP_STATUS_SUCCESS;
}//end function read_direct_keys

/*---------------------------------------------------------------------------
 Function:	xllp_keypad_config_controller
 Purpose:	This function configures the keypad controller
 Returns:	XLLPSTATUS_T
 Type No:	I/F proc 1
-----------------------------------------------------------------------------*/
void xllp_keypad_config_controller(
	XLLP_KEYPAD_REGS *	p_key_pad_regs, 
	XLLP_BOOL_T	use_default_debounces, 
	XLLP_UINT8_T	matix_debounce_ms, 
	XLLP_UINT8_T	direct_debounce_ms,
	XLLP_UINT8_T	max_row,
	XLLP_UINT8_T	max_col)
{
	/*Init the control regs for direct keys*/

    p_key_pad_regs->kpControlReg = (
		((max_row-1)<< 26) |  
		((max_col-1) << 23) |
		MAT_SCAN_LINE0 | 
		MAT_SCAN_LINE1 | 
		MAT_SCAN_LINE2 | 
		MAT_SCAN_LINE3 | 
		MAT_SCAN_LINE4 | 
		MAT_SCAN_LINE5 | 
		MAT_SCAN_LINE6 | 
		MAT_SCAN_LINE7 | 
		AUTO_SCAN_ON_ACTIVITY |
		MATRIX_INTR_ENABLE | 
		MATRIX_KP_ENABLE | 
		ROTARY_ENCODER_ZERO_DEB |
		DIRECT_KP_INTR_ENABLE | 
		DIRECT_KEY_NUMS |
		DIRECT_KP_ENABLE |
		ROTARY_ENCODER_0_ENABLE) ;    

	p_key_pad_regs->kpRotaryEncoderCountReg = START_VALUE;

	if (use_default_debounces) {
		p_key_pad_regs->kpKeyDebounceInterval = (
			DEFAULT_DIRECT_DEBOUNCE_MS << 8 | 
			DEFAULT_MATRIX_DEBOUNCE_MS);   
	}
	else {
		p_key_pad_regs->kpKeyDebounceInterval = (
		(direct_debounce_ms<<8)
		| matix_debounce_ms);
	}
}

