/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**  FILENAME:   xllp_ost.c
**
**  PURPOSE:    contains XLLP OST primitives.
**
******************************************************************************/

#include "xllp_defs.h"
#include "xllp_ost.h"


//
// Forward references
//
static void XllpOstDelayTicks(P_XLLP_OST_T pOstRegs, XLLP_UINT32_T ticks);


/// Configures the specified OST match register to cause the (compatible)
/// operating system timer to interrupt when the OST count equals that
/// assigned to the specified match register.
/// 
/// Upon completion of this routine, the OST match register is armed 
/// and enabled.
///
/// @param:
///   pOSVirtualAddress  OST  base register addresses.
///   matchreg       -  an enumeration specifiying the OST Match register to 
///                     be configured (match registers 0-3, only)
///   matchvalue     -  the unsigned 32-bit value to assign to the specified
///                     match register
/// @remarks
///		Not touch the Interrupt Controller, so enabling Interrupt Controller should be done by Caller. 
///		In Monahan, interrupt controller can be accessed only by coprocessor 6. So need supervise mode. 
///		Palm OS has interrupt controller node, which also dont expected bypass and directly call another module xllp function.
///

void XllpOstConfigureTimer 
     (P_XLLP_OST_T pOSVirtualAddress, XLLP_OST_MATCHREG matchreg, XLLP_UINT32_T matchvalue)
{

    P_XLLP_OST_T        pOSTRegs = pOSVirtualAddress;
    
    P_XLLP_VUINT32_T    pOSTMatchReg;
    XLLP_UINT32_T       ossrbit,
                        oierbit;
                        

    switch (matchreg)
    {
        default:
        case MatchReg0:
            ossrbit = XLLP_OSSR_M0;
            oierbit = XLLP_OIER_E0;
            pOSTMatchReg = &(pOSTRegs->osmr0);
            break;

        case MatchReg1:
            ossrbit = XLLP_OSSR_M1;
            oierbit = XLLP_OIER_E1;
            pOSTMatchReg = &(pOSTRegs->osmr1);
            break;

        case MatchReg2:
            ossrbit = XLLP_OSSR_M2;
            oierbit = XLLP_OIER_E2;
            pOSTMatchReg = &(pOSTRegs->osmr2);
            break;

        case MatchReg3:
            ossrbit = XLLP_OSSR_M3;
            oierbit = XLLP_OIER_E3;
            pOSTMatchReg = &(pOSTRegs->osmr3);
            break;

		case MatchReg4:
            ossrbit = XLLP_OSSR_M4;
            oierbit = XLLP_OIER_E4;
            pOSTMatchReg = &(pOSTRegs->osmr4);
            break;

		case MatchReg5:
            ossrbit = XLLP_OSSR_M5;
            oierbit = XLLP_OIER_E5;
            pOSTMatchReg = &(pOSTRegs->osmr5);
            break;
    }

    //
    // Disable interrupts on the specified Match register
    //
    pOSTRegs->oier &= ~(oierbit | XLLP_OIER_RESERVED_BITS);

    //
    // Clear any interrupt on the specified Match register
    //
    pOSTRegs->ossr = ossrbit; 

    // 
	// Set up the match register to expire when the oscr0 reaches
    // the next match interval.
	//
    *pOSTMatchReg = matchvalue; 

    //
    // Enable the Match register interrupt on 
    //
    pOSTRegs->oier|= oierbit;

    return;

}




/// XllpOstConfigureMatchReg assigns the sum of the specified matchincrement value
///
/// XllpOstConfigureMatchReg assigns the sum of the specified matchincrement value
/// and the (compatible) Operating System timer to the specified match register. 
/// An interrupt will occur when the compatible OST counter equals that assigned
/// to the specified match register.  
/// 
///	@par
/// Upon completion of this routine, the OST match register is armed 
/// and enabled.
///
///	@par
/// This is a variation of the XllpOstConfigureTimer().  The enhancement being that the
/// match increment value is computed here, saving the caller from doing it.
///
/// @param:
///   pOSVirtualAddress		OST  base register addresses.
///
///   matchreg				an enumeration specifiying the OST Match register to 
///							be configured (match registers 0-3, only)
///
///   matchincrement		this value and the compatible OS timer (OSCR0) are added.  The
///							 resulting value is assigned to the specified OST Match register. 
///@remarks
///		Not touch the Interrupt Controller, so enabling Interrupt Controller should be done by Caller. 
///		In Monahan, interrupt controller can be accessed only by coprocessor 6. So need supervise mode. 
///		Palm OS has interrupt controller node, which also dont expected bypass and directly call another module xllp function.
///

void XllpOstConfigureMatchReg
  (P_XLLP_OST_T pOSVirtualAddress, XLLP_OST_MATCHREG matchreg, XLLP_UINT32_T matchincrement)
{
    P_XLLP_OST_T        pOSTRegs = pOSVirtualAddress;
    XLLP_UINT32_T       matchvalue;

    //
    // Compute the new match value to load
    //
    matchvalue = pOSTRegs->oscr0 + matchincrement;

    //
    // Configure the timer to interrupt at that match value
    //
    XllpOstConfigureTimer (pOSVirtualAddress, matchreg, matchvalue);

}


/// Provides a delay in microseconds using the OST OSCR0 clock.
///
/// @param
///
///   pOstRegs          OST  base register addresses
///
///   microseconds      the unsigned 32-bit value containing the  number of
///                     microseconds to delay
///
void XllpOstDelayMicroSeconds 
     (P_XLLP_OST_T pOstRegs, XLLP_UINT32_T microseconds)
{
    XLLP_UINT32_T    ticks;

    ticks = microseconds * XLLP_OST_TICKS_US;  // approx. 3 ticks per microsecond. 
    XllpOstDelayTicks (pOstRegs, ticks);

}


/// Provides a delay using the OST OSCR0 clock.
///
/// @param
///   pOstRegs          OST  base register addresses
///
///   milliseconds      the unsigned 32-bit value containing the number of
///                     milliseconds to delay
///
void XllpOstDelayMilliSeconds 
     (P_XLLP_OST_T pOstRegs, XLLP_UINT32_T milliseconds)
{
    XLLP_UINT32_T    ticks;

    ticks = milliseconds * XLLP_OST_TICKS_MS;
    XllpOstDelayTicks (pOstRegs, ticks);
    return;
}

  
/// Delay the number of ticks specified using the OST OSCR0 clock.
///
/// @param
///   pOstRegs       OST  base register addresses
///
///   ticks          the unsigned 32-bit value containing the number of
///                  ticks to delay
///
static void XllpOstDelayTicks 
     (P_XLLP_OST_T pOstRegs, XLLP_UINT32_T ticks)
{    
    XLLP_UINT32_T    start_time;
    start_time = pOstRegs->oscr0;

    //
    // Unsigned math handles any wraparound.
    // Cant simplify the test because of possible wraparound.
    // Good for up to 2^32-1 ticks in a bare-metal 
    //     environment where the thread of execution 
    //     is never suspended.
    //

    while ((pOstRegs->oscr0 - start_time) < ticks);

	return;
}

/// Save state before system enter sleep state
///
/// @param 
///     pVirtual    OST base register addresses
///     pSave       Buff to save status
///
/// @return
///     0   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
XLLP_STATUS_T XllpOSTSave(P_XLLP_OST_T pVirtual, P_XLLP_OST_T pSave)
{
    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;

    *pSave = *pVirtual;

    return XLLP_STATUS_SUCCESS;
}

/// clear OSSR and restore state After system wakeup
///
/// @param 
///     pVirtual    OST base register addresses
///     pSave       Buff to save status
///
/// @return
///     0   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
XLLP_STATUS_T XllpOSTRestore(P_XLLP_OST_T pVirtual, P_XLLP_OST_T pSave)
{
    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;
    //OSSR is write 1 clear, clear OSSR register
    pSave->ossr = 0xfff ;
    *pVirtual=*pSave;

    return XLLP_STATUS_SUCCESS;
}
