/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/** FILENAME:	xllp_pm_dvfm.c
**
** PURPOSE:	contain APIs for frequency/voltage change.
**
******************************************************************************/

#include "xllp_pm_dvfm.h"
#include "xllp_pm_arava.h"

#define ASSERT_XL(xl)	(xl==0x4 || xl==0x8 || xl==0xc || xl==0x10 || xl== 0x18 || xl==0x1f)
#define ASSERT_XN(xn)	(xn==0x1 || xn==0x2 || xn==0x4)
#define ASSERT_MTS(mts)	(mts==0x1 || mts==0x2 || mts==0x4)

#define MIN(x, y) (((x) < (y))?(x):(y))
#define MAX(x, y) (((x) > (y))?(x):(y))

/* 
 * silicon bug 1605466: Initial design of the Monahans processor does 
 * not allow for more than 150mV maximum delta
 * between VCC_SRAM and VCC_APPS when caches are enabled. 
 * This requirement is in addition
 * to any requirements currently specified in the Monahans processor EMTS.
 *
 * Implication: Not keeping VCC_APPS and VCC_SRAM sufficiently close does not 
 * damage the cache in the physical sense. However, writes and/or reads with L1
 * or L2 caches might exhibit data failures. It exists in Monahans A0/A1
 */
#undef SI_BUG_1605466

/*
 * silicon bug 1636524: Continuous T-bit writes will cause the Monahans device 
 * to randomly hang if caches are on.
 *
 * Implication: T-bit and F-bit are not functional on Monahans A0/A1.
 */
#undef SI_BUG_1636524

extern int XllpPmEnterD0CS_a(XLLP_VUINT32_T *clkmgr, XLLP_VUINT32_T *dmemc);
extern int XllpPmExitD0CS_a(XLLP_VUINT32_T *clkmgr, XLLP_VUINT32_T *dmemc);

/*
 * get which frequencies and voltages will be changed
 */
XLLP_UINT32_T XllpPmFreqVolPreSet(
			P_XLLP_PM_FV_INFO cur, 
			P_XLLP_PM_FV_INFO next, 
			XLLP_PM_FV_MODE *pmode)
{
	XLLP_UINT32_T flag = 0, mode = *pmode;

	if (cur->vcc_core != next->vcc_core || cur->vcc_sram != next->vcc_sram)
		flag |= XLLP_FV_SET_CV;	
	if (cur->df_clk!= next->df_clk)
		flag |= XLLP_FV_SET_DFCLK;
	if (cur->empi_clk!= next->empi_clk)
		flag |= XLLP_FV_SET_EMPICLK;
	if (cur->d0cs != next->d0cs)
		flag |= XLLP_FV_SET_D0CS;

	if (next->d0cs)
		goto exit;/* don't change ACCR frequencies in D0CS mode */
	
	if (cur->xl != next->xl || cur->xn != next->xn)
		flag |= XLLP_FV_SET_CFS;
	if (cur->smcfs != next->smcfs)
		flag |= XLLP_FV_SET_SMCFS;
	if (cur->sflfs != next->sflfs)
		flag |= XLLP_FV_SET_SFLFS;
	if (cur->hss != next->hss)
		flag |= XLLP_FV_SET_HSS;
	if (cur->dmcfs != next->dmcfs)
		flag |= XLLP_FV_SET_DMCFS;

exit:
	if (!(flag & XLLP_FV_SET_CFS))
		mode = (mode & ~XLLP_PM_FV_MODE_SET_CFS_XSPCLK_MASK)
			| XLLP_PM_FV_MODE_SET_CFS_XSPCLK_DIS;
			
	*pmode = mode;
	
	return flag;	
}

/* 
 * This function is used to change the frequencies. 
 * Both frequencies and voltages changes will be performed here.
 */
XLLP_STATUS_T XllpPmFreqVolChange(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO cur,
			P_XLLP_PM_FV_INFO next,
			XLLP_PM_FV_MODE mode)
{
	XLLP_STATUS_T ret;
	XLLP_PM_FV_MODE fv_mode = mode;
	XLLP_UINT32_T flag;
	
	/* determine if voltages needs to be changed */
	flag = XllpPmFreqVolPreSet(cur, next, &fv_mode);	

	if ((flag & XLLP_FV_SET_CV) && (next->vcc_core >= cur->vcc_core))
	{
		ret = XllpPmVolChange(p_pm_dvfm_context, next->vcc_core, next->vcc_sram);
		if (ret)
			return ret;
	}
	
	ret = XllpPmFreqChange(p_pm_dvfm_context, cur, next, fv_mode);
	if (ret)
		return ret;

	if ((flag & XLLP_FV_SET_CV) && (next->vcc_core < cur->vcc_core))
	{
		ret = XllpPmVolChange(p_pm_dvfm_context, next->vcc_core, next->vcc_sram);
		if (ret)
			return ret;
	}
	
	return XLLP_STATUS_SUCCESS;
}

/* 
 * This function is used to change the frequencies except HSS. 
 * Both frequencies and voltages changes will be performed here.
 */
XLLP_STATUS_T XllpPmFreqVolChangeNoHss(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO cur,
			P_XLLP_PM_FV_INFO next,
			XLLP_PM_FV_MODE mode)
{
	XLLP_STATUS_T ret;
	XLLP_PM_FV_MODE fv_mode = mode;
	XLLP_UINT32_T flag;

	/* determine if voltages needs to be changed */
	flag = XllpPmFreqVolPreSet(cur, next, &fv_mode);	

	if ((flag & XLLP_FV_SET_CV) && (next->vcc_core >= cur->vcc_core))
	{
		ret = XllpPmVolChange(p_pm_dvfm_context, next->vcc_core, next->vcc_sram);
		if (ret)
			return ret;
	}
	
	ret = XllpPmFreqChangeNoHss(p_pm_dvfm_context, cur, next, fv_mode);
	if (ret)
		return ret;

	if ((flag & XLLP_FV_SET_CV) && (next->vcc_core < cur->vcc_core))
	{
		ret = XllpPmVolChange(p_pm_dvfm_context, next->vcc_core, next->vcc_sram);
		if (ret)
			return ret;
	}
	
	return XLLP_STATUS_SUCCESS;
}


/*
 * This function is used current frequencies and 
 * voltages information. Both frequencies and voltages 
 * info will be obtained here.
 */
XLLP_STATUS_T XllpPmFreqVolGet (
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO info,
			XLLP_PM_FV_MODE *mode)
{
	XLLP_STATUS_T ret;

	ret = XllpPmFreqGet(p_pm_dvfm_context, info, mode);
	if (ret)
		return ret;

	ret = XllpPmVolGet(p_pm_dvfm_context, &info->vcc_core, &info->vcc_sram);
	if (ret)
		return ret;

	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to change HSS. It is used by 
 * OS LCD driver's interrupt handler to call it when 
 * handling LCD EOF interrupts.
 */
XLLP_STATUS_T XllpPmHssChange(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			XLLP_UINT32_T hss)
{
	P_XLLP_CLKMGR_T	clkmgr;
	XLLP_VUINT32_T  accr;
	
	clkmgr = p_pm_dvfm_context->clkmgr;
	clkmgr->accr = XLLP_ACCR_S_HSS(clkmgr->accr, hss);
	accr = clkmgr->accr;

	/* 
	 * wait for HSS change completed 
	 * FIXME: do we need to add timeoout here?
	 */
	while((clkmgr->acsr & XLLP_ACCR_HSS_MASK)!= 
			(accr & XLLP_ACCR_HSS_MASK));	
	
	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to change the frequencies,
 * including HSS. Voltage changes won't be performed here. 
 */
XLLP_STATUS_T XllpPmFreqChange(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO cur,
			P_XLLP_PM_FV_INFO next,
			XLLP_PM_FV_MODE mode)
{
	XLLP_STATUS_T ret;
	XLLP_UINT32_T flag;
	XLLP_PM_FV_MODE fv_mode = mode;

	ret = XllpPmFreqChangeNoHss(p_pm_dvfm_context, cur, next, mode);
	if (ret)
		return ret;
	
	/* determine if HSS or D0CS needs operation */
	flag = XllpPmFreqVolPreSet(cur, next, &fv_mode);
	if (flag & XLLP_FV_SET_HSS) 
	{
		ret = XllpPmHssChange(p_pm_dvfm_context, next->hss);
		if (ret)
			return ret;
	}	

	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to change the frequencies except HSS.
 */
XLLP_STATUS_T XllpPmFreqChangeNoHss(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO cur,
			P_XLLP_PM_FV_INFO next,
			XLLP_PM_FV_MODE mode)
{
	XLLP_UINT32_T flag;
	XLLP_PM_FV_MODE fv_mode = mode;
	P_XLLP_CLKMGR_T	clkmgr;
	P_XLLP_SMEM_REGISTER_T smemc;
	XLLP_VUINT32_T accr;
	XLLP_VUINT32_T memclkcfg;
	P_XLLP_DMEM_REGISTER_T dmemc;
	XLLP_VUINT32_T mdrefr;
	
	/* determine how many frequecies and voltage needs to be changed */
	flag = XllpPmFreqVolPreSet(cur, next, &fv_mode);

	clkmgr = p_pm_dvfm_context->clkmgr;
	smemc = p_pm_dvfm_context->smemc;
	dmemc = p_pm_dvfm_context->dmemc;
	
	/* change core frequency */
	if (flag & XLLP_FV_SET_CFS) 
	{
		clkmgr->accr = XLLP_ACCR_S_XSPCLK(clkmgr->accr, 
								(fv_mode & XLLP_PM_FV_MODE_SET_CFS_XSPCLK_MASK) 
								>> XLLP_PM_FV_MODE_SET_CFS_XSPCLK_OFFSET);
		accr = clkmgr->accr;

		/* only enter/exit turbo mode? */
		accr = clkmgr->accr;
		accr = XLLP_ACCR_S_XN(accr, next->xn);
		accr = XLLP_ACCR_S_XL(accr, next->xl);
		clkmgr->accr = accr;
		accr = clkmgr->accr;/* readback */
		if (next->xl == cur->xl && next->xn != cur->xn) 
		{
#ifdef SI_BUG_1636524			
			XllpWriteClkCfg(XLLP_XCLKCFG_SET(1, 0));
#else
			XllpWriteClkCfg(XLLP_XCLKCFG_SET(0, 1));
#endif			
		}
		else
		{
			XllpWriteClkCfg(XLLP_XCLKCFG_SET(1, 0));
		}

		/* 
		 * wait for core frequency change completed 
		 * FIXME: do we need to add timeoout here?
		 */
		while((clkmgr->acsr & (XLLP_ACCR_XL_MASK | XLLP_ACCR_XN_MASK)) != 
			  (accr & (XLLP_ACCR_XL_MASK | XLLP_ACCR_XN_MASK)));

		/* refresh SDRAM after core frequency change*/
		mdrefr = dmemc->mdrefr;
		dmemc->mdrefr = mdrefr;
	}

	/* change static memory controller frequency */
	if (flag & XLLP_FV_SET_SMCFS)
	{
		accr = XLLP_ACCR_S_SMCFS(accr, next->smcfs);
		clkmgr->accr = accr;
		accr = clkmgr->accr;/* readback */
		/* 
		 * wait for core frequency change completed 
		 * FIXME: do we need to add timeoout here?
		 */
		while((clkmgr->acsr & XLLP_ACCR_SMCFS_MASK)!= 
			(accr & XLLP_ACCR_SMCFS_MASK));
	}

	/* change iSRAM frequency */
	if (flag & XLLP_FV_SET_SFLFS)
	{
		accr = XLLP_ACCR_S_SFLFS(accr, next->sflfs);
		clkmgr->accr = accr;
		accr = clkmgr->accr;/* readback */
		/* 
		 * wait for core frequency change completed 
		 * FIXME: do we need to add timeoout here?
		 */
		while((clkmgr->acsr & XLLP_ACCR_SFLFS_MASK)!= 
			(accr & XLLP_ACCR_SFLFS_MASK));
	}

	/* change dynamic memory controller frequency */
	if (flag & XLLP_FV_SET_DMCFS) 
	{
		accr = XLLP_ACCR_S_DMCFS(accr, next->dmcfs);
		clkmgr->accr = accr;
		accr = clkmgr->accr;/* readback */
		/* 
		 * wait for core frequency change completed 
		 * FIXME: do we need to add timeoout here?
		 */
		while((clkmgr->acsr & XLLP_ACCR_DMCFS_MASK)!= 
			(accr & XLLP_ACCR_DMCFS_MASK));		
	}
	
	/* change DFI clock */
	if (flag & XLLP_FV_SET_DFCLK)
	{
		memclkcfg = smemc->memclkcfg;
		memclkcfg = XLLP_MEMCLKCFG_S_DFCLKDIV(memclkcfg, next->df_clk);
		smemc->memclkcfg = memclkcfg;
		memclkcfg = smemc->memclkcfg;
	}

	/* change EMPI clock */
	if (flag & XLLP_FV_SET_EMPICLK)
	{
		memclkcfg = smemc->memclkcfg;
		memclkcfg = XLLP_MEMCLKCFG_S_EMPICLKDIV(memclkcfg, next->empi_clk);
		smemc->memclkcfg = memclkcfg;
		memclkcfg = smemc->memclkcfg;
	}

	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to get the frequencies including 
 * HSS frequency. 
 */ 
XLLP_STATUS_T XllpPmFreqGet(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			P_XLLP_PM_FV_INFO info,
			XLLP_PM_FV_MODE *mode)
{
	P_XLLP_CLKMGR_T	clkmgr;
	XLLP_VUINT32_T acsr;
	P_XLLP_SMEM_REGISTER_T smemc;
	XLLP_VUINT32_T memclkcfg;
	
	clkmgr = p_pm_dvfm_context->clkmgr;
	acsr = clkmgr->acsr;
	smemc = p_pm_dvfm_context->smemc;
	memclkcfg = smemc->memclkcfg;
		
	info->xl = XLLP_ACCR_G_XL(acsr);
	info->xn = XLLP_ACCR_G_XN(acsr);
	info->smcfs = XLLP_ACCR_G_SMCFS(acsr);
	info->sflfs = XLLP_ACCR_G_SFLFS(acsr);
	info->hss = XLLP_ACCR_G_HSS(acsr);
	info->dmcfs = XLLP_ACCR_G_DMCFS(acsr);
	info->df_clk = XLLP_MEMCLKCFG_G_DFCLKDIV(memclkcfg);
	info->empi_clk = XLLP_MEMCLKCFG_G_EMPICLKDIV(memclkcfg);
	info->d0cs = XLLP_ACCR_G_D0CS(acsr);
	if (mode)
		*mode = XLLP_ACCR_G_XSPCLK(acsr);

	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to change the VCC_CORE and VCC_SRAM. 
 */
XLLP_STATUS_T XllpPmVolChange(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			XLLP_UINT32_T vcc_core, 
			XLLP_UINT32_T vcc_sram)
{
	XLLP_STATUS_T ret;
	XLLP_UINT32_T cur_vcc_core, cur_vcc_sram;
#ifdef SI_BUG_1605466
	XLLP_UINT32_T next_vcc_core, next_vcc_sram;
#endif

	ret = XllpPmVolGet(p_pm_dvfm_context, &cur_vcc_core, &cur_vcc_sram);
	if (ret)
		return ret;

#ifdef SI_BUG_1605466
	next_vcc_core = cur_vcc_core;
	next_vcc_sram = cur_vcc_sram;
	if (vcc_core >= cur_vcc_core) 
	{
		while(next_vcc_core != vcc_core || next_vcc_sram != vcc_sram)
		{
			if (next_vcc_sram > next_vcc_core)
			{
				if (next_vcc_core != vcc_core)
				{
					if (vcc_core < next_vcc_sram)
					{
						next_vcc_core = vcc_core;
					}
					else
					{
						next_vcc_core = MIN(next_vcc_sram + 100
							- (next_vcc_sram - 850)%25, vcc_core);
					}
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccCore(next_vcc_core);
					if (ret)
						return ret;
				}
				else
				{
					next_vcc_sram = vcc_sram;
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccSram(next_vcc_sram);
					if (ret)
						return ret;				
				}
			}
			else
			{
				if (next_vcc_sram != vcc_sram)
				{
					if (vcc_sram < next_vcc_core)
					{
						next_vcc_sram = vcc_sram;
					}
					else
					{
						next_vcc_sram = MIN(next_vcc_core + 100 
							- (next_vcc_core - 1100)%50, vcc_sram);
					}
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccSram(next_vcc_sram);
					if (ret)
						return ret;				
				}
				else
				{
					next_vcc_core = vcc_core;
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccCore(next_vcc_core);
					if (ret)
						return ret;					
				}
			}
		}
	}
	else{
		while(next_vcc_core != vcc_core || next_vcc_sram != vcc_sram)
		{
			if (next_vcc_sram < next_vcc_core)
			{
				if (next_vcc_core != vcc_core)
				{
					if (vcc_core > next_vcc_sram)
					{
						next_vcc_core = vcc_core;
					}
					else
					{
						next_vcc_core = MAX(next_vcc_sram - 100 
							+ (next_vcc_sram - 850)%25, vcc_core);
					}
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccCore(next_vcc_core);
					if (ret)
						return ret;
				}
				else
				{
					next_vcc_sram = vcc_sram;
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccSram(next_vcc_sram);
					if (ret)
						return ret;				
				}
			}
			else
			{
				if (next_vcc_sram != vcc_sram)
				{
					if (vcc_sram > next_vcc_core)
					{
						next_vcc_sram = vcc_sram;
					}
					else
					{
						next_vcc_sram = MAX(next_vcc_core - 100 
							+ (next_vcc_core - 1100)%50, vcc_sram);
					}
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccSram(next_vcc_sram);
					if (ret)
						return ret;				
				}
				else
				{
					next_vcc_core = vcc_core;
					ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccCore(next_vcc_core);
					if (ret)
						return ret;
				}
			}
		}
	}
#else
	ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccCore(vcc_core);
	if (ret)
		return ret;
	
	ret = p_pm_dvfm_context->pmic_funcs.PmicSetVccSram(vcc_sram);
	if (ret)
		return ret;
#endif

	return XLLP_STATUS_SUCCESS;
}

/*
 * This function is used to get the VCC_CORE and VCC_SRAM.
 */
XLLP_STATUS_T XllpPmVolGet(
			P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context, 
			XLLP_UINT32_T *vcc_core, 
			XLLP_UINT32_T *vcc_sram)
{
	XLLP_STATUS_T ret;

	ret = p_pm_dvfm_context->pmic_funcs.PmicGetVccCore(vcc_core);
	if (ret)
		return ret;

	ret = p_pm_dvfm_context->pmic_funcs.PmicGetVccSram(vcc_sram);
	if (ret)
		return ret;

	return XLLP_STATUS_SUCCESS;
}

/*
 * Put system into D0CS mode. 
 */
XLLP_STATUS_T XllpPmEnterD0CS(P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context)
{
	P_XLLP_CLKMGR_T	clkmgr = p_pm_dvfm_context->clkmgr;
	P_XLLP_DMEM_REGISTER_T dmemc = p_pm_dvfm_context->dmemc;
	XLLP_VUINT32_T mdrefr;
	XLLP_VUINT32_T accr;
	
	if (XllpPmEnterD0CS_a((XLLP_VUINT32_T *)clkmgr, (XLLP_VUINT32_T *)dmemc))
		return XLLP_STATUS_FAILURE;
	
	/* disable system PLL and core PLL */
        /*
	 * delay 60us to ensure that ring oscillator is stable enough for 
	 * D0CKEN_k and D0CKEN_B clocks before turning off two PLLs
	 */
        XllpOstDelayMicroSeconds(p_pm_dvfm_context->ost, 60); 
	accr = clkmgr->accr;
	accr = XLLP_ACCR_S_SPDIS(accr, 0x1);
	accr = XLLP_ACCR_S_XPDIS(accr, 0x1);
	clkmgr->accr = accr;
	accr = clkmgr->accr;
	while((clkmgr->acsr & (XLLP_ACCR_SPDIS_MASK | XLLP_ACCR_XPDIS_MASK)) != 
		(clkmgr->accr & (XLLP_ACCR_SPDIS_MASK | XLLP_ACCR_XPDIS_MASK)));

	/* refresh SDRAM after core frequency change*/
	mdrefr = dmemc->mdrefr;
	dmemc->mdrefr = mdrefr;
	
	return XLLP_STATUS_SUCCESS;
}

/*
 * Let system exit D0CS mode. 
 */
XLLP_STATUS_T XllpPmExitD0CS(P_XLLP_PM_DVFM_CONTEXT_T p_pm_dvfm_context)
{
	P_XLLP_CLKMGR_T	clkmgr = p_pm_dvfm_context->clkmgr;
	P_XLLP_DMEM_REGISTER_T dmemc = p_pm_dvfm_context->dmemc;
	XLLP_VUINT32_T mdrefr;
	XLLP_VUINT32_T accr;
	
	/* enable system PLL and core PLL */
	accr = clkmgr->accr;
	accr = XLLP_ACCR_S_SPDIS(accr, 0x0);
	accr = XLLP_ACCR_S_XPDIS(accr, 0x0);
	clkmgr->accr = accr;
	accr = clkmgr->accr;
	while (clkmgr->acsr & (XLLP_ACCR_SPDIS_MASK | XLLP_ACCR_XPDIS_MASK));

	if (XllpPmExitD0CS_a((XLLP_VUINT32_T *)clkmgr, (XLLP_VUINT32_T *)dmemc))
		return XLLP_STATUS_FAILURE;

	/* refresh SDRAM after core frequency change*/
	mdrefr = dmemc->mdrefr;
	dmemc->mdrefr = mdrefr;
	
	return XLLP_STATUS_SUCCESS;
}

