/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/************************************************************************
FILENAME:	xllp_usbotg.c
PURPOSE:	implement usb otg primitives those are not platform related
************************************************************************/

#include "xllp_usbotg.h"

/***************************************************************************** 
 * FUNCTION:	XllpOtgConfigPort2
 * 
 * DESCRIPTION:	Configures USB host port2 to the desired mode
 *
 * PARAMETERS:
 *		P_XLLP_UDC_REGISTERS_T pUdcRegs - structure for USB client registers
 *		P_USB_OTG_MODE_T mode - the desired mode for USB host port2
 *          
 * RETURNS: 
 *		XLLP_OTG_SUCCESS
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgConfigPort2 (P_XLLP_UDC_REGISTERS_T  pUdcRegs,	
			XLLP_OTG_MODE_T mode)
{
	XLLP_VINT32_T	up2ocr = pUdcRegs->up2ocr;
	XLLP_OTG_ERROR_T status = XLLP_OTG_SUCCESS;

#ifdef XLLP_DEBUG_PARAM_CHECK
	if ( NULL == pUdcRegs )
		return XLLP_OTG_INVALID_PARAMETER;
#endif

	up2ocr &= ~(7 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
	up2ocr &= ~(XLLP_OTG_UP2OCR_HXS | XLLP_OTG_UP2OCR_HXOE);
	up2ocr &= ~(XLLP_OTG_UP2OCR_DPPUE | XLLP_OTG_UP2OCR_DMPUE);
	up2ocr &= ~(XLLP_OTG_UP2OCR_DMPDE|XLLP_OTG_UP2OCR_DPPDE);
	switch(mode) {
	/* Both differential port and single-ended port are turned off */
	case XLLP_USB_OTG_OFF:
		break;
	/* Differential Port is off, Single-Ended Port2 is Non-OTG USB Client */
	case XLLP_USB_NON_OTG_CLIENT_SEP:
		up2ocr |= (2 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
	
	/* Differential Port is off, Single-Ended Port2 is Non-OTG USB Host */	
	case XLLP_USB_NON_OTG_HOST_SEP:
		up2ocr |= (3 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
	
	/* Differential Port is off, Single-Ended Port2 is interfaced to external OTG transceiver
	 * and is directed into USB client controller */	
	case XLLP_USB_EXT_OTG_CLIENT_SEP:
		up2ocr |= (4 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
		
	/* Differential Port is off, Single-Ended Port2 is interfaced to external OTG transceiver 
	 * and is directed into USB host controller */	
	case XLLP_USB_EXT_OTG_HOST_SEP:
		up2ocr |= (5 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
	/* Differential Port is Non-OTG USB client, Single-Ended Port 2 is Non-OTG USB host */ 	
	case XLLP_USB_NON_OTG_HOST_SEP_CLIENT_DP:
		up2ocr |= XLLP_OTG_UP2OCR_HXOE;
		up2ocr |= (3 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
		
	/* Differential Port is Non-OTG USB host, Single-Ended Port 2 is Non-OTG USB client */	
	case XLLP_USB_NON_OTG_CLIENT_SEP_HOST_DP:
		up2ocr |= XLLP_OTG_UP2OCR_HXOE;
		up2ocr |= XLLP_OTG_UP2OCR_HXS;
		up2ocr |= (2 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		break;
	
	/* Differential Port is USB OTG host, Single-Ended Port 2 is 
	 * interfaced to external charge pump */		
	case XLLP_USB_INT_OTG_HOST_DP:
		up2ocr |= XLLP_OTG_UP2OCR_HXOE;
		up2ocr |= XLLP_OTG_UP2OCR_HXS;
		up2ocr |= (7 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		up2ocr |= (XLLP_OTG_UP2OCR_DPPDE | XLLP_OTG_UP2OCR_DMPDE);
		break;
	
	/* Differential Port is USB OTG client, Single-Ended Port 2 is 
	 * interfaced to external charge pump */	
	case XLLP_USB_INT_OTG_CLIENT_DP:
		up2ocr |= XLLP_OTG_UP2OCR_HXOE;
		up2ocr |= (6 << XLLP_OTG_UP2OCR_SEOS_SHIFT);
		up2ocr |= XLLP_OTG_UP2OCR_DPPUE;
		break;
	
	/* Differential Port is Non-OTG USB Client, Single-Ended Port2 is off */	
	case XLLP_USB_NON_OTG_CLIENT_DP:
		up2ocr |= XLLP_OTG_UP2OCR_HXOE;
		break;
		
	/* Differential Port is Non-OTG USB Host, Single-Ended Port2 is off */
	case XLLP_USB_NON_OTG_HOST_DP:
		up2ocr |= (XLLP_OTG_UP2OCR_HXOE | XLLP_OTG_UP2OCR_HXS);
		break;
		
	default:
		status = XLLP_OTG_INVALID_PARAMETER;
		break;
	}
	pUdcRegs->up2ocr = up2ocr;
	
	return status;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgRequireBus
 * 
 * DESCRIPTION:	Set field b_bus_required in XLLP_OTG_HANDLE_T to indicate
 *		B-Device hope to control bus.
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - the OTG handler
 *		XLLP_BOOL_T bRequire - indicate require or release bus
 *		XLLP_BOOL_T bSRP - indicate whethere srp is initiated 
 *		
 *          
 * RETURNS: 
 *		XLLP_OTG_SUCCESS
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgRequireBus (P_XLLP_OTG_T pOtgHandle, 
		XLLP_BOOL_T bRequire, XLLP_BOOL_T bSRP)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if ( NULL == pOtgHandle )
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if ( XLLP_TRUE == bRequire ) {
		pOtgHandle->b_bus_required = 1;
		/* If state is idle send SRP, tell host to wake up and 
		 * power VBUS */
		if ((pOtgHandle->state == XLLP_OTG_B_IDLE) && bSRP)
			XllpOtgVbusSRP(pOtgHandle);
	} else {
		pOtgHandle->b_bus_required = 0;
	}
		
	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgSetBHNPEnable 
 * 
 * DESCRIPTION:	Set or clear field a_set_b_hnp_en in XLLP_OTG_HANDLE_T
 *
 * PARAMETERS:
 *		P_XLLP_UDC_REGISTERS_T pUdcRegs - structure for USB client registers
 *		P_USB_OTG_MODE_T mode - the desired mode for USB host port2
 *          
 * RETURNS: 
 *		XLLP_OTG_SUCCESS
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgSetBHNPEnable (P_XLLP_OTG_T pOtgHandle, XLLP_BOOL_T bSet)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if ( NULL == pOtgHandle )
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if ( XLLP_TRUE == bSet ) {
		pOtgHandle->a_set_b_hnp_en = 1;
	} else {
		pOtgHandle->a_set_b_hnp_en = 0;
	}
		
	return XLLP_OTG_SUCCESS;	
}

/***************************************************************************** 
 * FUNCTION:	OtgTimerCallbackAIdleBdis
 * 
 * DESCRIPTION:	Internal function called when timer a_aidl_bdis_tmr expires
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		NONE	
 *
 *****************************************************************************/
static void OtgTimerCallbackAIdleBdis (unsigned long data)
{
	XLLP_OTG_T *pOtgHandle = (P_XLLP_OTG_T) data;
	if (XLLP_OTG_A_SUSPEND != pOtgHandle->state) {
		return;
	}
	/* According to OTG specification, Vbus should be turn off
	 * and A device become idle. In our implement, SRP is not feasible
	 * A device switching to a_wait_bcon is better */
#if 0	
	/* Turn off Vbus and return to idle state */
	XllpOtgVbusEnable(pOtgHandle, XLLP_FALSE);
	pOtgHandle->state = XLLP_OTG_A_IDLE;
#else
	pOtgHandle->state = XLLP_OTG_A_WAIT_BCON;
	XllpOtgStateChange(pOtgHandle);
#endif
	XllpOtgTimerExpired(pOtgHandle->a_aidl_bdis_tmr);
	pOtgHandle->a_aidl_bdis_tmr = -1;
}


/***************************************************************************** 
 * FUNCTION:	XllpOtgSuspendDevice 
 * 
 * DESCRIPTION:	This function is called after Host suspend USB device
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_SUCCESS
 *		XLLP_OTG_INVALID_PARAMETER
 *		XLLP_OTG_TIMER_FAILED
 *
 *****************************************************************************/

XLLP_OTG_ERROR_T XllpOtgSuspendDevice (P_XLLP_OTG_T pOtgHandle)
{
	XLLP_INT_T timer_id;
	
#ifdef XLLP_DEBUG_PARAM_CHECK
	if ( NULL == pOtgHandle )
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if (XLLP_OTG_A_HOST == pOtgHandle->state) {
		pOtgHandle->state = XLLP_OTG_A_SUSPEND;
		if (pOtgHandle->a_set_b_hnp_en) {
			timer_id = XllpOtgRegisterTimer(pOtgHandle, \
				T_A_AIDL_BDIS, OtgTimerCallbackAIdleBdis);
			if (timer_id < 0) {
				return XLLP_OTG_TIMER_FAILED;
			}
			
			pOtgHandle->a_aidl_bdis_tmr= timer_id;
		}
	}
	
	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	OtgTimerCallbackBasepBrst
 * 
 * DESCRIPTION:	Internal function called when timer b_ase0_brst_tmr expires
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *		XLLP_OTG_WRONG_STATE
 *
 *****************************************************************************/
static void OtgTimerCallbackBasepBrst (unsigned long data)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if (0 == data) 
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	P_XLLP_OTG_T pOtgHandle = (P_XLLP_OTG_T) data;
	if (XLLP_OTG_B_WAIT_ACON == pOtgHandle->state) {
		XllpOtgConfigPort2(pOtgHandle->pUdcRegs, XLLP_USB_INT_OTG_CLIENT_DP);
		pOtgHandle->state = XLLP_OTG_B_PERIPHERAL;
		XllpOtgStateChange(pOtgHandle);
	}
	XllpOtgTimerExpired(pOtgHandle->b_ase0_brst_tmr);
	pOtgHandle->b_ase0_brst_tmr = -1;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgRespondSuspend
 * 
 * DESCRIPTION:	Called by USB client driver when host suspends device
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgRespondSuspend (P_XLLP_OTG_T pOtgHandle)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if (NULL == pOtgHandle) 
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if (XLLP_OTG_B_PERIPHERAL == pOtgHandle->state) {
		if ( XLLP_UDCCR_BHNP & pOtgHandle->pUdcRegs->udccr) {
			pOtgHandle->b_hnp_enabled = 1;
		} else
			return XLLP_OTG_SUCCESS;

		if (pOtgHandle->b_bus_required) {	
			XllpOtgConfigPort2(pOtgHandle->pUdcRegs, XLLP_USB_INT_OTG_HOST_DP);
			pOtgHandle->state = XLLP_OTG_B_WAIT_ACON;
			/* Start timer Tb_ase0_brst_tmr */
			pOtgHandle->b_ase0_brst_tmr = 
				XllpOtgRegisterTimer(pOtgHandle, T_B_ASE0_BRST,
						OtgTimerCallbackBasepBrst);
		}
	}
	
	/* B-Device completes using the bus, A Device return to be host */
	if (XLLP_OTG_A_PERIPHERAL == pOtgHandle->state) {
		XllpOtgConfigPort2(pOtgHandle->pUdcRegs, XLLP_USB_INT_OTG_HOST_DP);
		pOtgHandle->state = XLLP_OTG_A_WAIT_BCON;
	}

	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgRespondConnect
 * 
 * DESCRIPTION:	Called by USB host controller driver when a device connects
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgRespondConnect (P_XLLP_OTG_T pOtgHandle)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if (NULL == pOtgHandle) 
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if (XLLP_OTG_A_WAIT_BCON == pOtgHandle->state) {
		XllpOtgCancelTimer(pOtgHandle->a_wait_bcon_tmr);
		pOtgHandle->a_wait_bcon_tmr = -1;
		pOtgHandle->state = XLLP_OTG_A_HOST;
	} else if (XLLP_OTG_B_WAIT_ACON == pOtgHandle->state) {
		XllpOtgCancelTimer(pOtgHandle->b_ase0_brst_tmr);
		pOtgHandle->b_ase0_brst_tmr = -1;
		pOtgHandle->state = XLLP_OTG_B_HOST;
	} else 
		return XLLP_OTG_WRONG_STATE;

	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgRespondDisconnect
 * 
 * DESCRIPTION:	Called by USB host controller driver when a device disconnects
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgRespondDisconnect (P_XLLP_OTG_T pOtgHandle)
{
#ifdef XLLP_DEBUG_PARAM_CHECK
	if (NULL == pOtgHandle) 
		return XLLP_OTG_INVALID_PARAMETER;
#endif
	if (XLLP_OTG_A_HOST == pOtgHandle->state) {
		pOtgHandle->state = XLLP_OTG_A_WAIT_BCON;
		/* Start timer */
	}
	
	if (XLLP_OTG_B_HOST == pOtgHandle->state) {
		XllpOtgConfigPort2(pOtgHandle->pUdcRegs, XLLP_USB_INT_OTG_CLIENT_DP);
		pOtgHandle->state = XLLP_OTG_B_PERIPHERAL;
		XllpOtgRequireBus(pOtgHandle, XLLP_FALSE, 0);
	}
	
	if ((XLLP_OTG_A_SUSPEND == pOtgHandle->state) && pOtgHandle->a_set_b_hnp_en) {
		XllpOtgConfigPort2(pOtgHandle->pUdcRegs, XLLP_USB_INT_OTG_CLIENT_DP);
		pOtgHandle->state = XLLP_OTG_A_PERIPHERAL;
		XllpOtgCancelTimer(pOtgHandle->a_aidl_bdis_tmr);
		pOtgHandle->a_aidl_bdis_tmr = -1;
	}
	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgSuspend
 * 
 * DESCRIPTION:	Called by OTG driver, it is a power management function.
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *
 *****************************************************************************/
XLLP_OTG_ERROR_T XllpOtgSuspend (P_XLLP_OTG_T pOtgHandle)
{
	return XLLP_OTG_SUCCESS;
}

/***************************************************************************** 
 * FUNCTION:	XllpOtgResume
 * 
 * DESCRIPTION:	Called by OTG driver, it is a power management function.
 *
 * PARAMETERS:
 *		P_XLLP_OTG_T pOtgHandle - a pointer to USB OTG's handle
 *          
 * RETURNS: 
 *		XLLP_OTG_INVALID_PARAMETER
 *
 *
 *****************************************************************************/	
XLLP_OTG_ERROR_T XllpOtgResume (P_XLLP_OTG_T pOtgHandle)
{
	return XLLP_OTG_SUCCESS;
}

