/*
 *  Copyright (C) Compaq Computer Corporation, 1998, 1999
 *  Copyright (C) Extenex Corporation 2001
 *  Copyright (C) Intrinsyc, Inc., 2002
 *
 *  usb_ctl.h
 *
 *  PRIVATE interface used to share info among components of the PXA USB
 *  core: usb_ctl, usb_ep0, usb_recv and usb_send. Clients of the USB core
 *  should use pxa_usb.h.
 *
 *  02-May-2002
 *   Frank Becker (Intrinsyc) - derived from sa1100 usb_ctl.h
 *
 */

#ifndef _USB_CTL_H
#define _USB_CTL_H

/* Interrupt mask bits and UDC enable bit */
#define UDCCR_MASK_BITS         (UDCCR_REM | UDCCR_SRM | UDCCR_UDE)

/*
 * These states correspond to those in the USB specification v1.0
 * in chapter 8, Device Framework.
 */
enum { 
	USB_STATE_NOTATTACHED	=0,
	USB_STATE_ATTACHED	=1,
	USB_STATE_POWERED	=2,
	USB_STATE_DEFAULT	=3,
	USB_STATE_ADDRESS	=4,
	USB_STATE_CONFIGURED	=5,
	USB_STATE_SUSPENDED	=6
};

struct usb_stats_t {
	 unsigned long ep0_fifo_write_failures;
	 unsigned long ep0_bytes_written;
	 unsigned long ep0_fifo_read_failures;
	 unsigned long ep0_bytes_read;
};

struct usb_info_t
{
	 char * client_name;
	 int state;
	 unsigned char address;
	 struct usb_stats_t stats;
};

/* in usb_ctl.c */
extern struct usb_info_t usbd_info;

/*
 * Function Prototypes
 */
enum { 
	kError		=-1,
	kEvSuspend	=0,
	kEvReset	=1,
	kEvResume	=2,
	kEvAddress	=3,
	kEvConfig	=4,
	kEvDeConfig	=5 
};
int usbctl_next_state_on_event( int event );

/* endpoint zero */
void ep0_reset(void);
void ep0_int_hndlr(void);

/* receiver */
void ep2_state_change_notify( int new_state );
int  ep2_recv(char *buf, int len, usb_callback_t callback);
int  ep2_init(int chn);
void ep2_int_hndlr(int status);
void ep2_reset(void);
void ep2_stall(void);

/* xmitter */
void ep1_state_change_notify( int new_state );
int  ep1_send(char *buf, int len, usb_callback_t callback);
void ep1_reset(void);
int  ep1_init(int chn);
void ep1_int_hndlr(int status);
void ep1_stall(void);

/* Bulverde USB register's definition */
#define UDCCR        __REG(0x40600000)
#define UDCICR0      __REG(0x40600004)
#define UDCICR1      __REG(0x40600008)
#define UDCISR0      __REG(0x4060000c)
#define UDCISR1      __REG(0x40600010)
#define UDCFNR       __REG(0x40600014)
#define UP2OCR       __REG(0x40600020)

#define UDCCSR0      __REG(0x40600100)
#define UDCCSRA      __REG(0x40600104)
#define UDCCSRB      __REG(0x40600108)
/* Omit endpoint C~X */

#define UDCBCR0      __REG(0x40600200)
#define UDCBCRA      __REG(0x40600204)
#define UDCBCRB      __REG(0x40600208)
/* Omit endpoint C~X */

#define UDCDR0       __REG(0x40600300)
#define UDCDRA       __REG(0x40600304)
#define UDCDRB       __REG(0x40600308)
#define UDCWAKEUP    __REG(0x40F00044)

/* Omit endpiont C~X */

#define UDCCRA       __REG(0x40600404)
#define UDCCRB       __REG(0x40600408)
/* Omit endpoint C~X */

/* USB register bit definitions */

#define UDCCR_UDE    (1 << 0)
#define UDCCR_UDA    (1 << 1)
#define UDCCR_UDR    (1 << 2)
#define UDCCR_EMCE   (1 << 3)
#define UDCCR_SMAC   (1 << 4)
#define UDCCR_AAISN  (1 << 5)
#define UDCCR_AIN    (1 << 8)
#define UDCCR_ACN    (1 << 11)
#define UDCCR_DWRE   (1 << 16)

#define UDCICR0_IE0  (1 << 0)
#define UDCICR0_IEA  (1 << 2)
#define UDCICR0_IEB  (1 << 4)
/* Omit endpoint C~X */

#define UDCICR1_IERS (1 << 27)
#define UDCICR1_IESU (1 << 28)
#define UDCICR1_IERU (1 << 29)
#define UDCICR1_IESOF (1 << 30)
#define UDCICR1_IECC (1 << 31)

#define UDCISR0_IR0  (1 << 0)
#define UDCISR0_IRA  (1 << 2)
#define UDCISR0_IRB  (1 << 4)
/* Omit endpiont C~X */

#define UDCISR1_IRRS (1 << 27)
#define UDCISR1_IRSU (1 << 28)
#define UDCISR1_IRRU (1 << 29)
#define UDCISR1_IRSOF (1 << 30)
#define UDCISR1_IRCC (1 << 31)

#define UP2OCR_CPVEN    (1 << 0)        /* Charge Pump Vbus Enable */
#define UP2OCR_CPVPE    (1 << 1)        /* Charge Pump Vbus Pulse Enable */
#define UP2OCR_DPPDE    (1 << 2)        /* Host Port 2 Transceiver D+ Pull Down Enable */
#define UP2OCR_DMPDE    (1 << 3)        /* Host Port 2 Transceiver D- Pull Down Enable */
#define UP2OCR_DPPUE    (1 << 4)        /* Host Port 2 Transceiver D+ Pull Up Enable */
#define UP2OCR_DMPUE    (1 << 5)        /* Host Port 2 Transceiver D- Pull Up Enable */
#define UP2OCR_DPPUBE   (1 << 6)        /* Host Port 2 Transceiver D+ Pull Up Bypass Enable */
#define UP2OCR_DMPUBE   (1 << 7)        /* Host Port 2 Transceiver D- Pull Up Bypass Enable */
#define UP2OCR_EXSP     (1 << 8)        /* External Transceiver Speed Control */
#define UP2OCR_EXSUS    (1 << 9)        /* External Transceiver Speed Enable */
#define UP2OCR_IDON     (1 << 10)       /* OTG ID Read Enable */
#define UP2OCR_HXS      (1 << 16)       /* Host Port 2 Transceiver Output Select */
#define UP2OCR_HXOE     (1 << 17)       /* Host Port 2 Transceiver Output Enable */
#define UP2OCR_SEOS     (1 << 24)       /* Single-Ended Output Select */

#define UDCCSR0_OPC  (1 << 0)
#define UDCCSR0_IPR  (1 << 1)
#define UDCCSR0_FTF  (1 << 2)
#define UDCCSR0_DME  (1 << 3)
#define UDCCSR0_SST  (1 << 4)
#define UDCCSR0_FST  (1 << 5)
#define UDCCSR0_RNE  (1 << 6)
#define UDCCSR0_SA   (1 << 7)

#define UDCCSR_FS    (1 << 0)
#define UDCCSR_PC    (1 << 1)
#define UDCCSR_TRN   (1 << 2)
#define UDCCSR_DME   (1 << 3)
#define UDCCSR_SST   (1 << 4)
#define UDCCSR_FST   (1 << 5)
#define UDCCSR_BNE   (1 << 6)
#define UDCCSR_BNF   (1 << 6)
#define UDCCSR_SP    (1 << 7)
#define UDCCSR_FEF   (1 << 8)
#define UDCCSR_DPE   (1 << 9)

#define UDCCRB_EE    (1 << 0)
#define UDCCRB_DE    (1 << 1)
#define UDCCRB_MPS   (1 << 2)
#define UDCCRB_ED    (1 << 12)
#define UDCCRB_ET    (1 << 13)
#define UDCCRB_EN    (1 << 15)
#define UDCCRB_AISN  (1 << 19)
#define UDCCRB_IN    (1 << 22)
#define UDCCRB_CN    (1 << 25)

#endif /* _USB_CTL_H */
