/*
 *  ether-pxausb.c : "eth-over-usb" driver for BLOB
 *
 *  Copyright (c) 2003, Intel Corporation (yu.tang@intel.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/types.h>
#include <blob/time.h>
#include <blob/serial.h>
#include <blob/util.h>
#include <blob/init.h>
#include <blob/pxa_usb.h>
#include <net/ether.h>
#include <net/net.h>

static int usb_rsize=64;
static int usb_wsize=64;

/* cable connection : defined in bvd_usb_ctl.c */
extern int usb_connected;

/* receive buffer management */
static u8  rx_buf[1024];
static int  rx_size = 0;
static int  rx_done = 0 ;

/* transmit buffer management */
static int tx_done = 0 ;

#define ETHERNET_VENDOR_ID 0x8086
#define ETHERNET_PRODUCT_ID 0x7d3

static void usb_eth_tx_callback(int flag, int size)
{
	tx_done = 1;
}

static void usb_eth_rx_callback(int flag, int size)
{
	int left;

	if( flag != 0 ) return;

	if (rx_done) { 
		/* drop */
		rx_done = 0;
		rx_size = 0;
	}

	rx_size += size;

	/* end-of-packet? */
	if(size == usb_rsize) {
		left = sizeof(rx_buf) - rx_size;
		if (!left) 
			goto done;

		ep2_recv(rx_buf+rx_size, (left<usb_rsize)?left:usb_rsize, usb_eth_rx_callback);
		return;
	}

done:
	rx_done = 1;

	/* setup to receive */
	ep2_recv(rx_buf, sizeof(rx_buf), usb_eth_rx_callback);

	return;
}

void usb_driver_reset(void)
{
	desc_t * pdesc = pxa_usb_get_descriptor_ptr();
	config_desc_t *cfg;
	intf_desc_t *intf;
	ep_desc_t *ep;

	/* setup device descriptor */
	pdesc->dev.idVendor	= ETHERNET_VENDOR_ID;
	pdesc->dev.idProduct    = ETHERNET_PRODUCT_ID;
	pdesc->dev.bNumConfigurations = 1;

	cfg = (config_desc_t*) (pdesc->cdb);

	cfg->bLength             = sizeof( config_desc_t );
	cfg->bDescriptorType     = USB_DESC_CONFIG;
	cfg->wTotalLength        = make_word_c( sizeof(config_desc_t) +
						   sizeof(intf_desc_t) * 1+
						   sizeof(ep_desc_t) * 2);
	cfg->bNumInterfaces      = 1;
	cfg->bConfigurationValue = 1;
	cfg->iConfiguration      = 0;
	cfg->bmAttributes        = USB_CONFIG_BUSPOWERED;
	cfg->MaxPower            = USB_POWER( 500 );

	intf = (intf_desc_t *) ( cfg + 1);
	intf->bLength            = sizeof( intf_desc_t );
	intf->bDescriptorType    = USB_DESC_INTERFACE;
	intf->bInterfaceNumber   = 0; 
	intf->bAlternateSetting  = 0;
	intf->bNumEndpoints      = 2;
	intf->bInterfaceClass    = 0xFF; 
	intf->bInterfaceSubClass = 0;
	intf->bInterfaceProtocol = 0;
	intf->iInterface         = 0;

	ep = (ep_desc_t *) (intf + 1);
	ep[0].bLength             = sizeof( ep_desc_t );
	ep[0].bDescriptorType     = USB_DESC_ENDPOINT;
	ep[0].bEndpointAddress    = USB_EP_ADDRESS( 1, USB_IN );
	ep[0].bmAttributes        = USB_EP_BULK;
	ep[0].wMaxPacketSize      = make_word( 64 );
	ep[0].bInterval           = 0;

	ep[1].bLength             = sizeof( ep_desc_t );
	ep[1].bDescriptorType     = USB_DESC_ENDPOINT;
	ep[1].bEndpointAddress    = USB_EP_ADDRESS( 2, USB_OUT );
	ep[1].bmAttributes        = USB_EP_BULK;
	ep[1].wMaxPacketSize      = make_word( 64 );
	ep[1].bInterval           = 0;

	/* reset buffer */
	rx_done = 0;
	rx_size = 0;

	/* setup to receive */
	ep2_recv(rx_buf, sizeof(rx_buf), usb_eth_rx_callback);
}

static int check_usb_connection(void)
{
	int cnt;

	if (usb_connected) return 0;

	/* waiting util connected */
	printf("***** Plug-in USB cable & config usbdnet now ****** \n");
	
	while (!usb_connected) {
		if ( ICPR & 0x800 ) 
			udc_int_hndlr(0x11, 0);
	}

	while ( cnt < 0x1000) {
		if ( ICPR & 0x800 ) {
			udc_int_hndlr(0x11, 0);
			cnt = 0;
		}
		cnt ++;
	}

	return 0;
}

static int usb_eth_init(u32 base)
{
	usb_connected = 0;

	/* initialize PXA USB controller */
	pxa_usb_open("usb-eth");
	pxa_usb_start();

	return 0;
}

static int usb_eth_rx(char *inbuf, int size)
{
	int i=0;

	check_usb_connection();

	for ( i=0; i<0x1000 && !rx_done; i++) {
		if ( ICPR & 0x800 ) {
			udc_int_hndlr(0x11, 0);
			i=0;
		}
	}

	if (size > rx_size) 
		size = rx_size;

	memcpy(inbuf, rx_buf, size);

	rx_done = 0;
	rx_size = 0;

	return size;
}

static int usb_eth_tx(char *outbuf, int size) 
{
	int tx_size = size;

	check_usb_connection();

	if ( 0 == (size % usb_wsize) ) 
		size ++;

	tx_done = 0;

	ep1_send(outbuf, size, usb_eth_tx_callback);

	while (!tx_done) {
		if ( ICPR & 0x800 ) 
			udc_int_hndlr(0x11, 0);
	}
	
	return tx_size;
}

static int usb_eth_set_addr( u8 hw_addr[6] )
{
	memcpy(hwaddress, hw_addr, 6);
	return 0;
}

static int usb_eth_get_addr( u8 hw_addr[6] )
{
	memcpy(hw_addr, hwaddress, 6);
	return 0;
}

/* export ethernet driver */
ether_driver_t bvdusb_ether_driver = {
	init:		usb_eth_init,
	rcv:		usb_eth_rx,
	snd:		usb_eth_tx,
	set_addr:	usb_eth_set_addr,
	get_addr:	usb_eth_get_addr,
	base:		0,
	name:		"usb-eth"
};
