/*------------------------------------------------------------------------
 . ether-smc91x.c
 . This is a driver for SMSC's 91C9x/91C1xx single-chip Ethernet devices.
 .
 . Copyright (C) 1996 by Erik Stahlman
 . Copyright (C) 2001 Standard Microsystems Corporation
 .	Developed by Simple Network Magic Corporation
 . Copyright (C) 2003 Monta Vista Software, Inc.
 .	Unified SMC91x driver by Nicolas Pitre
 .
 *
 * (C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 

 . This program is free software; you can redistribute it and/or modify
 . it under the terms of the GNU General Public License as published by
 . the Free Software Foundation; either version 2 of the License, or
 . (at your option) any later version.
 .
 . This program is distributed in the hope that it will be useful,
 . but WITHOUT ANY WARRANTY; without even the implied warranty of
 . MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 . GNU General Public License for more details.
 .
 . You should have received a copy of the GNU General Public License
 . along with this program; if not, write to the Free Software
 . Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 .
 . Arguments:
 . 	io	= for the base address
 .	irq	= for the IRQ
 .	nowait	= 0 for normal wait states, 1 eliminates additional wait states
 .
 . original author:
 . 	Erik Stahlman <erik@vt.edu>
 .
 . hardware multicast code:
 .    Peter Cammaert <pc@denkart.be>
 .
 . contributors:
 . 	Daris A Nevil <dnevil@snmc.com>
 .      Nicolas Pitre <nico@cam.org>
 .
 . History:
 .   08/20/00  Arnaldo Melo       fix kfree(skb) in smc_hardware_send_packet
 .   12/15/00  Christian Jullien  fix "Warning: kfree_skb on hard IRQ"
 .   03/16/01  Daris A Nevil      modified smc9194.c for use with LAN91C111
 .   08/22/01  Scott Anderson     merge changes from smc9194 to smc91111
 .   08/21/01  Pramod B Bhardwaj  added support for RevB of LAN91C111
 .   12/20/01  Jeff Sutherland    initial port to Xscale PXA with DMA support
 .   04/07/03  Nicolas Pitre      unified SMC91x driver, killed irq races,
 .                                more bus abstraction, big cleanup, etc.
 .   09/07/03  Yu Tang		  ported to BLOB from Linux.
 ----------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/util.h>
#include <blob/time.h>
#include <blob/errno.h>
#include <blob/serial.h>
#include <blob/command.h>
#include <net/ether.h>
#include <net/net.h>
#include <net/smc91x.h>


/***************************************************************************
 * defines
 */
#undef SMC_DEBUG
//#define SMC_DEBUG	1
#ifdef SMC_DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... ) ;
#endif

/*------------------------------------------------------------------------
 .
 . The internal workings of the driver.  If you are changing anything
 . here with the SMC stuff, you should have the datasheet and know
 . what you are doing.
 .
 -------------------------------------------------------------------------*/
#define CARDNAME "LAN91x"

// Use power-down feature of the chip
#define POWER_DOWN		1

/*
 . Wait time for memory to be free.  This probably shouldn't be
 . tuned that much, as waiting for this means nothing else happens
 . in the system
*/
#define MEMORY_WAIT_TIME	16

/*
 . This selects whether TX packets are sent one by one to the SMC91x internal
 . memory ans throttled until transmission completes.  This may prevent
 . RX overruns a litle by keeping much of the memory free for RX packets
 . but to the expense of reduced TX throughput and increased IRQ overhead.
 . Note this is not a cure for a too slow data bus or too high IRQ latency.
 */
#define THROTTLE_TX_PKTS	0

/**********************************************************************
 * Local 
 */
static int smc_chip_base = 0;
static int dbg = 0;

#if SMC_DEBUG
static void HEXDUMP(u8 *buf, int length)
{
	int i;
	int remainder;
	int lines;

	lines = length / 16;
	remainder = length % 16;

	for (i = 0; i < lines ; i ++) {
		int cur;
		for (cur = 0; cur < 8; cur++) {
			u8 a, b;
			a = *buf++;
			b = *buf++;
			printf("%02x%02x ", a, b);
		}
		printf("\n");
	}
	for (i = 0; i < remainder/2 ; i++) {
		u8 a, b;
		a = *buf++;
		b = *buf++;
		printf("%02x%02x ", a, b );
	}
	printf("\n");
}
#else
#define HEXDUMP(x...)  do { } while(0)
#endif

/* this enables an interrupt in the interrupt mask register */
#define SMC_ENABLE_INT(x) do {						\
	unsigned long flags;						\
	unsigned char mask;						\
	mask = SMC_GET_INT_MASK();					\
	mask |= (x);							\
	SMC_SET_INT_MASK(mask);						\
} while (0)

/* this disables an interrupt from the interrupt mask register */
#define SMC_DISABLE_INT(x) do {						\
	unsigned long flags;						\
	unsigned char mask;						\
	mask = SMC_GET_INT_MASK();					\
	mask &= ~(x);							\
	SMC_SET_INT_MASK(mask);						\
} while (0)

/* wait while MMU is busy */
#define SMC_WAIT_MMU_BUSY() do {					\
	if (SMC_GET_MMU_CMD() & MC_BUSY) {				\
		long timeout = 0x1000;					\
		while (SMC_GET_MMU_CMD() & MC_BUSY) {			\
			timeout --;					\
			if (!timeout) {					\
				printf("%s: timeout %s line %d\n",	\
					CARDNAME, __FILE__, __LINE__);	\
				break;					\
			}						\
		}							\
	}								\
} while (0)

#ifdef SMC_DEBUG
static void dump_banks(u32 ioaddr)
{
	int i,j;

	printf("Dump registers in ethernet\n");
	printf("offset\tbank0\tbank1\tbank2\tbank3\n");
	
	for(i=0; i < 8; i++) {
		printf("%d\t", i*2);
		for(j=0; j < 4; j++) {
			SMC_SELECT_BANK(j);
			printf("0x%04x\t", SMC_inw(ioaddr, 2 * i));
		}
		printf("\n");
	}
}
#else
static void dump_banks(u32 ioaddr) {}
#endif

/**********************************************************************
 * Core routines
 */

/* this does a soft reset on the device */
static void smc_reset(u32 ioaddr)
{
	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	/* This resets the registers mostly to defaults, but doesn't
	   affect EEPROM.  That seems unnecessary */
	SMC_SELECT_BANK( 0 );
	SMC_SET_RCR( RCR_SOFTRST );

	/* Setup the Configuration Register */
	/* This is necessary because the CONFIG_REG is not affected */
	/* by a soft reset */
	SMC_SELECT_BANK( 1 );
	SMC_SET_CONFIG( CONFIG_DEFAULT );

	/* Setup for fast accesses if requested */
	/* If the card/system can't handle it then there will */
	/* be no recovery except for a hard reset or power cycle */
	SMC_SET_CONFIG( SMC_GET_CONFIG() | CONFIG_NO_WAIT );

#ifdef POWER_DOWN
	/* Release from possible power-down state */
	/* Configuration register is not affected by Soft Reset */
	SMC_SELECT_BANK( 1 );
	SMC_SET_CONFIG( SMC_GET_CONFIG() | CONFIG_EPH_POWER_EN );
#endif

	/* this should pause enough for the chip to be happy */
	msleep(1);

	/* Disable transmit and receive functionality */
	SMC_SELECT_BANK( 0 );
	SMC_SET_RCR( RCR_CLEAR );
	SMC_SET_TCR( TCR_CLEAR );

	/* set the control register to automatically
	   release successfully transmitted packets, to make the best
	   use out of our limited memory */
	SMC_SELECT_BANK( 1 );
#if ! THROTTLE_TX_PKTS
	SMC_SET_CTL( SMC_GET_CTL() | CTL_AUTO_RELEASE );
#else
	SMC_SET_CTL( SMC_GET_CTL() & ~CTL_AUTO_RELEASE );
#endif

	/* Disable all interrupts */
	SMC_SELECT_BANK( 2 );
	SMC_SET_INT_MASK( 0 );

	/* Reset the MMU */
	SMC_SET_MMU_CMD( MC_RESET );
	SMC_WAIT_MMU_BUSY();
}

/* Enable Interrupts, Receive, and Transmit */
static void smc_enable(u32 ioaddr)
{
	int mask;

	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	/* see the header file for options in TCR/RCR DEFAULT*/
	SMC_SELECT_BANK( 0 );
	SMC_SET_TCR( TCR_DEFAULT);
	SMC_SET_RCR( RCR_DEFAULT);

	/* now, enable interrupts */
	mask = IM_EPH_INT|IM_RX_OVRN_INT|IM_RCV_INT;
	SMC_SELECT_BANK( 2 );
	SMC_SET_INT_MASK( mask );
}

/* this puts the device in an inactive state */
static void smc_shutdown(unsigned long ioaddr)
{
	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	/* no more interrupts for me */
	SMC_SELECT_BANK( 2 );
	SMC_SET_INT_MASK( 0 );

	/* and tell the card to stay away from that nasty outside world */
	SMC_SELECT_BANK( 0 );
	SMC_SET_RCR( RCR_CLEAR );
	SMC_SET_TCR( TCR_CLEAR );

#ifdef POWER_DOWN
	/* finally, shut the chip down */
	SMC_SELECT_BANK( 1 );
	SMC_SET_CONFIG( SMC_GET_CONFIG() & ~CONFIG_EPH_POWER_EN );
#endif
}

/* This is the procedure to handle the receipt of a packet. */
static inline int smc_rcv(u32 ioaddr, u8 *data, u16 size)
{
	unsigned int packet_number, status, packet_len;
	int ret = 0;

	DBG(3, "%s: %s\n", CARDNAME, __FUNCTION__);

	packet_number = SMC_GET_RXFIFO();
	if (packet_number & RXFIFO_REMPTY) {
		DBG(1, "%s: smc_rcv with nothing on FIFO.\n", CARDNAME);
		return;
	}

	/* read from start of packet */
	SMC_SET_PTR( PTR_READ | PTR_RCV | PTR_AUTOINC );

	/* First two words are status and packet length */
	SMC_GET_PKT_HDR(status, packet_len);
	packet_len &= 0x07ff;  /* mask off top bits */
	DBG(2, "%s: RX PNR 0x%x STATUS 0x%04x LENGTH 0x%04x (%d)\n",
		CARDNAME, packet_number, status,
		packet_len, packet_len);

	if (status & RS_ERRORS) {
		printf(" %s : RCV error\n", __FUNCTION__);
		ret = -1;
		goto done;
	}

	if (status & RS_ODDFRAME)
		packet_len ++;

	packet_len -= 6;

	SMC_PULL_DATA(data, packet_len);

	HEXDUMP(data, packet_len);

	ret = packet_len;
done:
	SMC_WAIT_MMU_BUSY();
	SMC_SET_MMU_CMD( MC_RELEASE );

	return ret;
}

/*
 * This is called to actually send a packet to the chip.
 * Returns non-zero when successful.
 */
static void smc_hardware_send_packet( u32 ioaddr , u8 *data , int size)
{
	unsigned int packet_no, len;
	unsigned char *buf;

	DBG(3, "%s: %s\n", CARDNAME, __FUNCTION__);

	packet_no = SMC_GET_AR();
	if (packet_no & AR_FAILED) {
		printf("%s: Memory allocation failed.\n", CARDNAME);
		return;
	}

	/* point to the beginning of the packet */
	SMC_SET_PN( packet_no );
	SMC_SET_PTR( PTR_AUTOINC );

	buf = data;
	len = size;
	DBG(2, "%s: TX PNR 0x%x lENGTH 0x%04x (%d) BUF 0x%p\n",
		CARDNAME, packet_no, len, len, buf);

	HEXDUMP(buf, len);

	/*
	 * Send the packet length ( +6 for status words, length, and ctl.
	 * The card will pad to 64 bytes with zeroes if packet is too small.
	 */
	SMC_PUT_PKT_HDR(0, len + 6);

	/* send the actual data */
	SMC_PUSH_DATA(buf, len & ~1);

	/* Send final ctl word with the last byte if there is one */
	SMC_outw( ((len & 1) ? (0x2000 | buf[len-1]) : 0), ioaddr, DATA_REG );

	/* and let the chipset deal with it */
	SMC_SET_MMU_CMD( MC_ENQUEUE );
}

/*
 . Since I am not sure if I will have enough room in the chip's ram
 . to store the packet, I call this routine which either sends it
 . now, or set the card to generates an interrupt when ready
 . for the packet.
 */
static int
smc_send( u32 ioaddr, u8 *data, int size )
{
	unsigned int numPages, poll_count, status, saved_bank;

	DBG(3, "%s: %s\n", CARDNAME, __FUNCTION__);

	/*
	** The MMU wants the number of pages to be the number of 256 bytes
	** 'pages', minus 1 ( since a packet can't ever have 0 pages :) )
	**
	** The 91C111 ignores the size bits, but the code is left intact
	** for backwards and future compatibility.
	**
	** Pkt size for allocating is data length +6 (for additional status
	** words, length and ctl!)
	**
	** If odd size then last byte is included in ctl word.
	*/
	numPages = ((size & ~1) + (6 - 1)) >> 8;
	if (numPages > 7) {
		printf("%s: Far too big packet error.\n", CARDNAME);
		return 0;
	}

	/* now, try to allocate the memory */
	saved_bank = SMC_CURRENT_BANK();
	SMC_SELECT_BANK( 2 );
	SMC_SET_MMU_CMD( MC_ALLOC | numPages );

	/*
	 * Poll the chip for a short amount of time in case the
	 * allocation succeeds quickly.
	 */
	poll_count = MEMORY_WAIT_TIME;
	do {
		status = SMC_GET_INT();
		if (status & IM_ALLOC_INT) {
			SMC_ACK_INT( IM_ALLOC_INT );
  			break;
		}
   	} while (--poll_count);

 	if (!poll_count) {
		printf("%s: memory allocation time out.\n", __FUNCTION__ );
		return -1;
 	}

	SMC_ENABLE_INT(IM_TX_INT | IM_TX_EMPTY_INT);
	smc_hardware_send_packet( ioaddr, data, size );

	/* waiting for transmit done */
	while ( !((status = SMC_GET_INT()) & (IM_TX_INT|IM_TX_EMPTY_INT)));

	if (status & IM_TX_INT) 
		SMC_ACK_INT(IM_TX_INT);
	if (status & IM_TX_EMPTY_INT) 
		SMC_ACK_INT(IM_TX_EMPTY_INT);

	SMC_SELECT_BANK( saved_bank );

	return 0;
}

/**********************************************************************
 * PHY Control and Configuration
 */

/*------------------------------------------------------------
 . Debugging function for viewing MII Management serial bitstream
 .-------------------------------------------------------------*/
#if SMC_DEBUG 
static void
PRINT_MII_STREAM(u8 *bits, int size)
{
	int i;

	printf("BIT#:");
	for (i = 0; i < size; ++i)
		printf("%d", i%10);

	printf("\nMDOE:");
	for (i = 0; i < size; ++i) {
		if (bits[i] & MII_MDOE)
			printf("1");
		else
			printf("0");
	}

	printf("\nMDO :");
	for (i = 0; i < size; ++i) {
		if (bits[i] & MII_MDO)
			printf("1");
		else
			printf("0");
	}

	printf("\nMDI :");
	for (i = 0; i < size; ++i) {
		if (bits[i] & MII_MDI)
			printf("1");
		else
			printf("0");
	}

	printf("\n");
}
#else
#define PRINT_MII_STREAM(x...)
#endif

/*------------------------------------------------------------
 . Reads a register from the MII Management serial interface
 .-------------------------------------------------------------*/
static int
smc_read_phy_register(unsigned long ioaddr, int phyaddr, int phyreg)
{
	int oldBank;
	int i, mask, mii_reg;
	u8 bits[64];
	int input_idx, phydata;
	int clk_idx = 0;

	// 32 consecutive ones on MDO to establish sync
	for (i = 0; i < 32; ++i)
		bits[clk_idx++] = MII_MDOE | MII_MDO;

	// Start code <01>
	bits[clk_idx++] = MII_MDOE;
	bits[clk_idx++] = MII_MDOE | MII_MDO;

	// Read command <10>
	bits[clk_idx++] = MII_MDOE | MII_MDO;
	bits[clk_idx++] = MII_MDOE;

	// Output the PHY address, msb first
	mask = 0x10;
	for (i = 0; i < 5; ++i) {
		if (phyaddr & mask)
			bits[clk_idx++] = MII_MDOE | MII_MDO;
		else
			bits[clk_idx++] = MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
	}

	// Output the phy register number, msb first
	mask = 0x10;
	for (i = 0; i < 5; ++i) {
		if (phyreg & mask)
			bits[clk_idx++] = MII_MDOE | MII_MDO;
		else
			bits[clk_idx++] = MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
	}

	// Tristate and turnaround (2 bit times)
	bits[clk_idx++] = 0;
	//bits[clk_idx++] = 0;

	// Input starts at this bit time
	input_idx = clk_idx;

	// Will input 16 bits
	for (i = 0; i < 16; ++i)
		bits[clk_idx++] = 0;

	// Final clock bit
	bits[clk_idx++] = 0;

	// Save the current bank
	oldBank = SMC_CURRENT_BANK();

	// Select bank 3
	SMC_SELECT_BANK( 3 );

	// Get the current MII register value
	mii_reg = SMC_GET_MII();

	// Turn off all MII Interface bits
	mii_reg &= ~(MII_MDOE|MII_MCLK|MII_MDI|MII_MDO);

	// Clock all 64 cycles
	for (i = 0; i < sizeof bits; ++i) {
		// Clock Low - output data
		SMC_SET_MII( mii_reg | bits[i] );
		msleep(1);

		// Clock Hi - input data
		SMC_SET_MII( mii_reg | bits[i] | MII_MCLK );
		msleep(1);
		bits[i] |= SMC_GET_MII() & MII_MDI;
	}

	// Return to idle state
	// Set clock to low, data to low, and output tristated
	SMC_SET_MII( mii_reg );
	msleep(1);

	// Restore original bank select
	SMC_SELECT_BANK( oldBank );

	// Recover input data
	phydata = 0;
	for (i = 0; i < 16; ++i) {
		phydata <<= 1;

		if (bits[input_idx++] & MII_MDI)
			phydata |= 0x0001;
	}

	DBG(3, "%s: phyaddr=0x%x, phyreg=0x%x, phydata=0x%x\n",
		__FUNCTION__, phyaddr, phyreg, phydata);
	PRINT_MII_STREAM(bits, sizeof(bits));

	return phydata;
}

/*------------------------------------------------------------
 . Writes a register to the MII Management serial interface
 .-------------------------------------------------------------*/
static void
smc_write_phy_register( unsigned long ioaddr, int phyaddr,
			int phyreg, int phydata )
{
	int oldBank;
	int i, mask, mii_reg;
	u8 bits[65];
	int clk_idx = 0;

	// 32 consecutive ones on MDO to establish sync
	for (i = 0; i < 32; ++i)
		bits[clk_idx++] = MII_MDOE | MII_MDO;

	// Start code <01>
	bits[clk_idx++] = MII_MDOE;
	bits[clk_idx++] = MII_MDOE | MII_MDO;

	// Write command <01>
	bits[clk_idx++] = MII_MDOE;
	bits[clk_idx++] = MII_MDOE | MII_MDO;

	// Output the PHY address, msb first
	mask = 0x10;
	for (i = 0; i < 5; ++i) {
		if (phyaddr & mask)
			bits[clk_idx++] = MII_MDOE | MII_MDO;
		else
			bits[clk_idx++] = MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
	}

	// Output the phy register number, msb first
	mask = 0x10;
	for (i = 0; i < 5; ++i) {
		if (phyreg & mask)
			bits[clk_idx++] = MII_MDOE | MII_MDO;
		else
			bits[clk_idx++] = MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
	}

	// Tristate and turnaround (2 bit times)
	bits[clk_idx++] = 0;
	bits[clk_idx++] = 0;

	// Write out 16 bits of data, msb first
	mask = 0x8000;
	for (i = 0; i < 16; ++i) {
		if (phydata & mask)
			bits[clk_idx++] = MII_MDOE | MII_MDO;
		else
			bits[clk_idx++] = MII_MDOE;

		// Shift to next lowest bit
		mask >>= 1;
	}

	// Final clock bit (tristate)
	bits[clk_idx++] = 0;

	// Save the current bank
	oldBank = SMC_CURRENT_BANK();

	// Select bank 3
	SMC_SELECT_BANK( 3 );

	// Get the current MII register value
	mii_reg = SMC_GET_MII();

	// Turn off all MII Interface bits
	mii_reg &= ~(MII_MDOE|MII_MCLK|MII_MDI|MII_MDO);

	// Clock all cycles
	for (i = 0; i < sizeof bits; ++i) {
		// Clock Low - output data
		SMC_SET_MII( mii_reg | bits[i] );
		msleep(1);

		// Clock Hi - input data
		SMC_SET_MII( mii_reg | bits[i] | MII_MCLK );
		msleep(1);
		bits[i] |= SMC_GET_MII() & MII_MDI;
	}

	// Return to idle state
	// Set clock to low, data to low, and output tristated
	SMC_SET_MII( mii_reg );
	msleep(1);

	// Restore original bank select
	SMC_SELECT_BANK( oldBank );

	DBG(3, "%s: phyaddr=0x%x, phyreg=0x%x, phydata=0x%x\n",
		__FUNCTION__, phyaddr, phyreg, phydata);
	PRINT_MII_STREAM(bits, sizeof(bits));
}


/*------------------------------------------------------------
 . Finds and reports the PHY address
 .-------------------------------------------------------------*/
static int smc_detect_phy(u32 ioaddr, u32 *pphyaddr)
{
	int phy_id1, phy_id2;
	int phyaddr;
	int found = 0;

	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	// Scan all 32 PHY addresses if necessary
	for (phyaddr = 0; phyaddr < 32; ++phyaddr) {
		// Read the PHY identifiers
		phy_id1 = smc_read_phy_register(ioaddr, phyaddr, PHY_ID1_REG);
		phy_id2 = smc_read_phy_register(ioaddr, phyaddr, PHY_ID2_REG);

		DBG(3, "%s: phy_id1=0x%x, phy_id2=0x%x\n",
			CARDNAME, phy_id1, phy_id2);

		// Make sure it is a valid identifier
		if ((phy_id2 > 0x0000) && (phy_id2 < 0xffff) &&
		    (phy_id1 > 0x0000) && (phy_id1 < 0xffff)) {
			if ((phy_id1 != 0x8000) && (phy_id2 != 0x8000)) {
				// Save the PHY's address
				*pphyaddr = phyaddr;
				found = 1;
				break;
			}
		}
	}

	if (!found) {
		DBG(1, "%s: No PHY found\n", CARDNAME);
		return(0);
	}

	// Set the PHY type
	if ( (phy_id1 == 0x0016) && ((phy_id2 & 0xFFF0) == 0xF840 ) ) {
		DBG(1, "%s: PHY=LAN83C183 (LAN91C111 Internal)\n", CARDNAME);
	}

	if ( (phy_id1 == 0x0282) && ((phy_id2 & 0xFFF0) == 0x1C50) ) {
		DBG(1, "%s: PHY=LAN83C180\n", CARDNAME);
	}

	return 1;
}

/*------------------------------------------------------------
 . Configures the specified PHY through the MII management interface
 . using Autonegotiation.
 . Calls smc_phy_fixed() if the user has requested a certain config.
 .-------------------------------------------------------------*/
static void
smc_phy_configure(u32 ioaddr)
{
	int timeout;
	int phyaddr;
	int my_phy_caps; // My PHY capabilities
	int my_ad_caps; // My Advertised capabilities
	int status;
	int failed = 0;
	int rpc_cur_mode = RPC_DEFAULT;

	unsigned char spinner[] = {'|', '/', '-', '\\'};
	int spin = 0;

	DBG(3, "%s:smc_program_phy()\n", CARDNAME);

	// Set the blocking flag
	//lp->autoneg_active = 1;

	// Find the address and type of our phy
	if (!smc_detect_phy(ioaddr, &phyaddr))
		goto smc_phy_configure_exit;

	// Reset the PHY, setting all other bits to zero
	smc_write_phy_register(ioaddr, phyaddr, PHY_CNTL_REG, PHY_CNTL_RST);

	// Wait for the reset to complete, or time out
	timeout = 6; // Wait up to 3 seconds
	while (timeout--) {
		if (!(smc_read_phy_register(ioaddr, phyaddr, PHY_CNTL_REG)
		    & PHY_CNTL_RST))
			// reset complete
			break;
		msleep(500); // wait 500 millisecs
	}

	if (timeout < 1) {
		printf("%s: PHY reset timed out\n", CARDNAME);
		goto smc_phy_configure_exit;
	}

	// Read PHY Register 18, Status Output
	//lp->lastPhy18 = smc_read_phy_register(ioaddr, phyaddr, PHY_INT_REG);

	// Enable PHY Interrupts (for register 18)
	// Interrupts listed here are disabled
	smc_write_phy_register(ioaddr, phyaddr, PHY_MASK_REG,
		PHY_INT_LOSSSYNC | PHY_INT_CWRD | PHY_INT_SSD |
		PHY_INT_ESD | PHY_INT_RPOL | PHY_INT_JAB |
		PHY_INT_SPDDET | PHY_INT_DPLXDET);

	/* Configure the Receive/Phy Control register */
	SMC_SELECT_BANK( 0 );
	SMC_SET_RPC( rpc_cur_mode );

	// Copy our capabilities from PHY_STAT_REG to PHY_AD_REG
	my_phy_caps = smc_read_phy_register(ioaddr, phyaddr, PHY_STAT_REG);
	my_ad_caps  = PHY_AD_CSMA; // I am CSMA capable

	if (my_phy_caps & PHY_STAT_CAP_T4)
		my_ad_caps |= PHY_AD_T4;

	if (my_phy_caps & PHY_STAT_CAP_TXF)
		my_ad_caps |= PHY_AD_TX_FDX;

	if (my_phy_caps & PHY_STAT_CAP_TXH)
		my_ad_caps |= PHY_AD_TX_HDX;

	if (my_phy_caps & PHY_STAT_CAP_TF)
		my_ad_caps |= PHY_AD_10_FDX;

	if (my_phy_caps & PHY_STAT_CAP_TH)
		my_ad_caps |= PHY_AD_10_HDX;

	// Disable capabilities not selected by our user
	//if (lp->ctl_rspeed != 100)
	//	my_ad_caps &= ~(PHY_AD_T4|PHY_AD_TX_FDX|PHY_AD_TX_HDX);

	//if (!lp->ctl_rfduplx)
	//	my_ad_caps &= ~(PHY_AD_TX_FDX|PHY_AD_10_FDX);

	// Update our Auto-Neg Advertisement Register
	smc_write_phy_register(ioaddr, phyaddr, PHY_AD_REG, my_ad_caps);

	// Read the register back.  Without this, it appears that when
	// auto-negotiation is restarted, sometimes it isn't ready and
	// the link does not come up.
	status = smc_read_phy_register(ioaddr, phyaddr, PHY_AD_REG);

	DBG(2, "%s: phy caps=%x\n", CARDNAME, my_phy_caps);
	DBG(2, "%s: phy advertised caps=%x\n", CARDNAME, my_ad_caps);

	// Restart auto-negotiation process in order to advertise my caps
	smc_write_phy_register( ioaddr, phyaddr, PHY_CNTL_REG,
		PHY_CNTL_ANEG_EN | PHY_CNTL_ANEG_RST );

	// Wait for the auto-negotiation to complete.  This may take from
	// 2 to 3 seconds.
	// Wait for the reset to complete, or time out
	printf("Searching for network... %c", spinner[spin++ % 4]);
	timeout = 20; // Wait up to 10 seconds
	while (timeout--) {
		status = smc_read_phy_register(ioaddr, phyaddr, PHY_STAT_REG);
		if (status & PHY_STAT_ANEG_ACK)
			// auto-negotiate complete
			break;

		printf("\b%c", spinner[spin++ % 4]);
		if (serial_poll())
			serial_read();
		msleep(50); // wait 500 millisecs

		// Restart auto-negotiation if remote fault
		if (status & PHY_STAT_REM_FLT) {
			DBG(1, "%s: PHY remote fault detected\n", CARDNAME);

			// Restart auto-negotiation
			DBG(1, "%s: PHY restarting auto-negotiation\n",
				CARDNAME);
			smc_write_phy_register( ioaddr, phyaddr, PHY_CNTL_REG,
				PHY_CNTL_ANEG_EN | PHY_CNTL_ANEG_RST |
				PHY_CNTL_SPEED | PHY_CNTL_DPLX);
		}
	}

	if (timeout < 1) {
	        printf("\bnot found.\n");
		DBG(1, "%s: PHY auto-negotiate timed out\n", CARDNAME);
		failed = 1;
	}

	// Fail if we detected an auto-negotiate remote fault
	if (status & PHY_STAT_REM_FLT) {
	        if (!failed) {
	                printf("\bnetwork found, but auto-negotiate failed.\n");
		}
		DBG(1, "%s: PHY remote fault detected\n", CARDNAME);
		failed = 1;
	}

	if (!failed) {
	        printf("\bsuccess!\n");
	}

	// Re-Configure the Receive/Phy Control register
	SMC_SET_RPC( rpc_cur_mode );

smc_phy_configure_exit:
	// Exit auto-negotiation
	return;
}

/*----------------------------------------------------------------------
 . Function: smc_probe( unsigned long ioaddr )
 .
 . Purpose:
 .	Tests to see if a given ioaddr points to an SMC91x chip.
 .	Returns a 0 on success
 .
 . Algorithm:
 .	(1) see if the high byte of BANK_SELECT is 0x33
 . 	(2) compare the ioaddr with the base register's address
 .	(3) see if I recognize the chip ID in the appropriate register
 .
 .---------------------------------------------------------------------
 */
/*---------------------------------------------------------------
 . Here I do typical initialization tasks.
 .
 . o  Initialize the structure if needed
 . o  print out my vanity message if not done so already
 . o  print out what type of hardware is detected
 . o  print out the ethernet address
 . o  find the IRQ
 . o  set up my private data
 . o  configure the dev structure with my subroutines
 . o  actually GRAB the irq.
 . o  GRAB the region
 .-----------------------------------------------------------------
*/
static int smc_probe(u32 ioaddr)
{
	unsigned int val;

	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	/* First, see if the high byte is 0x33 */
	val = SMC_CURRENT_BANK();
        val = SMC_CURRENT_BANK();
	DBG(2, "%s: val%x\n", __FUNCTION__, val);
	if ( (val & 0xFF00) != 0x3300 ) {
	        printf("Could not locate SMC91x at address 0x%08x\n", ioaddr);
		return -1;
	}

	/* The above MIGHT indicate a device, but I need to write to further
 	 	test this.  */
	SMC_SELECT_BANK(0);
	val = SMC_CURRENT_BANK();
	if ( (val & 0xFF00 ) != 0x3300 ) {
                printf("Could not locate SMC91x at address 0x%08x\n", ioaddr);
		return -1;
	}

	/* Get the MAC address */
	SMC_SELECT_BANK( 1 );
	SMC_GET_MAC_ADDR(hwaddress);
	printf("\nSMC91x detected with MAC address %02x-%02x-%02x-%02x-%02x-%02x\n", hwaddress[0], hwaddress[1],
			hwaddress[2], hwaddress[3], hwaddress[4], hwaddress[5]);

	/* now, reset the chip, and put it into a known state */
//	smc_reset( ioaddr );

	return 0;
}

static int 
smc_hw_reset(u32 ioaddr)
{
#if defined(LUBBOCK)
	{
		volatile int *attaddr = (int*)(ioaddr + 1<<25);
		long flags;

		/* first reset, then enable the device. Sequence is critical */
		attaddr[ECOR] |= ECOR_RESET;
		msleep(1);
		attaddr[ECOR] &= ~ECOR_RESET;
		attaddr[ECOR] |= ECOR_ENABLE;

		/* force 16-bit mode */
		attaddr[ECSR] &= ~ECSR_IOIS8;
		msleep(1);
	}
#endif
	return 0;
}

/**********************************************************************
 * ethernet driver interfaces
 */
static int 
smc_init(u32 ioaddr)
{
	int err=0;
	u8  version;

	DBG(2, "%s: %s\n", CARDNAME, __FUNCTION__);

	dump_banks(ioaddr);
	err = smc_hw_reset(ioaddr);
	if (err)
		return err;

	err = smc_probe(ioaddr);
	if (err) 
		return err;

	SMC_SELECT_BANK(3);
	version = SMC_GET_REV() & 0xff;
	DBG(2, "%s: version %x\n", __FUNCTION__, version);

	/* reset the hardware */
	smc_reset(ioaddr);
	
	dump_banks(ioaddr);
	smc_enable(ioaddr);

//	SMC_SELECT_BANK( 1 );
//	SMC_SET_MAC_ADDR(hwaddress);

	smc_chip_base = ioaddr;

	if (version > 0x70) 
		smc_phy_configure(ioaddr);
	
	dump_banks(ioaddr);
	return err;
}

static int readether( u8 *buf, int size )
{
	int ret = 0;
	u8  status, mask = 0xff;
	u16 saved_bank;
	u32 ioaddr = smc_chip_base;

	if ( !smc_chip_base ) 
		return -EINVAL;

	saved_bank = SMC_CURRENT_BANK();

	SMC_SELECT_BANK(2);
	mask = SMC_GET_INT_MASK();
	/* clear all interrupts */
	SMC_SET_INT_MASK(0);

	/* read the status flag, and mask it */
	status = SMC_GET_INT() & mask;

	if (!status) {
		ret = 0;
	} 
	else if ( status & IM_RCV_INT ) { 
		ret = smc_rcv( smc_chip_base, buf, size );
	}

	SMC_SET_INT_MASK(mask);
	SMC_SELECT_BANK(saved_bank);

	return ret;
}

static int writeether( u8 *buf, int size )
{
	if ( !smc_chip_base )
		return -EINVAL;

	memcpy( buf+6, hwaddress, 6 );

	return smc_send( smc_chip_base, buf, size);
}
#if 0
static int get_ethaddr( u8 dev_addr[6] )
{
	u32 ioaddr = smc_chip_base;

	if (!ioaddr) return -EINVAL;
	SMC_GET_MAC_ADDR(dev_addr);
	return 0;
}
#else
static int get_ethaddr( u8 dev_addr[6] )
{
	int 	i;
	u16	val;
	u32	ioaddr = smc_chip_base;
	
	for(i=0; i < 3; i++) {
		/* Set address to read */
		SMC_SELECT_BANK(2);
		SMC_SET_PTR( (SMC_GET_PTR() & 0xc0) | (0x20 + i));

		/* Set bit EEPROM SELECT in Control Register of Bank 1*/
		SMC_SELECT_BANK(1);
		SMC_SET_CTL( SMC_GET_CTL() |CTL_RELOAD | CTL_EEPROM_SELECT);

		while( SMC_GET_CTL() & CTL_RELOAD) {	
			DBG(3, "still reload:%x\n", SMC_GET_CTL());
			msleep(1);
		}
		
		/* Read the data */
		val = SMC_GET_GP();
		dev_addr[i*2] = *(u8*)&val;
		dev_addr[i*2 + 1] = *((u8*)&val+1);
		DBG(2,"%02x-%02x-",dev_addr[i*2], dev_addr[i*2 + 1]);
	}

	/* Clear  EEPROM SELECT */
	SMC_SELECT_BANK(1);
	DBG(2, "CTL_REG:0x%02x\n",SMC_GET_CTL());
	SMC_SET_CTL(SMC_GET_CTL() & (~(CTL_EEPROM_SELECT | CTL_RELOAD)));
	DBG(2, "CTL_REG:0x%02x\n",SMC_GET_CTL());

	return 0;
}
#endif

#if 0
static int set_ethaddr( u8 dev_addr[6] )
{
	u32 ioaddr = smc_chip_base;

	if (!ioaddr) return -EINVAL;

	SMC_SET_MAC_ADDR(dev_addr);
	return 0;
}
#else
static int set_ethaddr( u8 dev_addr[6] )
{
	int 	i;
	u16	val;
	u32	ioaddr = smc_chip_base;
	
	for(i=0; i < 3; i++) {
		/* Set address to write */
		SMC_SELECT_BANK(2);
		SMC_SET_PTR( (0x20 + i));
		
		/* Write the data */
		SMC_SELECT_BANK(1);
//		val = (dev_addr[i*2] | (dev_addr[i*2+1] << 8));
//		SMC_SET_GP(val);
		SMC_SET_GP((dev_addr[i*2] | (dev_addr[i*2+1] << 8)));
		DBG(3, "gp_reg:0x%02x|0x%02x\n", val, SMC_GET_GP());

		/* Set bit STORE in Control Register of Bank 1*/
		SMC_SELECT_BANK(1);
		SMC_SET_CTL( SMC_GET_CTL() | CTL_EEPROM_SELECT | CTL_STORE);
	
		while(SMC_GET_CTL() & CTL_STORE) {
			DBG(2, "ctl_reg:0x%02x\n",SMC_GET_CTL());
			msleep(1);
		}
	}

	/* Clear  EEPROM SELECT & reload configuration */
	SMC_SELECT_BANK(1);
	SMC_SET_CTL(SMC_GET_CTL() & (~(CTL_EEPROM_SELECT|CTL_STORE)) );
	
	DBG(3, "MAC address is modified, %x\n", SMC_GET_CTL());
#if 1
	/* Send reload command to reload configuration */
	SMC_SET_CTL(SMC_GET_CTL() | CTL_RELOAD);
	while(SMC_GET_CTL() & CTL_RELOAD) {
		DBG(2, "again ctl_reg:0x%02x\n",SMC_GET_CTL());
		msleep(1);
	}

#endif
	SMC_GET_MAC_ADDR(hwaddress);
	DBG(0, "%02x-%02x-%02x-%02x-%02x-%02x\n", hwaddress[0], hwaddress[1],
			hwaddress[2], hwaddress[3], hwaddress[4], hwaddress[5]);


	return 0;
}
#endif
/**********************************************************************
 * export ethernet driver 
 */
ether_driver_t smc91x_ether_driver = {
	init:		smc_init,
	rcv:		readether,
	snd:		writeether,
	set_addr:	set_ethaddr,
	get_addr:	get_ethaddr,
	base:		SMC_BASE,
};
