/**********************************************************************
 * generic_io.c
 *
 * generic io abstraction
 *
 * Copyright (C) 2002, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: generic_io.c,v 1.9 2003/08/21 12:47:10 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>
#include <blob/time.h>
#include <blob/serial.h>
#include <blob/arch.h>
#include <blob/init.h>

#include <blob/generic_io.h>

/**********************************************************************
 * Defines / Makros
 */

#define IO_DEBUG 1
#undef IO_QUIET

#if IO_DEBUG
#	define DBG( x, args... )	if ( io_dbg>x ) printf( args )
#	define DBGIO( x, io )		if ( io_dbg>x ) printf( "%s: " \
			"%s(%p)={ c=%p, r=%p, w=%p,\n\tp=%p, sz=%d, ch=%p,\n\tn='%s' }\n", \
			__FUNCTION__, #io, io, io->conf, io->read, io->write, \
			io->private_data, io->io_size, io->child_io, io->name );
#else
#	define DBG( x, args... )
#	define DBGIO( x, io )
#endif

#define ERR( x )	do { ret = x; goto DONE; } while (0)


/**********************************************************************
 * Typen
 */

/**********************************************************************
 * Programmglobale Variable
 */

/**********************************************************************
 * Modulglobale Variable
 */
static char *module_version = "$Id: generic_io.c,v 1.9 2003/08/21 12:47:10 seletz Exp $";

#ifdef	IO_DEBUG
static int io_dbg = 1;
#else
static int io_dbg = 0;
#endif

static io_driver_t *drivers[IO_MAX_DRIVER];


/**********************************************************************
 * Prototypen
 */

/**********************************************************************
 * Exported functions
 */

void io_dbg_set( int lvl )
{
	io_dbg = lvl;
}

void io_init( void )
{
	int i;

	for ( i=0; i<IO_MAX_DRIVER;i++)
		drivers[i] = NULL;

	return;
}
__initlist(io_init, INIT_LEVEL_OTHER_STUFF + 1);

int io_register( io_driver_t *io, char *name )
{
	int i;

	if ( !io || !name )
		return -EINVAL;

	for ( i=0; i<IO_MAX_DRIVER;i++) {
		if (!drivers[i])
			break;
	}
	if ( i==IO_MAX_DRIVER)
		return -EINVAL;

	drivers[i]=io;
	strncpy( io->name, name, IO_NAME_LEN );

	DBG( 1, "%s: driver '%s' registered #%d.\n", __FUNCTION__,
			name, i );

	return 0;
}

int io_configure( char *io_name, void *conf_data )
{
	io_driver_t *io;

	DBG( 1, "%s: io_name=%p, conf_data=%p\n", __FUNCTION__,
			io_name, conf_data );

	if ( !io_name )
		return -EINVAL;

	io = io_get_byname( io_name );
	if ( !io )
		return -EINVAL;

	DBGIO( 5, io );

	if (io->conf )
		return io->conf( io, conf_data );

	return 0;
}

io_driver_t *io_get_byname( char *name )
{
	int i;

	if ( !name )
		return NULL;

	for ( i=0; i<IO_MAX_DRIVER;i++) {
		if (drivers[i]) {
			if ( strncmp( io_get_name( drivers[i] ), name, IO_NAME_LEN ) == 0 )
				return drivers[i];
		}
	}

	return NULL;
}

int io_chain_driver( char *parent, char *new_child )
{
	io_driver_t *p, *c;

	if ( !parent || !new_child )
		return -EINVAL;

	p = io_get_byname( parent );
	if ( !p )
		return -EINVAL;

	DBGIO( 5, p );

	c = io_get_byname( new_child );
	if ( !c )
		return -EINVAL;

	DBGIO( 5, c );

	DBG( 1, "%s: chained Parent '%s' -> child '%s'.\n", __FUNCTION__,
			io_get_name( p ), io_get_name( c ) );

	p->child_io = c;

	return 0;
}

int io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *rd )
{
	DBG( 1, "%s: dest=%p, src=%p, amount=%d, io_name=%p\n", __FUNCTION__,
			dest, src, amount, rd );

	if ( !rd || !rd->read )
		return -EINVAL;

	DBGIO( 5, rd );

	/* check bounds */
	if ( (rd->io_size - (size_t)src) < amount ) {
		DBG( 1, "%s: out of bounds!\n", __FUNCTION__ );
		return -ERANGE;
	}

	return rd->read( dest, src, amount, rd );
}

int io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *rd )
{

	DBG( 1, "%s: dest=%p, src=%p, amount=%d, io_name=%p\n", __FUNCTION__,
			dest, src, amount, rd );

	if ( !rd || !rd->write  )
		return -EINVAL;

	DBGIO( 5, rd );

	/* check bounds */
	if ( (rd->io_size - (size_t)dest) < amount ) {
		DBG( 1, "%s: out of bounds!\n", __FUNCTION__ );
		return -ERANGE;
	}

	return rd->write( dest, src, amount, rd );
}

int io_copy( char *dest_name, char *src_name, size_t amount )
{
	static unsigned char buffer[512];
	int ret = 0;
	int blocks, block, to_copy;
	unsigned char *adr = NULL;
	io_driver_t *dest, *src;

	DBG( 1, "%s: dest_name=%p, src_name=%p, amount=%d\n", __FUNCTION__,
			dest_name, src_name, amount  );

	if ( !src_name || !dest_name )
		ERR( -EINVAL );

	src = io_get_byname( src_name );
	if ( !src )
		ERR( -EINVAL );

	dest = io_get_byname( dest_name );
	if ( !dest )
		ERR( -EINVAL );

	DBGIO( 5, src );
	DBGIO( 5, dest );

	if ( amount == 0 ) {
		amount = io_get_size( src );
	}

	if ( amount > io_get_size( dest ) ) {
		printf( "\n\nio_copy: src size (%d) > dest space (%d)!\n\n",
				amount, io_get_size( dest ) );
		ERR( -ETOOLONG );
	}

	blocks = amount / 512 + (amount%512?1:0);
	DBG( 1, "%s: %d blocks (%d bytes) to copy.\n", __FUNCTION__,
			blocks, amount );

	for ( block=0; block<blocks; block++ ) {
		if ( amount > 512 ) {
			to_copy = 512;
		} else {
			to_copy = amount;
		}
#if !defined(IO_QUIET)
		printf( "%s: block %d/%d (%d bytes to go)      \r", __FUNCTION__, block+1, blocks, amount );
#endif

		/* read from source */
		ret = io_read( buffer, adr, to_copy, src );
		if ( ret ) {
			DBG( 1, "\n%s: read failed at block %d (adr %p).\n",
					__FUNCTION__, block, adr );
			ERR( -EINVAL );
		}

		/* write to dest */
		ret = io_write( adr, buffer, to_copy, dest );
		if ( ret ) {
			DBG( 1, "\n%s: write failed at block %d (adr %p).\n",
					__FUNCTION__, block, adr );
			ERR( -EINVAL );
		}

		amount -= to_copy;
		adr += to_copy;
	}

	ret = 0;
DONE:
#if !defined(IO_QUIET)
	printf("\n");
#endif
	return ret;
}
