/**********************************************************************
 * GIO flash "driver"
 *
 * Copyright (C) 2003, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: gio_flash.c,v 1.1 2003/04/03 14:32:50 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>

#include <blob/init.h>
#include <blob/flash.h>
#include <blob/generic_io.h>

/**********************************************************************
 * defines
 */
#define DEBUG 0

#if DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

/**********************************************************************
 * module globals
 */
#if DEBUG
static int dbg = 1;
#else
static int dbg = 0;
#endif

static char module_version[] = "$Id: gio_flash.c,v 1.1 2003/04/03 14:32:50 seletz Exp $";

/**********************************************************************
 * static functions
 */

/**********************************************************************
 * flash io driver
 */
int flash_io_init( io_driver_t *io );
static int flash_io_conf( io_driver_t * io, void *conf );
static int flash_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );
static int flash_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );

/* initialize driver struct */
int flash_io_init( io_driver_t *io )
{
	if ( !io )
		return -EINVAL;

	io->private_data = NULL;
	io->io_size = 0;
	io->conf = flash_io_conf;
	io->read = flash_io_read;
	io->write = flash_io_write;
	io->child_io = NULL;

	return 0;
}

/* configure: select a base offset within the flash */
static int flash_io_conf( io_driver_t * io, void *conf )
{
	u32 offset;

	if ( !io )
		return -EINVAL;

	offset = (u32)conf;

	io->io_size = flash_get_size();

	if ( offset > io->io_size )
		return -ERANGE;
	
	io->private_data=(void *)offset;

	return 0;
}

static int flash_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	u32 offset;

	DBG( 5, "%s: dest=%p, src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, amount, io );

	if ( !io || !dest )
		return -EINVAL;


	offset = (u32)io->private_data;

	DBG( 5, "%s: using offset %x.\n", __FUNCTION__, offset );

	memcpy( dest, (unsigned char *)(offset + (u32)src), amount );

	return 0;
}

/* flash write: only on word boundaries! */
static int flash_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	u32 offset;
	u32 adr;

	DBG( 5, "%s: dest=%p, src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, amount, io );

	if ( !io || !src )
		return -EINVAL;


	offset = (u32)io->private_data;
	DBG( 5, "%s: using offset %x.\n", __FUNCTION__, offset );

	adr = (u32)dest + offset;

	/* check alignment */
	if ( adr%4 ) {
		printf( "%s: misaligned destination address (%x).\n",
				__FUNCTION__, adr );
		return -EALIGN;
	}
	if ( amount%4 ) {
		printf( "%s: warning: misaligned amount to write (%d).\n",
				__FUNCTION__, amount );
	}

	DBG( 5, "%s: adr=0x%08x, nwords=%d.\n", __FUNCTION__,
			adr, amount/4 + (amount%4?1:0));

	return flash_write_region( (u32 *)adr, (u32*)src, amount/4 + (amount%4?1:0));
}

/* register flash io driver */
void init_flash_io( void )
{
	int ret;
	static io_driver_t io_flash;
	
	ret = flash_io_init( &io_flash );
	if ( ret ) return ;

	ret = io_register( &io_flash, "FLASH" ); 
	if ( ret ) return;

	ret = io_configure( "FLASH", 0 );
	if ( ret ) return;

	return;
}
