/**********************************************************************
 * GIO ram "driver"
 *
 * Copyright (C) 2003, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: gio_ram.c,v 1.1 2003/04/03 14:32:50 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>

#include <blob/init.h>
#include <blob/generic_io.h>

/**********************************************************************
 * defines
 */
#define DEBUG 0

#if DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

#if !defined( RAM_START ) || !defined( RAM_SIZE )
#	error "define RAM_START and RAM_SIZE for your platform"
#endif

/**********************************************************************
 * module globals
 */
#if DEBUG
static int dbg = 1;
#else
static int dbg = 0;
#endif

static char module_version[] = "$Id: gio_ram.c,v 1.1 2003/04/03 14:32:50 seletz Exp $";

/**********************************************************************
 * static functions
 */
static int ram_io_conf( io_driver_t * io, void *conf );
static int ram_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );
static int ram_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );

/**********************************************************************
 *  A simple RAM driver
 *  - io driver for RAM access
 */
int ram_io_init( io_driver_t *io, ram_io_t *ram )
{
	if ( !io || !ram )
		return -EINVAL;

	io->private_data = (void *)ram;
	io->io_size = ram->len - ram->pos;
	io->conf = ram_io_conf;
	io->read = ram_io_read;
	io->write = ram_io_write;
	io->child_io = NULL;

	return 0;
}

static int ram_io_conf( io_driver_t * io, void *conf )
{
	ram_io_t *ram;

	if ( !io || !conf )
		return -EINVAL;

	ram = (ram_io_t*)io->private_data;
	if ( !ram )
		return -EINVAL;

	if ( (u32)conf >= ram->len )
		return -EINVAL;

	ram->pos = (u32)conf;

	io->io_size = ram->len - (u32)ram->pos;

	return 0;
}

static int ram_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	u32 offset;
	ram_io_t *ram;

	DBG( 1, "%s: dest=%p, src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, amount, io );

	if ( !io || !dest )
		return -EINVAL;

	ram = (ram_io_t*)io->private_data;
	if ( !ram )
		return -EINVAL;

	offset = ram->start + (u32)ram->pos;
	
	DBG( 1, "%s: dest=%p, src=%p(=%p), amount=%d, io=%p\n", __FUNCTION__, 
			dest, src, (void*)(src +offset), amount, io );

	memcpy( dest, (void *)(offset + (u32)src), amount);

	return 0;
}

static int ram_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	u32 offset;
	ram_io_t *ram;


	if ( !io || !src )
		return -EINVAL;

	ram = (ram_io_t*)io->private_data;
	if ( !ram )
		return -EINVAL;

	offset = ram->start + (u32)ram->pos;

	DBG( 1, "%s: dest=%p(=%p), src=%p, amount=%d, io=%p\n", __FUNCTION__, 
			dest, (void*)(dest +offset), src, amount, io );

	memcpy((void *)(offset + (u32)dest), src, amount);
	
	return 0;
}

/**********************************************************************
 *  init ram pools
 */
void init_ram_io( void )
{
	int ret;
	static ram_io_t		ram0;
	static io_driver_t	io_ram0;

	ram0.start=RAM_START;
	ram0.len=RAM_SIZE;
	ram0.pos=0;

	ret = ram_io_init( &io_ram0, &ram0 );
	if ( ret ) return;

	ret = io_register( &io_ram0, "RAM" ); 
	if ( ret ) return;
}
