/**********************************************************************
 * ide.c
 *
 * Basic ide drive framework. Based on Brad Parkers work.
 *
 * Copyright (C) 2001, Brad Parker (brad@heeltoe.com)
 * Copyright (C) 2002, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: ide.c,v 1.7 2002/05/10 15:33:39 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>
#include <blob/time.h>
#include <blob/serial.h>
#include <blob/arch.h>
#include <blob/ide.h>

/**********************************************************************
 * Defines / Makros
 */
#define WEAK_SYM __attribute__ (( weak ))

#define IDE_DEBUG 1

#ifdef IDE_DEBUG
#	define DBG( x, args... )	if ( ide_dbg>x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

#define IDE_DELAY 10

/**********************************************************************
 * Typen
 */

/**********************************************************************
 * Programmglobale Variable
 */

/**********************************************************************
 * Modulglobale Variable
 */
static char *version = "$Id: ide.c,v 1.7 2002/05/10 15:33:39 seletz Exp $";

#ifdef	IDE_DEBUG
static int ide_dbg = 1;
#else
static int ide_dbg = 0;
#endif


/**********************************************************************
 * Prototypen
 */
static int hd_identify(volatile u8 *port, volatile u8 *reg, char *buffer);
static int hd_recal(volatile u8 *port);
static int hd_busy_wait(volatile u8 *port);
static void ide_fixstring (u8 *s, const int bytecount, const int byteswap);

/**********************************************************************
 * Exported functions
 */

void ide_dbg_set( int level )
{
	ide_dbg = level;
}

/**********************************************************************
 * ide_init
 *
 * - Initialize ide drive struct
 * - calculatees and sets HD_PORT and HD_REG_RORT from a base address
 */
int ide_init( ide_drive_t *drive, u32 base_addr )
{
	if ( !base_addr )
		return -1;

	if ( !drive )
		return -1;

	DBG( 1, "ide: %s\n", version );

	drive->ide_port = (u8 *)(base_addr + HD_PORT);
	drive->ide_reg = (u8 *)(base_addr + HD_REG_PORT);
	drive->last_status = 0;

	DBG( 1, "%s: ide_port=%p\n", __FUNCTION__, drive->ide_port );
	DBG( 1, "%s: ide_reg=%p\n", __FUNCTION__, drive->ide_reg );

	return 0;
}

/**********************************************************************
 * ide_status_dump
 *
 * - Dump IDE ports
 */
int ide_status_dump( ide_drive_t *drive )
{
	u8 status;

	if ( !drive || !drive->ide_port )
		return -1;


	status = drive->ide_port[7];
	printf( "ide: status[%s %s %s %s %s %s %s %s]\n",
			status&BUSY			?"BSY":"   ",
			status&READY		?"RDY":"   ",
			status&WRITE_FAULT	?"DF ":"   ",
			status&SEEKOK		?"DSC":"   ",
			status&DATA_REQ		?"DRQ":"   ",
			status&CORRECTED	?"COR":"   ",
			status&REV_INDEX	?"IDX":"   ",
			status&ERROR		?"ERR":"   " 
		  );

	return 0;
}

/**********************************************************************
 * ide_reset
 *
 * - reset ide controller
 *
 * FIXME: resolve those magic numbers
 */
int ide_reset( ide_drive_t *drive )
{
	volatile u8 *port = NULL;
	volatile u8 *reg = NULL;

	if ( !drive || !drive->ide_port )
		return -1;

	port = drive->ide_port;
	reg = drive->ide_reg;

#if 1
	/* reset hd controller */
	reg[0] = 0x06;
	barrier();
	msleep(IDE_DELAY);
	reg[0] = 0x02;
	barrier();
	msleep(IDE_DELAY);
#else
	msleep(5*IDE_DELAY);
#endif

	port[1] = 0;
	port[2] = 0;
	port[3] = 0;
	port[4] = 0;
	port[5] = 0;
	port[6] = 0xa0;
	barrier();
	msleep(IDE_DELAY);

	return 0;
}

/**********************************************************************
 * ide_identify_drive
 *
 * - Try to identify hd
 * - Stores hd_driveid struct in ide drive struct
 * - gets sectors per track and heads of disk
 */
int ide_identify_drive( ide_drive_t *drive )
{
	struct hd_driveid *id;
	volatile u8 *port = NULL;
	volatile u8 *reg = NULL;
	static char id_buffer[IDE_BLOCK_SIZE];

	if ( !drive || !drive->ide_port )
		return -1;

	port = drive->ide_port;
	reg = drive->ide_reg;

	if (hd_identify(port, reg, id_buffer)) {
		return -1;
	}

	id = (struct hd_driveid *)id_buffer;

	ide_fixstring (id->model, sizeof(id->model), 1);

	printf("ide: model: '%s'\n", id->model);
	printf("ide: CHS:   %d/%d/%d\n", id->cyls, id->heads, id->sectors);
	printf("ide: lba_capacity %d (total %d bytes)\n", id->lba_capacity, id->lba_capacity*IDE_BLOCK_SIZE);

	if (id->sectors && id->heads) {
		drive->disk_sectors_per_track = id->sectors;
		drive->disk_heads = id->heads;
	} else {
		drive->disk_sectors_per_track = 0;
		drive->disk_heads = 0;
	}

	drive->driveid = *id;

	return 0;
}

/**********************************************************************
 * ide_setfeature
 *
 * - set features
 *
 */
int ide_setfeature( ide_drive_t *drive, unsigned char opcode, unsigned char param )
{
	volatile u8 *port = NULL;
	volatile u8 *reg = NULL;

	if ( !drive || !drive->ide_port )
		return -1;

	port = drive->ide_port;
	reg = drive->ide_reg;

	DBG( 1, "%s: HDC_SETFEATURE( 0x%02x, 0x%02x )\n", __FUNCTION__,
			opcode, param );

	port[1] = opcode;
	port[2] = param;
	port[3] = 0;
	port[4] = 0;
	port[5] = 0;
	port[6] = 0;
	port[7] = HDC_SETFEATURE;

	hd_busy_wait(port);

	return 0;
}

/**********************************************************************
 * hd_read
 *
 * - read head/sector/cyl of drive to a buffer. Reads 1 block (512 bytes) of
 * data.
 */
int hd_read( ide_drive_t *drive, int head, int sector, int cyl, char *buffer)
{
	int i, hd_drive, hd_cyl, hd_head, hd_sector;
	u8 status, hd_cmd[8];
	volatile u8 *port = NULL;
	volatile u8 *reg = NULL;

	if ( !drive || !drive->ide_port )
		return -1;

	port = drive->ide_port;
	reg = drive->ide_reg;

	hd_drive = 0;
	hd_cyl = cyl;
	hd_head = head;
	hd_sector = sector;

	hd_cmd[1] = 0;
	hd_cmd[2] = 1;			/* # sectors */
	hd_cmd[3] = hd_sector;		/* starting sector */
	hd_cmd[4] = cyl & 0xff;		/* cylinder low byte */
	hd_cmd[5] = (cyl >> 8) & 0xff;	/* cylinder hi byte */
	hd_cmd[6] = (hd_drive << 4) | (hd_head & 0x0f) | 0xa0;

	DBG( 2, "%s: h=%d s=%d c=%d\n", __FUNCTION__, hd_head, hd_sector,
			hd_cyl);

#if 0
	if (hd_head > 7) {
		//	hd_control |= 0x08;
		//	reg[0] = hd_control;
		//	hd_control &= 0xf7;
		reg[0] = 0x08;
	}
#endif

	/* --- */
	hd_busy_wait(port);

	for (i = 1; i < 7; i++) {
		port[i] = hd_cmd[i];
		barrier();
	}

	port[7] = HDC_READ;
	barrier();

	hd_busy_wait(port);

	for (i = 0; i < IDE_BLOCK_SIZE; i++) {
		buffer[i] = port[0];
		barrier();
	}

	drive->last_status = status = port[7];

	if (status & ERROR) {
		printf("read status: %2x\n", status);

		printf("h=%d s=%d c=%d\n", hd_head, hd_sector, hd_cyl);
		for (i = 1; i < 7; i++) {
			printf("set reg[%d] %2x\n", i, hd_cmd[i]);
		}
	}

	if (status & ERROR)
		return -1;

	return 0;
}

/**********************************************************************
 * hd_read_lba
 * - read using LBA mode
 */
int hd_read_lba( ide_drive_t *drive, u32 lba, char *buffer)
{
	int i, hd_drive;
	u8 status, hd_cmd[8];
	volatile u8 *port = NULL;
	volatile u8 *reg = NULL;

	if ( !drive || !drive->ide_port )
		return -1;

	port = drive->ide_port;
	reg = drive->ide_reg;

	lba &= 0x0fffffff;

	hd_drive = 0;

	hd_cmd[1] = 0;
	hd_cmd[2] = 1;			/* # sectors */
	hd_cmd[3] = lba&0xff;
	hd_cmd[4] = (lba>>8)&0xff;
	hd_cmd[5] = (lba>>16)&0xff;
	hd_cmd[6] = (lba>>24)&0xff;
	hd_cmd[6] |= ((hd_drive&1) << 4) | 0xe0;

	hd_busy_wait(port);

	for (i = 1; i < 7; i++) {
		port[i] = hd_cmd[i];
		barrier();
	}

	port[7] = HDC_READ;
	barrier();

	hd_busy_wait(port);

	for (i = 0; i < IDE_BLOCK_SIZE; i++) {
		buffer[i] = port[0];
		barrier();
	}

	drive->last_status = status = port[7];

	if (status & ERROR) {
		printf("hd_read_lba: read status: %2x\n", status);

		printf("hd_read_lba: lba=%08x\n", lba&0x0fffffff );
		for (i = 1; i < 7; i++) {
			printf("hd_cmd[%d] %2x\n", i, hd_cmd[i]);
		}
	}

	if (status & ERROR)
		return -1;

	return 0;
}

/**********************************************************************
 * hd_read_mapped
 *
 * - reads 512 bytes from sector <sector_num> from drive <drive>
 * - <sector_num> gets mapped to CHS, then hd_read() is called.
 */
int hd_read_mapped( ide_drive_t *drive, int sector_num, char *buffer)
{
	int head, sector, cyl;

	if ( !drive || !drive->ide_port )
		return -1;

	DBG( 2, "%s: sector_num=%d, buffer=%p\n", __FUNCTION__,
				sector_num, buffer);

#if 0
	sector = (sector_num % disk_sectors_per_track) + 1;
	sector_num /= disk_sectors_per_track;

	head = sector_num % disk_heads;
	sector_num /= disk_heads;

	cyl = sector_num;
#endif

	hd_map_sector( drive, sector_num, &head, &sector, &cyl );

	return hd_read( drive, head, sector, cyl, buffer);
}

/**********************************************************************
 * hd_map_sector
 *
 * - map continguous sector numbers to CHS values of <drive>
 */
int hd_map_sector( ide_drive_t *drive, int sector, int *head, int *sec, int *cyl )
{
	int _cyl, _head, _sec;

	if ( !drive )
		return -1;

	_sec = (sector % drive->disk_sectors_per_track) + 1;
	sector /= drive->disk_sectors_per_track;

	_head = sector % drive->disk_heads;
	sector /= drive->disk_heads;

	_cyl = sector;

	DBG( 1, "%s: => (h:%d, s:%d, c:%d)\n", __FUNCTION__, _head, _sec, _cyl );

	if ( head ) *head = _head;
	if ( sec ) *sec = _sec;
	if ( cyl ) *cyl = _cyl;

	return 0;
}
/**********************************************************************
 * Statische Funktionen
 */

/**********************************************************************
 * hd_identify
 *
 * - performs HDC_IDENTIFY command and reads result to buffer
 */
static int hd_identify(volatile u8 *port, volatile u8 *reg, char *buffer)
{
	int i;
	unsigned char status;

	hd_busy_wait(port);

	port[1] = 0;
	port[2] = 1;
	port[3] = 0;
	port[4] = 0;
	port[5] = 0;
	port[6] = 0xa0;
	port[7] = HDC_IDENTIFY;
	barrier();

	hd_busy_wait(port);

	for (i = 0; i < IDE_BLOCK_SIZE; i++) {
		barrier();
		buffer[i] = port[0];
	}

	status = port[7];
	DBG( 1, "%s: status=0x%02x\n", __FUNCTION__, status );

	if (status & ERROR) {
		printf("ide: identify status: %2x\n", status);
		return -1;
	}

	return 0;
}

static int hd_recal(volatile u8 *port)
{
	int i, hd_drive, hd_head, sector;
	u8 hd_cmd[8];

	hd_drive = 0;
	sector = 0;
	hd_head = sector & 0xffff;

	hd_cmd[1] = 0;
	hd_cmd[2] = 1;			/* # sectors */
	hd_cmd[3] = 0;
	hd_cmd[4] = 0;
	hd_cmd[5] = 0;
	hd_cmd[6] = (hd_drive << 4) | (hd_head & 0x0f) | 0xa0;

	hd_busy_wait(port);

	for (i = 1; i < 7; i++) {
		hd_busy_wait(port);
		port[i] = hd_cmd[i];
		barrier();
	}

	port[7] = HDC_RECAL;
	barrier();

	if (hd_busy_wait(port))
		return -1;

	return 0;
}

static int hd_busy_wait(volatile u8 *port)
{
	int count;
	u8 status, err;

	count = 0;
	while (count++ < 2500) {
		status = port[7];
		barrier();
		if ((status & BUSY) == 0)
			break;
		msleep(1);
	}

	DBG( 1, "%s: port=%p, count=%d, status=%02x\n", __FUNCTION__,
			port, count, status );

	if ( (status & ERROR) || (status & WRITE_FAULT) ) {
		err = port[1];
		barrier();
		printf( "ide: busy_wait: err = %2x\n", err );
		return -1;
	}

	return 0;
}

static void ide_fixstring (u8 *s, const int bytecount, const int byteswap)
{
	u8 *p = s, *end = &s[bytecount & ~1]; /* bytecount must be even */

	if (byteswap) {
		/* convert from big-endian to host byte order */
		for (p = end ; p != s;) {
			unsigned short *pp = (unsigned short *) (p -= 2);
			//			*pp = ntohs(*pp);
			*pp = ((*pp >> 8) & 0xff) | ((*pp & 0xff) << 8);
		}
	}

	/* strip leading blanks */
	while (s != end && *s == ' ')
		++s;

	/* compress internal blanks and strip trailing blanks */
	while (s != end && *s) {
		if (*s++ != ' ' || (s != end && *s && *s != ' '))
			*p++ = *(s-1);
	}

	/* wipe out trailing garbage */
	while (p != end)
		*p++ = '\0';
}
