/**********************************************************************
 * tar archive framework
 *
 * Copyright (C) 2002, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: tar.c,v 1.5 2002/07/25 14:08:00 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>

#include <blob/init.h>
#include <blob/generic_io.h>
#include <blob/tar.h>

/**********************************************************************
 * defines
 */
#define TAR_DEBUG 0

#if TAR_DEBUG
#	define DBG( x, args... )	if ( tar_dbg>x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

/**********************************************************************
 * module globals
 */
#if TAR_DEBUG
static int tar_dbg = 1;
#else
static int tar_dbg = 0;
#endif

static char module_version[] = "$Id: tar.c,v 1.5 2002/07/25 14:08:00 seletz Exp $";

/**********************************************************************
 * static functions
 */
static int tar_skip_file( tar_arch_t *arch );
static int tar_file_search( tar_arch_t *arch, char *file );
static int tar_xtract_blocks( tar_arch_t *arch, char *dest );
static int tar_fileinfo_get( tar_arch_t *arch );
static int tar_hdr_chk( tar_hdr_posix_t *hdr );
static int tar_hdr_fname_get( tar_hdr_posix_t *hdr, char **fname );
static int tar_hdr_fsize_get( tar_hdr_posix_t *hdr, unsigned long *size );
static int tar_from_oct( unsigned char *start, size_t digits, unsigned long *value );

/* io driver stuff */
static int tar_io_conf( io_driver_t * io, void *conf );
static int tar_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );
static int tar_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io );

/**********************************************************************/
/**********************************************************************/
/**********************************************************************/

/* set debug level */
void tar_dbg_set( int lvl )
{
	tar_dbg = lvl;
}

/**********************************************************************
 * io driver for tar archives
 */
static void init_tar_default_io( void )
{
	int ret;
	static io_driver_t	def_tar_io_driver;
	static tar_arch_t	def_tar_arch;

	tar_dbg_set(0);

	/* init default tar drv */
	ret = tar_io_init( &def_tar_io_driver, &def_tar_arch );
	if ( ret ) return;

	ret = io_register( &def_tar_io_driver, "TAR" );
	if ( ret ) return;
}
__initlist(init_tar_default_io, INIT_LEVEL_OTHER_STUFF + 2);

int tar_io_init( io_driver_t *io, tar_arch_t *arch )
{
	if ( !io || !arch )
		return -EINVAL;

	DBG( 5, "%s: io=%p, arch=%p\n", __FUNCTION__,
	  io, arch );

	io->conf = tar_io_conf;
	io->read = tar_io_read;
	io->write = tar_io_write;
	io->private_data = (void *)arch;

	/* filled by configure */
	io->io_size = 0;
	
	/* filled for us by io_chain() */
	io->child_io = NULL;

	return 0;
}

/* the user gives us a filename in conf */
static int tar_io_conf( io_driver_t * io, void *conf )
{
	int ret;
	char *name;
	io_driver_t *child;
	tar_arch_t *arch;

	DBG( 5, "%s: io=%p, conf=%p\n", __FUNCTION__,
	  io, conf );

	if ( !io || !conf )
		return -EINVAL;

	arch = (tar_arch_t*)io->private_data;
	if ( !arch )
		return -EINVAL;

	child = io_get_child( io );
	if ( !child )
		return -EINVAL;

	name = (char *)conf;
	DBG( 5, "%s: name=%s\n", __FUNCTION__, name );

	/* init new archive struct */
	ret = tar_init( arch, child );
	if ( ret )
		return -EINVAL;

	/* try to find file */
	ret = tar_file_search( arch, name );
	if ( ret )
		return -EINVAL;

	/* set io_size to filesize */
	io->io_size = arch->f_size;

	return 0;
}

static int tar_io_read( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	io_driver_t *child;
	tar_arch_t *arch;
	u32 offset;

	arch = (tar_arch_t*)io->private_data;
	if ( !arch )
		return -EINVAL;

	child = io_get_child( io );
	if ( !child )
		return -EINVAL;

	offset = arch->f_start_block*TAR_BLOCKSIZE;

	DBG( 3, "%s: dest=%p, src=%p(=%p), amount=%d, child=%p\n", __FUNCTION__,
			dest, src, (char *)((u32)src + offset), amount, child );

	return io_read(dest, (unsigned char *)(offset + (u32)src), amount, child );
}

static int tar_io_write( unsigned char *dest, unsigned char *src, size_t amount, io_driver_t *io )
{
	return -EINVAL;
}

/**********************************************************************
 * TAR functions
 */
/* init archive struct */
int tar_init( tar_arch_t *arch, io_driver_t *io )
{
	DBG( 1, "%s: arch=%p, io=%p\n", __FUNCTION__, arch, io );

	DBG( 1, "tar: version %s\n", module_version );

	if ( !arch ) {
		return -EINVAL;
	}

	arch->io_driver=io;
	arch->block = 0;
	arch->f_size=0l;
	arch->f_blocks=0l;
	arch->filename[0]=0;

	return 0;
}

/* do a "tar tv" */
int tar_tell( tar_arch_t *arch )
{
	int ret = 0;
	int fini = 0;
	int files = 0;

	DBG( 1, "%s: arch=%p\n", __FUNCTION__, arch );

	if ( !arch )
		return -EINVAL;

	/* reset to block 0 */
	arch->block = 0;

	while ( !fini ) {
		ret = tar_fileinfo_get( arch );
		if ( ret ) {
			fini = 1;
			continue;
		}

		files++;

		printf( "%s %ld bytes, %d blocks\n",
				arch->filename,
				arch->f_size,
				arch->f_blocks );

		/* skip file */
		ret = tar_skip_file( arch );
		if ( ret ) {
			printf( "tar: tar_skip_file: %d\n", ret );
			return -EINVAL;
		}
	}

	printf( "tar: %d files.\n", files );

	return 0;
}

/* do a "tar xv" */
int tar_xtract( tar_arch_t *arch, char *filename, char *dest )
{
	int ret = 0;

	DBG( 1, "%s: arch=%p, filename=%p, dest=%p\n", __FUNCTION__, arch, filename,
	  dest );

	if ( !arch || !filename )
		return -EINVAL;

	/* reset to block 0 */
	arch->block = 0;

	ret = tar_file_search( arch, filename );
	if ( ret ) {
		printf( "tar: cant find file '%s' in archive (%d).\n",
			   filename, ret );	
		return -EINVAL;
	}

	ret = tar_xtract_blocks( arch, dest );
	if ( ret ) {
		printf( "tar: cant extract file '%s' from archive (%d).\n",
			   filename, ret );	
		return -EINVAL;
	}

	printf( "tar: extracted file '%s', %ld bytes (%d blocks).\n",
			arch->filename, arch->f_size, arch->f_blocks );

	return 0;
}

/**********************************************************************
 * Static functions
 */

static int tar_file_search( tar_arch_t *arch, char *file )
{
	int ret = 0;
	int fini = 0;
	int found = 0;

	DBG( 1, "%s: arch=%p, file=%p\n", __FUNCTION__, arch, file );

	if ( !arch || !file )
		return -EINVAL;

	while ( !fini ) {
		ret = tar_fileinfo_get( arch );
		if ( ret ) {
			fini = 1;
			continue;
		}

		DBG( 5, "%s %ld bytes, %d blocks\n",
				arch->filename,
				arch->f_size,
				arch->f_blocks );

		if ( strncmp( arch->filename, file, 255 ) == 0 ) {
			found = 1;
			fini = 1;
			continue;
		}

		/* skip file */
		ret = tar_skip_file( arch );
		if ( ret ) {
			printf( "tar: tar_skip_file: %d\n", ret );
			return -EINVAL;
		}
	}

	return found?0:1;
}
/* skip to next file, if any */
static int tar_skip_file( tar_arch_t *arch )
{
	DBG( 1, "%s: arch=%p\n", __FUNCTION__, arch );

	if ( !arch )
		return -EINVAL;

	/* skip data */
	arch->block += arch->f_blocks;
	return 0;
}

/* extract current file to memory. */
static int tar_xtract_blocks( tar_arch_t *arch, char *dest )
{
	int ret = 0;
	unsigned long size;

	DBG( 1, "%s: arch=%p, dest=%p\n", __FUNCTION__, arch, dest );

	if ( !arch || !dest )
		return -EINVAL;

	size = arch->f_size;
	if ( !size )
		return -EINVAL;

	ret = io_read( dest, (unsigned char*)(arch->f_start_block*TAR_BLOCKSIZE), arch->f_size, arch->io_driver );
	if ( ret )
		return ret;

	return 0;
}

/* get current file infos out of tar header */
static int tar_fileinfo_get( tar_arch_t *arch )
{
	int ret = 0;
	tar_hdr_posix_t *phdr = NULL;

	if ( !arch )
		return -EINVAL;

	/* get header */
	ret = io_read( arch->buffer, (unsigned char *)(arch->block*TAR_BLOCKSIZE), TAR_BLOCKSIZE, arch->io_driver );
	if ( ret ) {
		printf( "tar: cant get block: %d\n", ret );
		goto DONE;
	}
	arch->block++;

	/* and decode infos */
	phdr = (tar_hdr_posix_t*)arch->buffer;
	ret = tar_hdr_chk( phdr );
	if ( ret ) {
		printf( "tar: no posix tar header: %d\n", ret );
		ret = -EINVAL;
		goto DONE;
	}

	ret = tar_hdr_fsize_get( phdr, &arch->f_size );
	if ( ret ) {
		printf( "tar: cant decode file size: %d\n", ret );
		ret = ret;
		goto DONE;
	}

	arch->f_start_block = arch->block;
	arch->f_blocks = (arch->f_size/TAR_BLOCKSIZE) + (arch->f_size%TAR_BLOCKSIZE?1:0);
	printf( "tar: %s %ld bytes, %d blocks\n",  phdr->name,
			arch->f_size, arch->f_blocks );

	strncpy( arch->filename, phdr->name, 100 );
	arch->filename[100] = 0;

	ret = 0;
DONE:
	/* reset fileinfos on error */
	if ( ret ) {
		arch->f_size=0;
		arch->f_start_block = 0;
		arch->f_blocks=0;
		arch->filename[0]=0;
	}
	return ret;
}

/* check wether or not we have a posix tar header */
static int tar_hdr_chk( tar_hdr_posix_t *hdr )
{
	int ret = 0;

	if ( !hdr )
		return -EINVAL;

	/* check magic */
	ret = strncmp( hdr->magic, TMAGIC, 5 ); // "ustar"
	if ( ret != 0 )
		return -EINVAL;

	return 0;
}

/* get file size from header */
static int tar_hdr_fsize_get( tar_hdr_posix_t *hdr, unsigned long *size )
{
	return tar_from_oct( hdr->size, 11, size );
}

/* get filename from header */
static int tar_hdr_fname_get( tar_hdr_posix_t *hdr, char **fname )
{
	if ( !hdr || !fname )
		return -EINVAL;

	*fname = hdr->name;
	return 0;
}

/* decode octal BCD data from header */
static int tar_from_oct( unsigned char *start, size_t digits, unsigned long *value )
{
	unsigned long val = 0l;

	DBG( 5, "%s: start=%p, digits=%d, value=%p\n", __FUNCTION__,
			start, digits, value );

	if ( !start || !value )
		return -EINVAL;

	while ( digits ) {
		if ( (*start > '7') || (*start < '0') )
			return -EINVAL;

		DBG( 5, "%s: %c, val=%ld\n", __FUNCTION__, *start, val );

		val = ( val << 3 ) | (*start - '0');
		--digits;
		++start;
	}

	*value = val;
	return 0;
}

