/*
 *  tftp.c
 *
 *  Copyright 1996-2001 Russell King
 *  Copyright 2002,2003 Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <blob/util.h>
#include <blob/serial.h>
#include <blob/errno.h>

#include <net/net.h>
#include <net/ip_bits.h>
#include <net/arp.h>
#include <net/tftp.h>

#undef TFTP_DEBUG
//#define TFTP_DEBUG
#ifdef TFTP_DEBUG
static int dbg = 10;
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... ) ;
#endif

static const unsigned char tftp_txpacket[] =
	{					/*+destination				*/
	 					/*+source				*/
	 					/* proto				*/
	 /* IP */
	 0x45, 0x00, 0x00, 0xa4,		/* ver/ihl, TOS, length			*/
	 0x00, 0x00,				/*+id					*/
	 0x00, 0x00, 0x40, 0x11,		/* frag off, ttl, proto			*/
	 0x00, 0x00,				/*+checksum				*/
	 0x00, 0x00, 0x00, 0x00,		/*+source address			*/
	 0x00, 0x00, 0x00, 0x00,		/*+destination address			*/
	 /* UDP */
	 0x80, 0x00,				/* source port				*/
	 0x00, 0x45,				/* destination port			*/
	 0x00, 0x90,				/* length (inc header)			*/
	 0x00, 0x00,				/*+checksum				*/
	 /* TFTP */
	 0x00, 0x01				/* opcode: RRQ				*/
	 					/* filename				*/
	};

static unsigned char serverhwaddress[6];
static char octet[] = "octet";
static unsigned char tftp_tx[200];
static unsigned char tftp_rx[800];
static          int  serverport;
static unsigned int  tftp_length;
static unsigned int  block;
static unsigned long tftp_received;

int tftp_rcv(unsigned long *addr)
{
	unsigned int tftp_off, tftp_sz, nblk;
	unsigned char *tftp;

	DBG( 5, "%s: addr=%p *addr=0x%08x\n", __FUNCTION__, addr, *addr );

	tftp_off = ipbits_rcv (tftp_rx, 800, &tftp_sz, serverport, 0x8000);
	if (!tftp_off || tftp_rx[tftp_off] != 0) 
		return 0;

	tftp = tftp_rx + tftp_off;
	switch (tftp[1]) {
	case 5: /* error */
		printascii ("\nTFTP error: ");
		printascii ((const char *)(tftp + 4));
		printch ('\n');
		return -2;
		break;

	case 3:	/* OpCode: DATA, acknowledge block */
		if (block == 0) {
			serverport = (tftp[-8] << 8) | tftp[-7];
							/* copy replied-source port	*/
			tftp_tx[OFF_UDP + 2] = serverport >> 8;
			tftp_tx[OFF_UDP + 3] = serverport;
			tftp_tx[OFF_UDP + 5] = 12;
			tftp_tx[OFF_IP + 3] = 32;
			tftp_tx[OFF_TFTP + 1] = 4;	/* OpCode: ACK */
			block = 1;
			tftp_length = 12;
		}
		nblk = (tftp[2] << 8) | tftp[3];
		if (nblk == (block%(1<<16))) {
			tftp_tx[OFF_TFTP + 2] = nblk >> 8;
			tftp_tx[OFF_TFTP + 3] = nblk;
			ipbits_tx (tftp_tx, tftp_length );
			memcpy((void *)*addr, tftp + 4, tftp_sz - 4);
			*addr += tftp_sz - 4;
			tftp_received += tftp_sz - 4;
			if ((block & 31) == 1) {
				printch ('^');
				DBG( 5, "%s: nblk=%d, block=%d, sz=%d\n", __FUNCTION__, nblk, block, tftp_sz );
			}
			block ++;
		} else {
			if (nblk == ((block-1)%(1<<16)))
				ipbits_tx (tftp_tx, tftp_length );
		}
		if (tftp_sz != 516)
			return 1;
		return -1;
	default:
		return 0;
	}
}	

static int tftp_received_len = 0;
int get_tftp_received_len( void )
{
	return tftp_received_len;
}

int do_tftp(char *file, unsigned long addr, unsigned long *size )
{
	int ret;
	int finished;
	unsigned int i;
	int retry = 8;
	
	DBG( 1, "%s: file='%s', addr=%08lx\n", __FUNCTION__, file, addr );

	if (do_arp(serverhwaddress, serveripaddress) < 0) {
		DBG( 1, "%s, failed to call do_arp\n");
		return -ETIMEOUT;
	}
	DBG( 1, "succeed in do_arp\n");
	serverport = -1;
	tftp_length = 144;
	block = 0;
	tftp_received = 0;

	memzero(tftp_tx, sizeof (tftp_tx));
	memcpy(tftp_tx + 14, tftp_txpacket, sizeof (tftp_txpacket));
	memcpy(tftp_tx + OFF_ETHER, serverhwaddress, 6);
	memcpy(tftp_tx + OFF_IP + 12, clientipaddress, 4);
	memcpy(tftp_tx + OFF_IP + 16, serveripaddress, 4);

	for (i = 0; file[i]; i++)
		tftp_tx[OFF_TFTP + 2 + i] = file[i];
	memcpy(tftp_tx + OFF_TFTP + 3 + i, octet, 6);

	printascii("TFTPing ");
	printascii(file);

	finished = 0;
	do {
		printch('*');
		ipbits_tx(tftp_tx, tftp_length);
		for (i = 0; i < 0x00040000 && !finished; i++) {
			ret = tftp_rcv(&addr);
			switch (ret) {
			case 1:
				finished = 1;
				break;
			case -1:
				i = 0;
				break;
			case -2:
				i = 0x00040000;
				retry = 0;
				break;
			}
		}
	} while (retry-- && !finished);

	if ( finished ) {
		printf( " OK.\n" );
		printf( "received %d blocks (%ld bytes)\n", block, tftp_received );
		*size = tftp_received;
		tftp_received_len = tftp_received;
		ret = 0;
	} else {
		printf( " ERROR\n" );
		ret = -1;
	}
	return ret;
}
