/**************************************************************************
**
**  (C) Copyright December 2006 Marvell International Ltd.
**
**  All Rights Reserved.
**
**  This software file (the "File") is distributed by Marvell International
**  Ltd. under the terms of the GNU General Public License Version 2, June
**  1991 (the "License").  You may use, redistribute and/or modify this File
**  in accordance with the terms and conditions of the License, a copy of
**  which is available along with the File in the license.txt file or by
**  writing to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
**  Boston, MA 02111-1307 or on the worldwide web at
**  http://www.gnu.org/licenses/gpl.txt.
**
**  THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE
**  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
**  PURPOSE ARE EXPRESSLY DISCLAIMED.  The License provides additional
**  details about this warranty disclaimer.
**
**  Filename: TBBLIB.cpp
**
**  This file contains miscellaneous functions for the TIM binary image
**  build tool, TBBTIM.
**
******************************************************************************/

#define TBBLIB
#define WTPDEBUG 1
#define DBGVERIFY 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if LINUX
  #include <sys/types.h>
  #include <fcntl.h>
  #include <sys/stat.h>
  #include <limits.h>
  #include "typedef.h"
#endif

#if WINDOWS

#endif

#if TRUSTED
  #include "ippcp.h"
  #include "TBBLIB.h"
#endif

#include "NTBBLIB.h"
#include "tim.h"


extern bool bVerbose;

#pragma warning ( disable : 4096 4081 ) // Disable warning messages

bool GenerateHash (FILE *hImageFile,char *szImageFilename,IMAGE_INFO *pImageInfo)
{
	int BlockNo, iSize, Rem, i;
 Ipp8u pSrc[8192];
 IppStatus Status;
 IppsSHA1State *pState = NULL;
 IppsSHA256State *pState256 = NULL;

	// A hash is generated from each image in its entirety and stored in the
	// Hash array of the IMAGE_INFO struct for that image in the TIM.

	fseek (hImageFile,0L,SEEK_SET); // Set position to SOF

	BlockNo = pImageInfo->ImageSizeToHash / 8192;
	Rem = pImageInfo->ImageSizeToHash % 8192;

	// SHA Initialization

	if (pImageInfo->HashAlgorithmID == 32)
	{
		Status = ippsSHA256GetSize (&iSize);
		pState256 = (IppsSHA256State *) malloc (iSize);
		if (pState256 == NULL)
		{
			printf ("////////////////////////////////////////////////////////////\n");
			printf ("  Error: Memory allocation for pState256 failed in GenerateHash()!\n");
			printf ("////////////////////////////////////////////////////////////\n");
			return FALSE;
		}
		Status = ippsSHA256Init (pState256);
	}
	else
	{
		Status = ippsSHA1GetSize (&iSize);
		pState = (IppsSHA1State *) malloc (iSize);
		if (pState == NULL)
		{
			printf ("////////////////////////////////////////////////////////////\n");
			printf ("  Error: Memory allocation for pState failed in GenerateHash()!\n");
			printf ("////////////////////////////////////////////////////////////\n");
			return FALSE;
		}
		Status = ippsSHA1Init (pState);
	}

	if (Status != ippStsNoErr)
	{
		if (pState != NULL) free (pState);
		if (pState256 != NULL) free (pState256);
		printf ("  Error: Failed generating hash data for file: %s\n",szImageFilename);
		return FALSE;
	}

	// SHA Update

	if (pImageInfo->HashAlgorithmID == 32)
	{
		for (i = 0; i < BlockNo; i++)
		{
			iSize = (int)fread (pSrc,sizeof(char),8192,hImageFile);
			ippsSHA256Update (pSrc,8192,pState256);
		}
		iSize = (int)fread (pSrc,sizeof(char),Rem,hImageFile);
		Status = ippsSHA256Update (pSrc,Rem,pState256);
	}
	else
	{
		for (i = 0; i < BlockNo; i++)
		{
			iSize = (int)fread (pSrc,sizeof(char),8192,hImageFile);
			for(int q=0; q<8192;q+=4)
				Endian_Convert(*(unsigned int *)&pSrc[q], (unsigned int *)&pSrc[q]);
			ippsSHA1Update (pSrc,8192,pState);
		}
		iSize = (int)fread (pSrc,sizeof(char),Rem,hImageFile);
		for(int q=0; q<Rem;q+=4)
				Endian_Convert(*(unsigned int *)&pSrc[q], (unsigned int *)&pSrc[q]);
		Status = ippsSHA1Update (pSrc,Rem,pState);
	}

	if (Status != ippStsNoErr)
	{
		if (pState != NULL) free (pState);
		if (pState256 != NULL) free (pState256);
		printf ("  Error: Failed generating hash data for file: %s\n",szImageFilename);
		return FALSE;
	}

	// SHA Final

	if (pImageInfo->HashAlgorithmID == 32)
	{
		Status = ippsSHA256Final ((Ipp8u *)pImageInfo->Hash,pState256);
		free (pState256);
	}
	else
	{
		Status = ippsSHA1Final ((Ipp8u *)pImageInfo->Hash,pState);
		free (pState);
	}
	if (Status != ippStsNoErr)
	{
		printf ("  Error: Failed generating hash data for file: %s\n",szImageFilename);
		return FALSE;
	}

	for (i = 0; i < 5; i++)
	{
		Endian_Convert (((UINT_T *)pImageInfo->Hash)[i],&(((UINT_T *)pImageInfo->Hash)[i]));
	}

	printf("\n\n");
	#if WTPDEBUG
		DumpHashToFile (szImageFilename,pImageInfo);
	#endif

	return TRUE;
}

bool DumpHashToFile (char *szImageFilename,IMAGE_INFO *pImageInfo)
{
	char szHashFileName[MAX_PATH];
	FILE *hHash = NULL;
	char HashValue[12],TempValue[12];

	if (strstr (szImageFilename,".") == NULL)
	{
		sprintf (szHashFileName,"%s_hash.txt",szImageFilename);
	}
	else
	{
		char *TempName = strtok (szImageFilename,".");
		sprintf (szHashFileName,"%s_hash.txt",TempName);
	}

	if ((hHash = fopen (szHashFileName,"w+")) == NULL)
	{
		printf ("\n  Error: fopen failed to open hash file: %s.\n",szHashFileName);
		return FALSE;
	}

	for (int i = 0; i < 5; i++)
	{
		sprintf (TempValue,"%X",pImageInfo->Hash[i]);

		if (strlen (TempValue) < 8)
			sprintf (HashValue,"0%s\n",TempValue);
		else
			sprintf (HashValue,"%s\n",TempValue);

		fwrite (HashValue,strlen (HashValue),1,hHash);
	}

	return TRUE;
}

bool GenerateKeyHash (KEY_MOD *key)
{
	int i;
	unsigned int count, KeyLength;
	IppStatus ippstatus;
	unsigned int KeyData[128]; //Max 2048 bit key * 2  (for exponent and modulus)

	// Extract the KeyLength value from TIM Digital Signature in words
	KeyLength = key->ModulusSize / 4;

	// Fill in the Exponent
	for (i = 0; i < (int)KeyLength; i++)
		Endian_Convert(key->RSAPublicExponent[i], &KeyData[i]);

	// Fill in the modulus
	for (i = 0; i < (int)KeyLength; i++)
		Endian_Convert(key->RSAModulus[i], &KeyData[KeyLength + i]);

	if (key->HashAlgorithmID == 32)
	{
		ippstatus = ippsSHA256MessageDigest ((Ipp8u *)KeyData, KeyLength*4*2,(Ipp8u *)key->KeyHash);
	}
	else
	{
		ippstatus = ippsSHA1MessageDigest ((Ipp8u *)KeyData, KeyLength*4*2,(Ipp8u *)key->KeyHash);
	}

	// Do an endian conversion 1 word (4 bytes) at a time
	for (count = 0; count < (key->HashAlgorithmID)/4; count++ )
		Endian_Convert (*(UINT_T *)(&key->KeyHash[count]),(UINT_T *)(&key->KeyHash[count]));

	if (ippstatus != ippStsNoErr) return FALSE;

	return TRUE;
}

int GetTimHashSize (CTIM *pTimHeader)
{
	int iTotalHashSize;

	iTotalHashSize = sizeof (CTIM);
	iTotalHashSize += sizeof (IMAGE_INFO) * pTimHeader->NumImages;
	iTotalHashSize += sizeof (KEY_MOD) * pTimHeader->NumKeys;
	iTotalHashSize += pTimHeader->SizeOfReserved;
	iTotalHashSize += sizeof (PLAT_DS) - (sizeof (UINT_T) * MAXKEYSIZEWORDS);

	return iTotalHashSize;
}

bool GenerateTimHash (FILE *hTimFile,CTIM *pTimHeader,PLAT_DS *pDs,Ipp8u *pDsHash, unsigned int HashSize)
{
// A hash of the TIM minus the digital signature is generated by
// this function.
	Ipp8u TimHash[32];
	Ipp8u PaddedTimHash[256],PaddedTimHash_WordSwapped[256];
	Ipp8u DER_SHA1_T[] = {0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 0x1a, 0x05, 0x00, 0x04, 0x14};
	Ipp8u DER_SHA256_T[] = {0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0x04, 0x20};
	Ipp8u *DER_T;
	long lFileSize;
	unsigned char *TimBuffer;
	FILE *hTimHash;
	char HashValue[80];
	int i,j,paddingsize;
	unsigned long TimSizeWithNoDS;
	unsigned int SHAHashSize; //size of digest in words
	unsigned int sizeofT;

	//get the max size of the TIM to be hashed
	TimSizeWithNoDS = GetTimHashSize (pTimHeader);

	fseek (hTimFile,0L,SEEK_END); // Set position to EOF

	//determine how much of the TIM needs to be read into buffer
    if ( HashSize == TimSizeWithNoDS )
		lFileSize = ftell (hTimFile); // Get position of file, thus the file size.
	else
		lFileSize = HashSize;

//	printf("size read from file: %d\n", lFileSize);

	fseek (hTimFile,0L,SEEK_SET); // Set position SOF

	if (NULL == (TimBuffer = (unsigned char *)malloc (lFileSize)))
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: Memory allocation for TimBuffer failed in GenerateTimHash()!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}
	//read in TIM file
	if (fread (TimBuffer,1,lFileSize,hTimFile) != (unsigned int)lFileSize)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: GenerateTimHash() couldn't read all of the %ld bytes\n",lFileSize);
		printf ("from the TIM binary image!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (TimBuffer != NULL) free (TimBuffer);
		return FALSE;
	}
	//endian convert TIM to big endian for IPP reasons(LSW->MSW)
	for(int c=0; c<lFileSize; c+=4){
		Endian_Convert(*(unsigned int *)&TimBuffer[c], (unsigned int *)&TimBuffer[c]);
	}

	memset (PaddedTimHash,0,256);//clear the hash buffers
	memset (TimHash,0,32);//clear the hash buffers

	//Calculate the SHA hash of the TIM file
	if (pDs->HashAlgorithmID == 32){
		ippsSHA256MessageDigest (TimBuffer,lFileSize,TimHash);
		SHAHashSize = 8;
		sizeofT = 19;
		DER_T = DER_SHA256_T;
	}
	else
	{
		ippsSHA1MessageDigest (TimBuffer,lFileSize,TimHash);
		SHAHashSize = 5;
		sizeofT = 15;
		DER_T = DER_SHA1_T;
	}

	if (TimBuffer != NULL) free (TimBuffer);

	if (pDs->DSAlgorithmID == Intel_DS )
	{
		//endian convert hash back to little endian to satisfy Caddo
		for (i = 0; (unsigned int)i < SHAHashSize; i++)
			Endian_Convert (((UINT_T *)TimHash)[i],&((UINT_T *)TimHash)[i]);

		//add TIM hash to the padded TIM hash buffer
		for (i = 0; i < SHAHashSize*4; i++)
			PaddedTimHash[i] = TimHash[i];

		//now add the padding
		for (i = SHAHashSize;(unsigned int)i < ((pDs->ModulusSize / 4) - 1); i++)
			((Ipp32u *)PaddedTimHash)[i] = 0xBEBEBEBE;

		((Ipp32u *)PaddedTimHash)[i] = 0x6EBEBEBE;
	}
	else if ( pDs->DSAlgorithmID == PKCS1_v1_5 )
	{

		// Encryption block(EB) formatting for PKCS
		//=================================
		// EB = 00 || BT || PS || 00 || T || D
		//==================================
		// BT = 00, 01(private key operation) or 02(public key operation)
		// PS = 00, FF, randomly generated(based on BT values)
		// T = DER encoding of DigestInfo
		// D = SHA hash
		// All values are in octets (octet means byte)

		//PS must be k-3-|T|-|D| bytes long
		paddingsize = pDs->ModulusSize - 3 - sizeofT - SHAHashSize * 4;

		PaddedTimHash[0] = 0x00; //byte 0 is always 0x0
		PaddedTimHash[1] = 0x01;//private key op, with Padding = 0xFF
		for (j = 0, i = 2; j < paddingsize; j++, i++)
			PaddedTimHash[i] = 0xFF;

		PaddedTimHash[i++] = 0x00;

		for(j = 0; j < sizeofT; j++, i++)
			PaddedTimHash[i] = DER_T[j];

		//now add Tim Hash to the Padded hash
		for (j = 0; j < SHAHashSize * 4; j++, i++){
			PaddedTimHash[i] = TimHash[j];
		}
		// //PKCS_1.5:Convert padded Tim hash to BE and from MSW to LSW.
		for (i = 0; i < pDs->ModulusSize; i++)
			PaddedTimHash_WordSwapped[i] = PaddedTimHash[pDs->ModulusSize - 1 - i];

		for (i = 0; (unsigned int)i < pDs->ModulusSize; i++)
			PaddedTimHash[i] = PaddedTimHash_WordSwapped[i];

	}
	else
	{
	   printf("  Error: DSA Algorithm needs to be Marvell method or PKCS!\n\n");
	   return FALSE;
	}

	#if DBGVERIFY
	memset (HashValue,'\0',80);

	if ((hTimHash = fopen ("TimHash.txt","w+")) == NULL)
	{
		printf ("\n  Error: fopen failed to TimHash.txt\n");
	}

	if (bVerbose)
		printf ("\nTIM Hash prior to Encryption for size %d\n",pDs->ModulusSize);

	for (i = 0; (unsigned int)i < ((pDs->ModulusSize / 4)); i++)
	{
		if (((i / 4) * 4 == i) && (i != 0))
		{
			if (bVerbose) printf ("\n");
			sprintf (HashValue,"\n0x%08x ",((Ipp32u *) PaddedTimHash)[i]);
		}
		else sprintf (HashValue,"0x%08x ",((Ipp32u *) PaddedTimHash)[i]);
		if (bVerbose) printf ("0x%08x ",((Ipp32u *)PaddedTimHash)[i]);

		if (hTimHash != NULL) fwrite (HashValue,strlen (HashValue),1,hTimHash);
	}
	if (bVerbose) printf ("\n");

	if (hTimHash != NULL) fclose (hTimHash);

	#endif

	memcpy (pDsHash,PaddedTimHash,256);

	return TRUE;
}

bool DumpTimHashToBinFile (PLAT_DS *pDs,Ipp8u *pDsHash, char *HashFileName)
{
	FILE *hTimHash;
	Ipp8u DsHash[256];

	memset (DsHash,0,256);
	memcpy (DsHash,pDsHash,256);

	if ((hTimHash = fopen (HashFileName,"wb")) == NULL)
	{
		printf ("\n  Error: fopen failed to TimHash.bin\n");
		return FALSE;
	}

	fwrite (DsHash,sizeof (Ipp8u),pDs->ModulusSize,hTimHash);

	if (hTimHash != NULL) fclose (hTimHash);

	return TRUE;
}

bool EncryptSignature (bool encrypt,PLAT_DS *pDs,Ipp8u *pDsHash)
{
#ifdef TBBDS_LIB
	int kn = 4096, kp = 2048, size, length = 128;
	IppsRSAState *key = NULL;
	IppsBigNumState *x, *y;
	IppsBigNumSGN sgn;
	IppStatus Status;
	Ipp8u DsHash[256];
	FILE *hEncrypted;
	char HashValue[80];
//	Ipp32u DigitalSignatureWordSwapped[MAXKEYSIZEWORDS];

	// This function encrypts the hash of the entire TIM minus installation
	// of the digital signature. Information from the DSA algorithm module, as
	// defined in the descriptor text file, is also utilized. The starting point
	// in the descriptor would look something like this, ( DSA Algorithm ID: 1 ).
	// The result from this is the digital signature which is appended
	// to the end of the TIM by another function.


	// retrieve the size required to configure IppsRSAState context
	Status = ippsRSAGetSize ( kn, kp, IppRSApublic, &size );
	if (Status != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsRSAGetSize has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}
	// allocate required memory space for IppsRSAState
	key = (IppsRSAState *)malloc (size);

	// Initialize the context for the RSA encryption with the allocated memory.
	Status = ippsRSAInit (kn, kp,IppRSApublic,key);
	if (Status != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsRSAInit has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		return FALSE;
	}

	// get size of the ippsBigNumState context
	if (ippsBigNumGetSize (length,&size) != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsBigNumGetSize has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		return FALSE;
	}

	x = (IppsBigNumState *)malloc (size);
	y = (IppsBigNumState *)malloc (size);

	if (ippsBigNumInit (length,x) != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsBigNumInit has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	if (ippsBigNumInit (length,y) != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsBigNumInit has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	ippsSet_BN (IppsBigNumPOS,(int)(pDs->ModulusSize / 4),(Ipp32u *)pDs->RSAModulus,x);
	Status = ippsRSASetKey (x,IppRSAkeyN,key);

	if (Status != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsRSASetKey has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	if (encrypt == true )
	{
		ippsSet_BN (IppsBigNumPOS,(int)(pDs->ModulusSize / 4),(Ipp32u *)RsaPrivateKey,x);
		Status = ippsRSASetKey (x,IppRSAkeyE,key);
	}
	else
	{
		ippsSet_BN (IppsBigNumPOS,1,(Ipp32u *)&pDs->RSAPublicExponent,x);
		Status = ippsRSASetKey (x,IppRSAkeyE,key);
	}

	if (Status != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsRSASetKey has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	memcpy (DsHash,pDsHash,256);
	ippsSet_BN (IppsBigNumPOS,(int)(pDs->ModulusSize / 4),(Ipp32u *)DsHash,x);

	Status = ippsRSAEncrypt (x,y,key);
	if (Status != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsRSAEncrypt has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	memset (DigitalSignature,0,MAXKEYSIZEWORDS);

	if (ippsGet_BN (&sgn, &length, DigitalSignature, y) != ippStsNoErr)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: EncryptSignature Error:\n");
		printf ("ippsGet_BN has failed!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (key != NULL) free (key);
		if (x != NULL) free (x);
		if (y != NULL) free (y);
		return FALSE;
	}

	if (key != NULL) free (key);
	if (x != NULL) free (x);
	if (y != NULL) free (y);

/*	if (pDs->DSAlgorithmID == PKCS1_v1_5)
	{
		unsigned int j;
		//PKCS_1.5:Convert Digital signature from MSW to LSW.
		for (unsigned int i=0, j=(int)(pDs->ModulusSize / 4)-1; i<(int)(pDs->ModulusSize / 4);i++,j--)
		{
			DigitalSignatureWordSwapped[j]=DigitalSignature[i];
		}
		//PKCS_1.5:Copy swapped array into the global array.
		for (int i=0;i<(int)(pDs->ModulusSize / 4);i++)
		{
			DigitalSignature[i]=DigitalSignatureWordSwapped[i];
		}
	}*/

	#if DBGVERIFY
	memset (HashValue,'\0',80);
	if ( encrypt == true )
	{
		if ((hEncrypted = fopen ("EncryptedTimHash.txt","w+")) == NULL)
		{
			printf ("\n  Error: fopen failed to EncryptTimHash.txt\n");
		}

		if (bVerbose) printf ("\nSignature After the Encryption\n");

		for (int i = 0; i < (int)(pDs->ModulusSize / 4); i++)
		{
			if ((i & 3) == 0 && i != 0)
			{
				if (bVerbose) printf ("\n");
				sprintf (HashValue,"%x",DigitalSignature[i]);

				if (strlen (HashValue) < 7)
					sprintf (HashValue,"\n0x00%x ",DigitalSignature[i]);
				else if (strlen (HashValue) < 8)
					sprintf (HashValue,"\n0x0%x ",DigitalSignature[i]);
				else
					sprintf (HashValue,"\n0x%x ",DigitalSignature[i]);
			}
			else
			{
				sprintf (HashValue,"%x",DigitalSignature[i]);

				if (strlen (HashValue) < 7)
					sprintf (HashValue,"0x00%x ",DigitalSignature[i]);
				else if (strlen (HashValue) < 8)
					sprintf (HashValue,"0x0%x ",DigitalSignature[i]);
				else
					sprintf (HashValue,"0x%x ",DigitalSignature[i]);
			}

			if (bVerbose) printf ("0x%8x ", DigitalSignature[i]);
			if (hEncrypted != NULL) fwrite (HashValue,strlen (HashValue),1,hEncrypted);
		}
		if (bVerbose) printf ("\n\n");

		if (hEncrypted != NULL) fclose (hEncrypted);
	}
	else
	{
		if ((hEncrypted = fopen ("DecryptedTimHash.txt","w+")) == NULL)
		{
			printf ("\n  Error: fopen failed to EncryptTimHash.txt\n");
		}

		if (bVerbose) printf ("\nSignature After the Encryption\n");

		for (int i = 0; i < (int)(pDs->ModulusSize / 4); i++)
		{
			if ((i & 3) == 0 && i != 0)
			{
				if (bVerbose) printf ("\n");
				sprintf (HashValue,"%x",DigitalSignature[i]);

				if (strlen (HashValue) < 7)
					sprintf (HashValue,"\n0x00%x ",DigitalSignature[i]);
				else if (strlen (HashValue) < 8)
					sprintf (HashValue,"\n0x0%x ",DigitalSignature[i]);
				else
					sprintf (HashValue,"\n0x%x ",DigitalSignature[i]);
			}
			else
			{
				sprintf (HashValue,"%x",DigitalSignature[i]);

				if (strlen (HashValue) < 7)
					sprintf (HashValue,"0x00%x ",DigitalSignature[i]);
				else if (strlen (HashValue) < 8)
					sprintf (HashValue,"0x0%x ",DigitalSignature[i]);
				else
					sprintf (HashValue,"0x%x ",DigitalSignature[i]);
			}

			if (bVerbose) printf ("0x%8x ", DigitalSignature[i]);
			if (hEncrypted != NULL) fwrite (HashValue,strlen (HashValue),1,hEncrypted);
		}
		if (bVerbose) printf ("\n\n");

		if (hEncrypted != NULL) fclose (hEncrypted);
	}
	#endif

	return TRUE;
#else //TBBIM_LIB
  printf ("////////////////////////////////////////////////////////////\n");
  printf ("  Error: EncryptSignature Error:\n");
  printf ("  TBBIM lib don't support!\n");
  printf ("////////////////////////////////////////////////////////////\n");
  return FALSE;
#endif // end of TBBDS_LIB
}

void ConvertMSWToLSWAndFlip(PLAT_DS *pDs)
{
	unsigned int i, j;
	Ipp32u DigitalSignatureWordSwapped[MAXKEYSIZEWORDS];

	if (pDs->DSAlgorithmID == PKCS1_v1_5)
	{

		//PKCS_1.5:Convert Digital signature from MSW to LSW.
		for (unsigned int i=0, j=(int)(pDs->ModulusSize / 4)-1; i<(int)(pDs->ModulusSize / 4);i++,j--)
		{
			DigitalSignatureWordSwapped[j]=DigitalSignature[i];
		}
		//PKCS_1.5:Copy swapped array into the global array.
		for (int i=0;i<(int)(pDs->ModulusSize / 4);i++)
		{
			DigitalSignature[i]=DigitalSignatureWordSwapped[i];
		}
	}

}

bool InstallDigitalSignature (char *szFilename,bool bIsAppend)
{
	FILE *hFile;

	// This function has two purposes. It can either append a
	// digital signature to the end of a Trusted Image Module
	// or save the signature in a seperate file.

	if (bIsAppend)
	{
		char Attribute[] = "ab+";
		hFile = fopen (szFilename,Attribute);
	}
	else
	{
		char Attribute[] = "wb";
		hFile = fopen (szFilename,Attribute);
	}

	if (hFile == NULL)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: InstallDigitalSignature Error:\n");
		printf ("Can not open file name %s !",szFilename);
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}

	fwrite (DigitalSignature,sizeof (Ipp32u),MAXKEYSIZEWORDS,hFile);

	if (hFile != NULL) fclose (hFile);

	return TRUE;
}

bool GenerateOtpHash (CTIM *pTimHeader,PLAT_DS *pDs)
{
	UINT_T i, KeyLengthMask = 0x000000FF;
	char HashValue[12],TempValue[12];
	FILE *hOtpHash;

	// Used for OEM Verfication Key
	UINT_T OEMKeyData[66];
	UINT_T OEMKeyDataC[5];
	UINT_T TempOEMKeyData[66];
	// Used for JTAG Verfication Key
	UINT_T 	JTAGKeyData[128];
	UINT_T 	JTAGKeyDataC[5];

	// Key Lengths
    UINT16_T TIM_KeyLength;     // Key Length in words
    UINT16_T KeyHashLength;     // Length of target area to be hashed in words

	// Extract the KeyLength value from TIM Digital Signature in ISRAM in words
	TIM_KeyLength    = (((pDs->ModulusSize) & KeyLengthMask) / 4);

	/**************** Calculate DS Hash for the OEM Verification Keys ********/

	KeyHashLength = (TIM_KeyLength * 2); // Length in words

	for (i = 0; i < TIM_KeyLength; i++)
	{
		OEMKeyData[i] = pDs->RSAPublicExponent[i];
		OEMKeyData[(i + TIM_KeyLength)] = pDs->RSAModulus[i];
	}

 	// Do Endian conversion, 1 word at a time
	for (i = 0; i < KeyHashLength; i++)
	{
     	Endian_Convert (OEMKeyData[i], &OEMKeyData[i]);
	}

	// Do the OEM Hash
	ippsSHA1MessageDigest((unsigned char *)OEMKeyData,(4 * KeyHashLength),(unsigned char *)OEMKeyDataC);

	for (i = 0; i < 5; i++)
	{
		pDs->Hash[i] = OEMKeyDataC[i];
		Endian_Convert (OEMKeyDataC[i], &pDs->Hash[i]);

	}

	if (bVerbose) printf ("\nOEM Hash of 20 bytes\n");
	for (i = 0; i < 5; i++)
	{
		if (bVerbose) printf ("%X\n",OEMKeyDataC[i]);
	}

	/**************** Calculate Hash for the JTAG *********************/

	KeyHashLength = TIM_KeyLength * 2; // Length is in words

	for (i = 0; i < KeyHashLength; i++)
	{
		JTAGKeyData[i] = 0x0;
	}

	// Now fill in the Exponent
	JTAGKeyData[0] = pDs->RSAPublicExponent[0];

	// Fill in the modules
	for (i = 0; i < TIM_KeyLength; i++)
	{
		JTAGKeyData[(TIM_KeyLength + i)] = pDs->RSAModulus[i];
	}

	// Do Endian conversion, 1 word at a time
	for(i=0; i < KeyHashLength; i++)
	{
     	Endian_Convert(JTAGKeyData[i], &JTAGKeyData[i]);
	}

	// Do the JTAG Hash
	ippsSHA1MessageDigest((unsigned char *)JTAGKeyData,(4 * KeyHashLength),(unsigned char *)JTAGKeyDataC);

	if (bVerbose) printf ("\nJTAG Hash of 20 bytes\n");
	for (i = 0; i < 5; i++)
	{
		if (bVerbose) printf ("%X\n",JTAGKeyDataC[i]);
	}

	/**************** Dump combined hashes to file ********************/

	if ((hOtpHash = fopen ("OtpHash.txt","w+")) == NULL)
	{
		printf ("\n  Error: fopen failed to OtpHash.txt\n");
		return FALSE;
	}

	memset (TempValue,'\0',12);
	memset (HashValue,'\0',12);

	for (i = 0; i < 5; i++)
	{
		sprintf (TempValue,"%X",OEMKeyDataC[i]);

		if (strlen (TempValue) < 8)
			sprintf (HashValue,"0%s\n",TempValue);
		else
			sprintf (HashValue,"%s\n",TempValue);

		if (hOtpHash != NULL) fwrite (HashValue,strlen (HashValue),1,hOtpHash);
	}

	for (i = 0; i < 5; i++)
	{
		sprintf (TempValue,"%X",JTAGKeyDataC[i]);

		if (strlen (TempValue) < 8)
			sprintf (HashValue,"0%s\n",TempValue);
		else
			sprintf (HashValue,"%s\n",TempValue);

		if (hOtpHash != NULL) fwrite (HashValue,strlen (HashValue),1,hOtpHash);
	}

	if (hOtpHash != NULL) fclose (hOtpHash);
	return TRUE;
}

bool VerifyTimHash (char *szTimFilename,PLAT_DS *pDs,Ipp8u *pDsHash, long TIMSizeNoDS )
{
	FILE *hTimFile;
	Ipp8u DsHash[256], DsHashOriginal[256];
	unsigned char *TimBuffer;
	int i, j;

	memset (DsHash,0,256);

	if (pDs->DSAlgorithmID == Intel_DS )
	{
		for (i = 0; (unsigned int)i < ((pDs->ModulusSize / 4) - 1); i++)
			((Ipp32u *)DsHash)[i] = 0xBEBEBEBE;

		((Ipp32u *)DsHash)[i] = 0x6EBEBEBE;
	}
	else if (pDs->DSAlgorithmID == PKCS1_v1_5 )
	{
		// Encryption block(EB) formatting for PKCS
		//=================================
		// EB = 00 || BT || PS || 00 || D
		//==================================
		// BT = 00, 01(public key operation) or 02(private key)
		// PS = 00, FF, randomly generated(based on BT values)
		// D = Data quantity
		// All values are in octets
		DsHash[0] = 0x0;
		DsHash[1] = 0x01;
		for (j = 2; (unsigned int)j < ((pDs->ModulusSize / 4) - 1); j++)
			DsHash[j] = 0xFF;
		DsHash[j] = 0x0;
		DsHash[j+1] = (Ipp8u)TIMSizeNoDS;
	}
	else
	{
	   printf("  Error: DSA Algorithm needs to be Marvell method or PKCS!\n\n");
	   return FALSE;
	}

	if ((hTimFile = fopen (szTimFilename,"rb")) == NULL)
	{
		printf ("  Error: Cannot open file name %s !",szTimFilename);
		return FALSE;
	}

	fseek (hTimFile,0L,SEEK_SET); // Set position to SOF

	if (NULL == (TimBuffer = (unsigned char *)malloc (TIMSizeNoDS)))
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: Memory allocation for TimBuffer failed in VerifyTimHash()!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (hTimFile != NULL) fclose (hTimFile);
		return FALSE;
	}

	if (fread (TimBuffer,1,TIMSizeNoDS,hTimFile) != (unsigned int)TIMSizeNoDS)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: VerifyTimHash() couldn't read all of the %ld bytes.\n",TIMSizeNoDS);
		printf ("from the TIM binary image!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		if (TimBuffer != NULL) free (TimBuffer);
		if (hTimFile != NULL) fclose (hTimFile);
		return FALSE;
	}

	for(int c=0; c<TIMSizeNoDS; c+=4)
		Endian_Convert(*(unsigned int *)&TimBuffer[c], (unsigned int *)&TimBuffer[c]);

	if (hTimFile != NULL) fclose (hTimFile);

	if (pDs->HashAlgorithmID == 32)
		ippsSHA256MessageDigest (TimBuffer,TIMSizeNoDS,DsHash);
	else
		ippsSHA1MessageDigest (TimBuffer,TIMSizeNoDS,DsHash);

	if (TimBuffer != NULL) free (TimBuffer);

	for (i = 0; (unsigned int)i < pDs->HashAlgorithmID; i += 4 )
		Endian_Convert (*(UINT_T *)&DsHash[i],(UINT_T *)&DsHash[i]);

	// copy previously calculated digital signature to DsHashOriginal
	memcpy (DsHashOriginal,(Ipp8u *)DigitalSignature,pDs->HashAlgorithmID);

	if (memcmp (DsHash,DsHashOriginal,pDs->HashAlgorithmID) != 0)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("  Error: TIM Hash verification fails in VerifyTimHash()!\n");
		printf ("////////////////////////////////////////////////////////////\n");
		return FALSE;
	}
	else
	{
		printf("**TIM Hash verification successful!!!**\n\n");
	}

	#if DBGVERIFY
	FILE *hTimHash;
	char HashValue[80];

	memset (HashValue,'\0',80);

	if ((hTimHash = fopen ("TimHashVerify.txt","w+")) == NULL)
	{
		printf ("\n  Error: fopen failed to TimHashVerify.txt\n");
	}

	if (bVerbose)
		printf ("\nRe-verified hash for size %d\n",pDs->ModulusSize);

	for (i = 0; (unsigned int)i < ((pDs->ModulusSize / 4)); i++)
	{
		if (((i / 4) * 4 == i) && (i != 0))
		{
			if (bVerbose) printf ("\n");
			sprintf (HashValue,"\n0x%x ",((Ipp32u *) DsHash)[i]);
		}
		else sprintf (HashValue,"0x%x ",((Ipp32u *) DsHash)[i]);
		if (bVerbose) printf ("0x%x ",((Ipp32u *)DsHash)[i]);

		if (hTimHash != NULL) fwrite (HashValue,strlen (HashValue),1,hTimHash);
	}
	if (bVerbose) printf ("\n");

	if (hTimHash != NULL) fclose (hTimHash);
	#endif

	return TRUE;
}

bool FillDataArrays (FILE *hKeyFile,DWORD *dwArray,int iSize)
{
	char fLine[100], *pData = NULL;
	DWORD ImageOffset;
	int iTotal = 0;

	// This function grabs one line at a time and parses out tokens that
	// are delimited by spaces, tabs or carraige returns. This process will
	// continue until it reads a line containing a semicolon or it reaches
	// the end of the file. The semicolon signifies a new section of the
	// descriptor text file.

	while (fgets (fLine,100,hKeyFile) != NULL)
	{
		if ((iTotal * 4) > iSize)
		{
			printf ("////////////////////////////////////////////////////////////\n");
			printf ("Data bytes read > max allowable\n");
			return FALSE;
		}
		if (strstr (fLine,":") != NULL)
		{
			fseek (hKeyFile,ImageOffset,SEEK_SET); // Set position to Image offset
			break;
		}

		pData = strtok (fLine," \n\t\r");

		if (pData != NULL)
		{
			*dwArray++ = Translate (pData); iTotal++;

			while ((pData = strtok (NULL," \n\t\r")) != NULL)
			{
				*dwArray++ = Translate (pData); iTotal++;
			}
		}
		ImageOffset = ftell (hKeyFile);
	}

	if ((iTotal * 4) != iSize)
	{
		printf ("////////////////////////////////////////////////////////////\n");
		printf ("Data bytes read = %d\n",iTotal * 4);
		return FALSE;
	}

	return TRUE;
}

bool VerifyNumberOfKeys (FILE *hKeyFile)
{
	char fLine[100];
	char szTemp[100];
	int i, NumberFound = 0;

	while (fgets (fLine,100,hKeyFile) != NULL)
	{
		if (strstr (fLine,"Number of Keys") != NULL)
		{
			if (GetParameter (fLine,szTemp,"Number of Keys",FALSE) == FALSE) return FALSE;

			i = 0;
			while (szTemp[i] != '\0')
			{
				if ( !(isdigit (szTemp[i])) )
				{
					printf ("////////////////////////////////////////////////////////////\n");
					printf ("  Error: Number of Keys: declaration statement can only be set to numbers!\n");
					printf ("Format: Number of Keys: 2\n");
					printf ("////////////////////////////////////////////////////////////\n");
					rewind (hKeyFile);
					return FALSE;
				}

				++i;
			}

			for (i = 0,NumberFound = 0; i < atoi (szTemp); i++)
			{
				while (fgets (fLine,100,hKeyFile) != NULL)
				{
					if (strstr (fLine,"Key ID:") != NULL)
					{
						NumberFound++;
					}
				}
			}

			if (NumberFound != atoi (szTemp))
			{
				printf ("////////////////////////////////////////////////////////////\n");
				printf ("  Error: The number of keys found does not match the number of keys\n");
				printf ("set in Number of Keys: declaration statement!\n\n");
				printf ("Number of Keys: %d\n",atoi (szTemp));
				printf ("Number of keys found = %d\n",NumberFound);
				printf ("////////////////////////////////////////////////////////////\n");
				rewind (hKeyFile);
				return FALSE;
			}
		}
	}

	rewind (hKeyFile);

	return TRUE;
}


