/******************************************************************************
**	(C) Copyright 2007 Marvell International Ltd. 
**	All Rights Reserved
******************************************************************************/

/******************************************************************************
**
**  COPYRIGHT (C) 2000, 2001 Intel Corporation.
**
**  This software as well as the software described in it is furnished under
**  license and may only be used or copied in accordance with the terms of the
**  license. The information in this file is furnished for informational use
**  only, is subject to change without notice, and should not be construed as
**  a commitment by Intel Corporation. Intel Corporation assumes no
**  responsibility or liability for any errors or inaccuracies that may appear
**  in this document or any software that may be provided in association with
**  this document.
**  Except as permitted by such license, no part of this document may be
**  reproduced, stored in a retrieval system, or transmitted in any form or by
**  any means without the express written consent of Intel Corporation.
**
**  FILENAME:       xllp_udc.c
**
**  PURPOSE:   		This files contains the driver for USB Device controller
**                  (UDC).
**
**  LAST MODIFIED: $Modtime: 4/04/06 4:26p $
**
******************************************************************************/

/*
*******************************************************************************
*   HEADER FILES
*******************************************************************************
*/

#include "xllp_udc.h"
#include "usbdriver.h"
#include "misc.h"
#include "usb_descriptors.h"

extern OPT_USB_DESCRIPTOR_LOADS USBDescriptors[NUM_USB_DESCRIPTORS];
extern UINT_T	USBStringIndexTable[MAX_USB_STRINGS];
extern const XLLP_UDC_EP_CONFIG_TABLE_T defaultUdcEndpointConfigTable[];
extern UINT_T FindVendorRequestInTIM (XLLP_UINT8_T bRequest, XLLP_UINT8_T bmRequestType, XLLP_UINT32_T *pTxBuffEp0, XLLP_INT32_T *ppacketLength);

/******************************************************************************

  Function Name:
    XllpUdcComputeConfigRegisterValue

  Description:
    This function computes the configuration values that will be loadded to
    the Endpoints Configuration Registers UDCCRA-UDCCRX, based on the entries
    in the Endpoints Configuration table. It stores the configuration value in
    the Endpoints Configuration table. It also returns the configuration register's
    value.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_EP_CONFIG_TABLE_T pEndpointsConfigTable - a pointer to the Endpoints
                                                         Configuration Table.
  Output Arguments:
    P_XLLP_UINT32_T pEndpointConfigRegValue

  Return Value:
    None

*******************************************************************************/

void XllpUdcComputeConfigRegisterValue (P_XLLP_UDC_EP_CONFIG_TABLE_T pEndpointsConfigTable,
                                        P_XLLP_UINT32_T pEndpointConfigRegValue)
{
    // Compute the Configuration register value
//    pEndpointsConfigTable->endpointConfigValue =
    *pEndpointConfigRegValue =
    // Usb Configuration Number: (1-3)
    pEndpointsConfigTable->usbConfigNum << XLLP_UDC_UDCCRZ_CN_SHIFT  |
    // Usb Interface Number: (1-7)
    pEndpointsConfigTable->usbInterfaceNum << XLLP_UDC_UDCCRZ_IN_SHIFT |
    // Usb Interface Alternate Settings Number: (1-7)
    pEndpointsConfigTable->usbIntAltSettingsNum << XLLP_UDC_UDCCRZ_AISN_SHIFT |
    // Usb Endpoint Number: (1-15)
    pEndpointsConfigTable->usbEndpointNum << XLLP_UDC_UDCCRZ_EN_SHIFT |
    // Usb Endpoint type: Bulk, Iso, Interrupt
    pEndpointsConfigTable->usbEndpointType << XLLP_UDC_UDCCRZ_ET_SHIFT |
    // Usb Endpoint direction: IN, OUT
    pEndpointsConfigTable->usbEndpointDirection << XLLP_UDC_UDCCRZ_ED_SHIFT |
    // Max. Packet Size: (1-1023)
    ((pEndpointsConfigTable->maxPacketSize << XLLP_UDC_UDCCRZ_MPS_SHIFT) &
                                                    XLLP_UDC_UDCCRZ_MPS_MASK) |
    // Double Buffering enabled\disabled and enable Endpoint
    (pEndpointsConfigTable->doubleBuffEnabled << XLLP_UDC_UDCCRZ_DE_SHIFT) |
                                                    XLLP_UDC_UDCCRZ_EE;

    // Return the Configuration register value
//    *pEndpointConfigRegValue = pEndpointsConfigTable->endpointConfigValue;
}

/******************************************************************************

  Function Name:
    XllpUdcConfigureEndpoints

  Description:
    This function configures and enables all of the Endpoints that are
    used for each of the USB Configurations and Interfaces.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T XllpUdcConfigureEndpoints (P_XLLP_UDC_T pUdcHandle)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    P_XLLP_UDC_EP_CONFIG_TABLE_T pConfigTable = (P_XLLP_UDC_EP_CONFIG_TABLE_T)
                                                           pUdcHandle->pConfigTable;
    XLLP_UINT32_T j, i = 1;
    XLLP_STATUS_T error = 0;
    XLLP_UINT32_T   endpointConfigRegValue;
    //XLLP_UINT32_T Tmp = 0;

    // Program the Endpoint Configuration Registers:
    // Disable UDC controller by writing zero to the UDE bit (UDCCR.UDE = 0),
    // that enables the Endpoint Configuration Registers for write operation.
    pRegs->UDCCR = 0;

    while (pConfigTable[i].usbEndpointNum != 0)
    {
        // Compute the values that would be loaded to the Endpoints Configuration
        // Registers UDCCRA-UDCCRX, from the data in the Endpoints Configuration Table.
        XllpUdcComputeConfigRegisterValue (&pUdcHandle->pConfigTable[i],
                                           &endpointConfigRegValue);

        // Write the configuration value to the Endpoints Configuration Registers UDCCRA-UDCCRX.
        j = pConfigTable[i].udcEndpointNum;
        pRegs->UDCCRZ[j] = endpointConfigRegValue;
        ++i;
    }

    // Enable UDC controller by writing one to the bit UDE (UDCCR.UDE = 1),
    // that disables the Endpoint Configuration Registers for write operation.
    pRegs->UDCCR = XLLP_UDC_UDCCR_UDE;

    // Check if the Endpoint Memory Configuration is valid by reading the status of
    // the Endpoint Memory Configuration Error UDCCR.EMCE bit.
    if (pRegs->UDCCR & XLLP_UDC_UDCCR_EMCE)
    {
        // This bit is set if the endpoint memory configuration has an error.
        // The driver must clear UDCCR.EMCE bit by writing one to it before again enabling the UDC.
        pRegs->UDCCR = XLLP_UDC_UDCCR_EMCE;
        error = XLLP_TRUE;
    }

    return error;
}

/******************************************************************************

  Function Name:
    XllpUdcEnableInterrupt

  Description:
    This function enables the Endpoints or Events interrupt.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcEnableInterrupt (P_XLLP_UDC_T pUdcHandle,
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

    // Set the corresponding bit in the UDC's interrupt control registers UDCICR0 and UDCICR1
    // to enable the endpoint's and event's interrupt requests.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
   	    pRegs->UDCICR0 |= (1 << bitNum);
 	}
    else if (udcEndpointNum < INT_ENDPOINT_X + 1)
    {
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
   	    pRegs->UDCICR1 |= (1 << bitNum);
    }
    else if (udcEndpointNum >= INT_RESET)
    {
   	    bitNum = (udcEndpointNum & 0x1f);
   	    pRegs->UDCICR1 |= (1 << bitNum);
    }
}

/******************************************************************************

  Function Name:
    XllpUdcDisableInterrupt

  Description:
    This function disables the Endpoints or Events interrupts.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcDisableInterrupt (P_XLLP_UDC_T pUdcHandle,
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

    // Clear the corresponding bit in the UDC's interrupt control registers UDCICR0 and UDCICR1
    // to disable the endpoint's and event's interrupt requests.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
        pRegs->UDCICR0 &= ~(1 << bitNum);
 	}
    else if (udcEndpointNum < (INT_ENDPOINT_X + 1))
    {
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
        pRegs->UDCICR1 &= ~(1 << bitNum);
    }
    else if (udcEndpointNum >= INT_RESET)
    {
     	bitNum = (udcEndpointNum  & 0x1f);
        pRegs->UDCICR1 &= ~(1 << bitNum);
    }
}

/******************************************************************************

  Function Name:
    XllpUdcClearInterrupt

  Description:
    This function clears the Endpoints or Events interrupts.
    It also clears the Endpoint FIFO Error bit in the Endpoint's UDCCSR register.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcClearInterrupt (P_XLLP_UDC_T pUdcHandle,
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

    // Write one to the corresponding bit of the UDC's interrupt status registers
    // UDCISR0 and UDCISR1 to clear the endpoint's and event's interrupt.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
        pRegs->UDCISR0 = (1 << bitNum);
 	}
    else if (udcEndpointNum < (INT_ENDPOINT_X + 1))
    {
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
        pRegs->UDCISR1 = (1 << bitNum);

        // Clear the Endpoint FIFO Error bit in the Endpoint's UDCCSR register
        pRegs->UDCCSR[udcEndpointNum] =
          (pRegs->UDCCSR[udcEndpointNum] & XLLP_UDC_UDCCSR_DME) | XLLP_UDC_UDCCSR_EFE;
    }
    else if (udcEndpointNum >= INT_RESET)
    {
     	bitNum = (udcEndpointNum  & 0x1f);
        pRegs->UDCISR1 = (1 << bitNum);
    }
}



/******************************************************************************

  Function Name:
    XllpUdcWriteDwordFifo

  Description:
    This function writes the specified number of bytes to the Endpoint's FIFO.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_VUINT32_T pReg    - a pointer to the Endpoints FIFO
    void *src               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to write

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcWriteDwordFifo(P_XLLP_VUINT32_T pReg, void *src, XLLP_INT32_T len)
{
	XLLP_INT32_T    i;
	XLLP_UINT32_T	temp[16];
	P_XLLP_UINT32_T p1;
    P_XLLP_VUINT8_T pBuff, pTemp;

	if (len == 0)
	{
		return;
	}

	if (len > sizeof(temp) && ((XLLP_UINT32_T)src & 0x3))
	{
		// ERROR alignment, return
		return;
	}

	if ((XLLP_UINT32_T) src & 0x3)
	{
		memcpy (temp, src, len);
		p1 = temp;
	}
	else
	{
		p1 = src;
	}

	if (((XLLP_INT32_T) len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++)
		{
			*pReg = *p1++;
	    }
	}
	else
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*pReg = *p1++;
	    }

        pBuff = (P_XLLP_VUINT8_T)p1;
        pTemp = (P_XLLP_VUINT8_T)pReg;
	    for (i=0; i<(len&0x3); i++)
	    {
		    *pTemp = *pBuff++;
	    }
	}
}


/******************************************************************************

  Function Name:
    XllpUdcReadDwordFifo

  Description:
    This function reads the specified number of bytes to the Endpoint's FIFO.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_VUINT32_T pReg    - a pointer to the Endpoints FIFO
    void *buf               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to read

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcReadDwordFifo (P_XLLP_VUINT32_T pReg, void *buf, XLLP_INT32_T len)
{
	XLLP_INT32_T    i;
	XLLP_UINT32_T	temp[16], rdat;
	P_XLLP_UINT32_T p1;
    XLLP_BOOL_T     copyback;
    P_XLLP_VUINT8_T pBuff, pTemp;

	if (((XLLP_UINT32_T) buf & 0x3) != 0  && len > sizeof(temp))
	{
		// ERROR alignment, return
		return;
	}

	if (((XLLP_UINT32_T)buf & 0x3) != 0)
	{
		p1 = temp;
		copyback = XLLP_TRUE;
	}
	else
	{
		p1 = buf;
		copyback = XLLP_FALSE;
	}


	if (((XLLP_INT32_T)len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++)
		{
			*p1++ = *pReg;
	    }
	}
	else
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*p1++ = *pReg;
	    }
	    rdat = *pReg;

        pBuff = (P_XLLP_VUINT8_T)p1;
        pTemp = (P_XLLP_VUINT8_T)&rdat;

	    for (i=0; i<(len&0x3); i++)
	    {
           *pBuff++ = *pTemp++;
	    }
	}

	if (copyback)
	{
		memcpy (buf, temp, len);
	}
}


/******************************************************************************

  Function Name:
    XllpUdcBuildListOfActiveEndpoints

  Description:
    This function fills the list of the endpoints that are used in the active
    configuration, active interface, and active alternate interface. This list
    will be processed from the main UDC's interrupt handler to accelerate the
    determination which of the Endpoints A-X generated the interrupt.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UINT32_T configuration
    XLLP_UINT32_T interface
    XLLP_UINT32_T settings

  Output Arguments:
    P_XLLP_UINT32_T pNumActiveEndpoints - the total number of the endpoints in
                                          the list of the active endpoints will
                                          be returned.
  Return Value:
    None

*******************************************************************************/

void XllpUdcBuildListOfActiveEndpoints (P_XLLP_UDC_T pUdcHandle,
                                        XLLP_UINT32_T configuration,
                                        XLLP_UINT32_T interface,
                                        XLLP_UINT32_T settings,
                                        P_XLLP_UINT32_T pNumActiveEndpoints)
{
    XLLP_UINT32_T i = 1, j = 0;
    P_XLLP_UDC_EP_CONFIG_TABLE_T pConfigTable = (P_XLLP_UDC_EP_CONFIG_TABLE_T)
                                                                    pUdcHandle->pConfigTable;

    // Find all of the Endpoints that supports the active configuration and interface
    // and build the table of active endpoints.
    while (pConfigTable[i].usbConfigNum > 0)
    {
        if (pConfigTable[i].usbConfigNum == configuration)
        {
            if (pConfigTable[i].usbInterfaceNum == interface)
            {
                if (pConfigTable[i].usbIntAltSettingsNum == settings)
                {
                    pUdcHandle->listOfActiveEndpoints[j].udcEndpointNum =
                                                                pConfigTable[i].udcEndpointNum;
                    pUdcHandle->listOfActiveEndpoints[j].usbEndpointNum =
                                                                pConfigTable[i].usbEndpointNum;

                    // Enable endpoint's packet complete interrupt
                    XllpUdcEnableInterrupt (pUdcHandle,
                                            (XLLP_UDC_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                            PACKET_COMPL_INT);

                    // Enable endpoint's FIFO error interrupt
                    XllpUdcEnableInterrupt (pUdcHandle,
                                            (XLLP_UDC_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                            FIFO_ERROR_INT);
                    ++j;
                }
            }
        }
        ++i;
    }

    *pNumActiveEndpoints = j;
}


/******************************************************************************

  Function Name:
    XllpUdcControlProcessIdle

  Description:
    This function is an entry point of Endpoint 0 state machine.
    It processes the setup stage of the setup transaction.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - this argument is not used

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/
void XllpUdcControlProcessIdle (P_XLLP_UDC_T pUdcHandle,
                                P_XLLP_UINT32_T pControlXferStatus)
{
    XLLP_USB_SETUP_DATA_T reqData;
    P_XLLP_UINT32_T rxBuffP = (P_XLLP_UINT32_T)&reqData;
    P_XLLP_UINT32_T pTxBuffEp0;
    XLLP_UINT32_T tempDataLen;
    XLLP_UINT32_T stringIndex;
    XLLP_UINT32_T stringLength;
    XLLP_UINT32_T tempReqType;
    XLLP_INT32_T packetLength;

	volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

    // Determine if UDCCS0.OPC and UDCCS0.SA both are set.
    // It means that OUT packet in the Endpoint 0 buffer identifying SETUP transaction,
    if (pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA))
    {
        // Read data from Endpoint 0 data buffer
        // The data packet of the setup stage is always 8 bytes
        XllpUdcReadDwordFifo (&pRegs->UDCDR[UDC_EP_0], rxBuffP, 8);

        // Clear UDCCS0.OPC and UDCCS0.SA
        pRegs->UDCCSR[UDC_EP_0] = (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA);

        // Parse the data that is in the buffer
        // Identify RequestType
        tempReqType = (reqData.bmRequestType & XLLP_UDC_REQ_TYPE_MASK) >> 5;
        if (tempReqType == STANDARD_REQ)
        {
		    // Identify Standard Request type
            switch (reqData.bRequest)
            {
                case GET_DESCRIPTOR:
                    // Get the number of data bytes in the data stage that follows
                    // This is the maximum number, and the device can return this
                    // number of bytes or fewer.
                    pUdcHandle->controlXfer.dataLengthEp0 = reqData.wLength;

                    // Identify type of descriptor
                    switch (reqData.wValue >> 8)
                    {
                        case DEVICE_DESCRIPTOR:
                           	// Read the length of the device descriptor
                            	// and compare with number of bytes in the data stage
                            	tempDataLen = USBDescriptors[USB_DEVICE_DESCRIPTOR & 0xFF].DesSize;
                            	if (pUdcHandle->controlXfer.dataLengthEp0 > tempDataLen)
                              		pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                            	// Initialize the Tx buffer pointer with the device descriptor's address
                            	pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_DEVICE_DESCRIPTOR & 0xFF].pDesPayload;
                           break;
                        case CONFIG_DESCRIPTOR:
                            // Read the total length of the configuration descriptor
                            // and compare with number of bytes in the data stage
                            tempDataLen = USBDescriptors[USB_CONFIG_DESCRIPTOR & 0xFF].DesSize;
                            if (pUdcHandle->controlXfer.dataLengthEp0 > tempDataLen)
                              pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                            // Initialize the Tx buffer pointer with the config. descriptor's address
                            pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_CONFIG_DESCRIPTOR & 0xFF].pDesPayload;
                            break;
                        case STRING_DESCRIPTOR:
							stringIndex = ((reqData.wValue) & 0xFF);

							// Bounds protection
							if (stringIndex >= MAX_USB_STRINGS)
							{
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].pDesPayload;
	                           		break;
	                        }

							switch (USBStringIndexTable[stringIndex])
							{
								case USB_LANGUAGE_STRING_DESCRIPTOR:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_LANGUAGE_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_LANGUAGE_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_LANGUAGE_STRING_DESCRIPTOR & 0xFF].pDesPayload;
                            		break;

								case USB_MANUFACTURER_STRING_DESCRIPTOR:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_MANUFACTURER_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_MANUFACTURER_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_MANUFACTURER_STRING_DESCRIPTOR & 0xFF].pDesPayload;
                            		break;
								case USB_PRODUCT_STRING_DESCRIPTOR:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_PRODUCT_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_PRODUCT_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_PRODUCT_STRING_DESCRIPTOR & 0xFF].pDesPayload;
  									break;
								case USB_SERIAL_STRING_DESCRIPTOR:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_SERIAL_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_SERIAL_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;
                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_SERIAL_STRING_DESCRIPTOR & 0xFF].pDesPayload;
  									break;

								case USB_INTERFACE_STRING_DESCRIPTOR:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_INTERFACE_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_INTERFACE_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_INTERFACE_STRING_DESCRIPTOR & 0xFF].pDesPayload;
 									break;
								default:
									stringLength = ((reqData.wLength) & 0xFFFF);
									if (stringLength > USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize)
		                            	tempDataLen = USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize;
		                            else
		                            	tempDataLen = stringLength;

                		            pUdcHandle->controlXfer.dataLengthEp0 = tempDataLen;
                        		    // Initialize the Tx buffer pointer with the config. descriptor's address
                            		pUdcHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].pDesPayload;
	                           		break;
								}
							break;
                    }

                    // Preload EP0 FIFO to get ready for IN transaction that follows.
                    // Write up to sixteen bytes to the Endpoint 0 data buffer
                    packetLength = pUdcHandle->controlXfer.dataLengthEp0;
                    pTxBuffEp0 = pUdcHandle->controlXfer.pTxBuffEp0;
                    // Write to the Endpoint 0 data buffer
                    if (packetLength >= MPS_CNTRL_16)
                    {
                        XllpUdcWriteDwordFifo (&pRegs->UDCDR[UDC_EP_0], pTxBuffEp0, MPS_CNTRL_16);
                        pTxBuffEp0 = (P_XLLP_UINT32_T)((XLLP_UINT32_T)pTxBuffEp0 + MPS_CNTRL_16);
                        pUdcHandle->controlXfer.dataLengthEp0 -= MPS_CNTRL_16;
                        // Update pointer to the transmit buffer
                        pUdcHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;
                    }
                    else
                    {
                        XllpUdcWriteDwordFifo (&pRegs->UDCDR[UDC_EP_0], pTxBuffEp0, packetLength);
                        pUdcHandle->controlXfer.dataLengthEp0 -= packetLength;
                    }

                    // Set UDCCSR0.IPR = 1 if the packet smaller than 16 bytes has been loaded
                    // to the Endpoint 0 FIFO
                    if (packetLength < MPS_CNTRL_16)
                    {
                        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
                        // Set the status of the Endpoint 0 state machine to XFER data
                        pUdcHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
                    }
                    else
                    {
                        // Set the status of the Endpoint 0 state machine to IN data
                        pUdcHandle->controlXfer.statusEp0 = EP0_IN_DATA_STATE;
                    }
                    break;

                 case SET_DESCRIPTOR:
                    // Save descriptor index (lower byte of wValue)
                    pUdcHandle->controlXfer.descIndexEp0 = reqData.wValue & 0xff;
                    // Save descriptor type (upper byte of wValue)
                    pUdcHandle->controlXfer.descTypeEp0 = reqData.wValue >> 8;
                    // Get the number of data bytes in the data stage that follows
                    pUdcHandle->controlXfer.dataLengthEp0 = reqData.wLength;
                    // Initialize Rx buffer pointer with address where the descriptor will be stored
                    pUdcHandle->controlXfer.pRxBuffEp0 = pUdcHandle->controlXfer.outDataEp0;
                    // Set UDCCSR0.IPR in case of premature STATUS IN stage
                    pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
                    // Set the status of the Endpoint 0 state machine to OUT data
                    pUdcHandle->controlXfer.statusEp0 = EP0_OUT_DATA_STATE;
            }
        }

#if 1
        // Process Vendor Request
        else if (tempReqType == VENDOR_REQ)
        {
			// Set this flag to indicate processing a vendor request
            pUdcHandle->vendorReq.vendorReqComplete = XLLP_FALSE;

            // Get ready to process the data that would be sent in the data stage
            pUdcHandle->vendorReq.recipient = reqData.bmRequestType;
            pUdcHandle->vendorReq.vendorReqType = reqData.bRequest;
            pUdcHandle->vendorReq.vendorReqEndpointNum = reqData.wValue;

            // Get the number of data bytes in the data stage that follows
            pUdcHandle->controlXfer.dataLengthEp0 = reqData.wLength;

            // 0: Host to device
            if ( !(reqData.bmRequestType & 0x80) )
            {
            	// Set UDCCSR0.IPR in case of premature STATUS IN stage
            	pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;

				// Get ready to input from the host
                pUdcHandle->controlXfer.statusEp0 = EP0_OUT_DATA_STATE;
            	pUdcHandle->controlXfer.pRxBuffEp0 = pUdcHandle->controlXfer.outDataEp0;

				// Don't allow an input buffer overflow
				if (reqData.wLength > 64*4)
					pUdcHandle->controlXfer.dataLengthEp0 = 64*4;
            }
            // 1: Device To Host	Transfer Type
            else
                XllpUdcProcessVendorRequest (pUdcHandle); // ROM compression
        }
#endif

        else if (tempReqType == CLASS_REQ)
        {}
    }


    *pControlXferStatus = 0;
}

#if 1
/******************************************************************************

  Function Name:
    XllpUdcProcessVendorRequest

  Description:
    Vendor Request format:
        bmRequestType: bits 6 and 5 = 10 (vendor defined)
                       bits 4 through 0 = 00010 (endpoint)
        bRequest:      SETUP IN EP   = 0x01
                       SETUP OUT EP  = 0x02
                       SETUP INT EP  = 0x03
                       SETUP LOOPBACK  = 0x04
                       bRequest is passed in requestType

        wValue:        LSB - Rx(OUT) Endpoint number
                       MSB - Tx(IN) Endpoint number
                       wValue is used to pass IN and/or OUT Endpoints
        wIndex:        Not Used
        wLength:       4 bytes will be sent in the data stage to the UDC
                       these bytes can be used to control various functions

    To setup loopback: 1. Send SETUP OUT EPx request
                       2. Send SETUP LOOPBACK on EPx and EPy

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None.

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/
XLLP_STATUS_T XllpUdcProcessVendorRequest (P_XLLP_UDC_T pUdcHandle)
{
	XLLP_UINT32_T Retval;
    P_XLLP_UINT32_T pTxBuffEp0;
	XLLP_INT32_T packetLength;
	volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

	// Are we in a host to device request?
	if (!(pUdcHandle->vendorReq.recipient & 0x80))
	{
    	// No more data left, force UDC to send zero lenght packet.
        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
        // Set Endpoint 0 state machine to the End of transfer state
        pUdcHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
	}
	else
	{
		// Is this request registered in the TIM?
		Retval = FindVendorRequestInTIM (pUdcHandle->vendorReq.vendorReqType, pUdcHandle->vendorReq.recipient,
						(XLLP_UINT32_T *) &pTxBuffEp0, &packetLength);

		// Preload EP0 FIFO to get ready for IN transaction that follows.
    	// Write up to sixteen bytes to the Endpoint 0 data buffer
    	pUdcHandle->controlXfer.dataLengthEp0 = packetLength;
        pUdcHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;

		// Write to the Endpoint 0 data buffer
        if (packetLength >= MPS_CNTRL_16)
        {
        	XllpUdcWriteDwordFifo (&pRegs->UDCDR[UDC_EP_0], pTxBuffEp0, MPS_CNTRL_16);
            pTxBuffEp0 = (P_XLLP_UINT32_T)((XLLP_UINT32_T)pTxBuffEp0 + MPS_CNTRL_16);
            pUdcHandle->controlXfer.dataLengthEp0 -= MPS_CNTRL_16;
            // Update pointer to the transmit buffer
            pUdcHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;
        }
        else
        {
            XllpUdcWriteDwordFifo (&pRegs->UDCDR[UDC_EP_0], pTxBuffEp0, packetLength);
            pUdcHandle->controlXfer.dataLengthEp0 -= packetLength;
        }

        // Set UDCCSR0.IPR = 1 if the packet smaller than 16 bytes has been loaded
        // to the Endpoint 0 FIFO
        if (packetLength < MPS_CNTRL_16)
        {
             pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
             // Set the status of the Endpoint 0 state machine to XFER data
             pUdcHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
        }
        else
        {
             // Set the status of the Endpoint 0 state machine to IN data
             pUdcHandle->controlXfer.statusEp0 = EP0_IN_DATA_STATE;
        }
	}
    pUdcHandle->vendorReq.vendorReqComplete = XLLP_TRUE;

	return 0;
}
#endif


/******************************************************************************

  Function Name:
    XllpUdcControlProcessOutData

  Description:
    This function services the data OUT state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpUdcControlProcessOutData (P_XLLP_UDC_T pUdcHandle,
                                   P_XLLP_UINT32_T pControlXferStatus)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    XLLP_INT32_T packetLength;

    // Check UDCCS0.OPC and UDCCS0.SA
    if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) ==
            (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA) )
    {
        // UDCCSR0.OPC and UDCCSR0.SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
    }

    else if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) == 0)
    {
        // UDCCSR0.OPC = 0 and UDCCSR0.SA = 0 this is premature STATUS stage of
        // the SETUP transaction.
        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
    }

    else if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) == XLLP_UDC_UDCCSR0_OPC)
    {
        // Read data from the Endpoint 0 FIFO
        if (pRegs->UDCCSR[UDC_EP_0] & XLLP_UDC_UDCCSR0_RNE)
        {
            packetLength = pRegs->UDCBCR[UDC_EP_0] & XLLP_UDC_UDCBCR_BC_MASK;
            XllpUdcReadDwordFifo (&pRegs->UDCDR[UDC_EP_0], pUdcHandle->controlXfer.pRxBuffEp0, packetLength);
            pUdcHandle->controlXfer.dataLengthEp0 -= packetLength;
            *pControlXferStatus = 0;
        }

         // Clear UDCCS0.OPC
        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_OPC;
        // Set UDCCS0.IPR in case of premature STATUS IN stage
        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
    }
}

/******************************************************************************

  Function Name:
    XllpUdcControlProcessInData

  Description:
    This function services the data IN state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpUdcControlProcessInData (P_XLLP_UDC_T pUdcHandle,
                                  P_XLLP_UINT32_T pControlXferStatus)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    XLLP_INT32_T packetLength;

    // Check UDCCSR0.OPC and UDCCSR0.SA
    if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) ==
            (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA) )
    {
        // UDCCSR0.OPC and UDCCSR0.SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
    }

    else if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) == XLLP_UDC_UDCCSR0_OPC)
    {
        // UDCCSR0.OPC = 1 and UDCCSR0.SA = 0 this is premature STATUS stage of
        // the SETUP transaction. Clear UDCCSR0.OPC and empty transmit FIFO
        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_FTF;

        // Reset all state machine variables
        pUdcHandle->controlXfer.dataLengthEp0 = 0;
        pUdcHandle->controlXfer.pTxBuffEp0 = NULL;
        pUdcHandle->controlXfer.pRxBuffEp0 = NULL;

        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
    }

    else if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) == 0)
    {
        packetLength = pUdcHandle->controlXfer.dataLengthEp0;
        if ( packetLength > MPS_CNTRL_16 )
               packetLength = MPS_CNTRL_16;

        XllpUdcWriteDwordFifo (&pRegs->UDCDR[UDC_EP_0], pUdcHandle->controlXfer.pTxBuffEp0, packetLength);

        pUdcHandle->controlXfer.dataLengthEp0 -= packetLength;
        if ( MPS_CNTRL_16 != packetLength )
        {
            // short packet or ZLP

            // No more data left, force UDC to send zero lenght packet.
            // Set Endpoint 0 state machine to the End of transfer state
            pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_IPR;
            pUdcHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
        }
        else
        {
            pUdcHandle->controlXfer.pTxBuffEp0 += (MPS_CNTRL_16 >> 2);
            // It needs more data to transmit,
            // so on the next interrupt go back to the same state.
        }

        *pControlXferStatus = 0;
    }
}

/******************************************************************************

  Function Name:
    XllpUdcControlProcessEndXfer

  Description:
    This function services the Status state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpUdcControlProcessEndXfer (P_XLLP_UDC_T pUdcHandle,
                                   P_XLLP_UINT32_T pControlXferStatus)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;

    // Check UDCCSR0.OPC and UDCCSR0.SA
    if ((pRegs->UDCCSR[UDC_EP_0] & (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) ==
            (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA) )
    {
        // UDCCSR0.OPC and UDCCSR0.SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
		return;
    }

    else if ((pRegs->UDCCSR[UDC_EP_0] &
         (XLLP_UDC_UDCCSR0_OPC | XLLP_UDC_UDCCSR0_SA)) == XLLP_UDC_UDCCSR0_OPC)
    {
        // UDCCSR0.OPC = 1 and UDCCSR0.SA = 0 this is STATUS stage of
        // the IN transaction. Clear UDCCSR0.OPC.
        pRegs->UDCCSR[UDC_EP_0] = XLLP_UDC_UDCCSR0_OPC;
        // Initialize Endpoint 0 state machine to the Idle state
        pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
		return;
    }
    *pControlXferStatus = 0;
}



/******************************************************************************

  Function Name:
    XllpUdcEp0InterruptHandler

  Description:
    This function is an interrupt handler for Endpoint 0 of UDC.
    It dispatches the functions of the Endpoint 0 in according to the status
    of the Endpoint 0 state machine.
    It handles the Get_Descriptor and Set_Descriptor standard requests
    and the Vendor specific requests from the USB Host.

    The Set_Configuration and Set_Interface Requests are serviced upon
    the invocation of the Configuration interrupt.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcEp0InterruptHandler (P_XLLP_UDC_T pUdcHandle)
{
    XLLP_UINT32_T controlXferStatus = 0, statusEp0;

    // Determine the status of the Endpoint 0 state machine
    // and call the apropriate function
    do
    {
       statusEp0 = pUdcHandle->controlXfer.statusEp0;
       if (statusEp0 == EP0_IDLE_STATE)
       {
           XllpUdcControlProcessIdle (pUdcHandle, &controlXferStatus);
       }
       else if (statusEp0 == EP0_OUT_DATA_STATE)
       {
           XllpUdcControlProcessOutData (pUdcHandle, &controlXferStatus);

#if 1
           if (pUdcHandle->vendorReq.vendorReqComplete == XLLP_FALSE && pUdcHandle->controlXfer.dataLengthEp0 == 0 )
           {
               XllpUdcProcessVendorRequest (pUdcHandle);
           }
#endif

       }
       else if (statusEp0 == EP0_IN_DATA_STATE)
       {
            XllpUdcControlProcessInData (pUdcHandle, &controlXferStatus);
       }
       else if (statusEp0 == EP0_END_XFER_STATE)
       {
            XllpUdcControlProcessEndXfer (pUdcHandle, &controlXferStatus);
       }
    } while (controlXferStatus == EP0_JUMP_TO_IDLE);
}

/******************************************************************************

  Function Name:
    XllpUdcConfigInterruptHandler

  Description:
    This is the UDC's Configuration interrupt handler. This function is called
    when the USB Host issues the Set_Config or Set_Interface requests on the USB bus.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcConfigInterruptHandler (P_XLLP_UDC_T pUdcHandle)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    XLLP_UINT32_T numActiveEndpoints;
    XLLP_UINT32_T configuration, interface, settings;

    // Get Active UDC Configuration number
    configuration = (pRegs->UDCCR & XLLP_UDC_UDCCR_ACN_MASK) >> XLLP_UDC_UDCCR_ACN_SHIFT;

    // Get Active UDC Interface Number
    interface = (pRegs->UDCCR & XLLP_UDC_UDCCR_AIN_MASK) >> XLLP_UDC_UDCCR_AIN_SHIFT;

    // Get Active UDC Alternative Interface Setting Number
    settings = (pRegs->UDCCR & XLLP_UDC_UDCCR_AAISN_MASK) >> XLLP_UDC_UDCCR_AAISN_SHIFT;

    // Switch Endpoint memory to active configuration
    pRegs->UDCCR |= XLLP_UDC_UDCCR_SMAC;

    if (pUdcHandle->usbConfigNumActive != configuration)
    {
        // Build the list of the endpoints in active configuration
        XllpUdcBuildListOfActiveEndpoints (pUdcHandle,
                                           configuration, interface, settings,
                                           &numActiveEndpoints);

        // Update active configuration, interface and alt. settings in UDC handler structure
        pUdcHandle->usbConfigNumActive = configuration;
        pUdcHandle->usbInterfaceNumActive = interface;
        pUdcHandle->usbIntAltSettingsNumActive = settings;
    }
}


/******************************************************************************

  Function Name:
    XllpUdcHWSetup

  Description:
    This function initilizes UDC Controller's hardware.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None


  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T XllpUdcHWSetup (P_XLLP_UDC_T pUdcHandle)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    XLLP_STATUS_T error = 0;

    // Enable USB Client Unit Clock in Clock-Enable Register, CKEN [11] = 1.
    //pClkMgrRegs->CKEN |= XLLP_CLK_MGR_REGS_CKEN_UDC_EN;
	//error = pClkMgrRegs->CKEN;
	//error |= XLLP_CLK_MGR_REGS_CKEN_UDC_EN;
	//pClkMgrRegs->CKEN = error;

    // Program the Endpoint Configuration Registers.
    error = XllpUdcConfigureEndpoints (pUdcHandle);
    if (!error)
    {

        // Enables Reset, Resume, and Configuration change interrupts by
        // setting the corresponding bits to one in the UDC Interrupt Control Register UDCICR1,
        // and disables Start of the frame  and suspend interrupts.
        // pRegs->UDCICR1 = XLLP_UDC_UDCICR1_EVENTS & ~XLLP_UDC_UDCICR1_IESOF;
        pRegs->UDCICR1 = (XLLP_UDC_UDCICR1_IERS | XLLP_UDC_UDCICR1_IERU | XLLP_UDC_UDCICR1_IECC) &
                          ~XLLP_UDC_UDCICR1_IESOF & ~XLLP_UDC_UDCICR1_IESU;


        // Enable the Endpoint 0 interrupt by setting IE0[0] and IE0[1] in the UDC
        // Interrupt Control Register UDCICR0, so it will respond to the USB Host requests.
        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_0,
                                PACKET_COMPL_INT);

        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_0,
                                FIFO_ERROR_INT);

		// Enable the Endpoint A interrupt by setting IEA[0] and IEA[1] in the UDC
		// Interrupt Control Register UDCICR0, so it will respond to the USB Host requests.
        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_A,
                                PACKET_COMPL_INT);

        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_A,
                                FIFO_ERROR_INT);

		// Enable the Endpoint B interrupt by setting IEB[0] and IEB[1] in the UDC
		// Interrupt Control Register UDCICR0, so it will respond to the USB Host requests.
        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_B,
                                PACKET_COMPL_INT);

        XllpUdcEnableInterrupt (pUdcHandle,
                                INT_ENDPOINT_B,
                                FIFO_ERROR_INT);

		pRegs->UDCISR0 = 0xffffffff;
		pRegs->UDCISR1 = 0xffffffff;

        // Enable the first level UDC interrupt.
 		error = EnablePeripheralIRQInterrupt(USB_CLIENT_INT);
        // Set the setup complete flag in the UDCs handler structure to signal
        // that hardware setup was successful.
        pUdcHandle->setupComplete = XLLP_TRUE;
    }

    return error;
}


/******************************************************************************

  Function Name:
    XllpUdcHWShutdown

  Description:
    This function will shutdown the UDC Controller's hardware and clears the UDC
    handle's variables.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T XllpUdcHWShutdown (P_XLLP_UDC_T pUdcHandle)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    XLLP_STATUS_T error = 0;

    // Disable USB Client Unit Clock in Clock-Enable Register, CKEN [11] = 0.
    //pClkMgrRegs->CKEN &= XLLP_CLK_MGR_REGS_CKEN_UDC_DIS;
    //pClkMgrRegs->CKEN &= ~XLLP_CLK_MGR_REGS_CKEN_UDC_EN;
	//error = pClkMgrRegs->CKEN;
	//error &= ~XLLP_CLK_MGR_REGS_CKEN_UDC_EN;
	//pClkMgrRegs->CKEN = error;

    // Disable the interrupt requests from endpoints A - X, and all Event interrupts
    pRegs->UDCICR0 &= ~XLLP_UDC_UDCICR0_ENABLE_ALL;
    pRegs->UDCICR1 &= ~(XLLP_UDC_UDCICR1_EVENTS | XLLP_UDC_UDCICR1_ENABLE_ALL);

	//make sure all interrupts are cleared
	pRegs->UDCISR0 |= pRegs->UDCISR0;
	pRegs->UDCISR1 |= 0xF800FFFF;

    // Disable first level UDC interrupt
    error = DisablePeripheralIRQInterrupt(USB_CLIENT_INT);

	// Unregister the UDC interrupt handler
    // error = XllpUdcUnRegisterFirstLevelInterrupt(); // Wacked to make room for ROM

    // Clear the rest of the UDC handles variables
	pUdcHandle->setupComplete = XLLP_FALSE;
    pUdcHandle->enumerateComplete = XLLP_FALSE;
    pUdcHandle->cableAttached = XLLP_FALSE;
    pUdcHandle->enableDma = XLLP_FALSE;

    pUdcHandle->usbConfigNumActive = 0;
    pUdcHandle->usbInterfaceNumActive = 0;
    pUdcHandle->usbIntAltSettingsNumActive = 0;

    return error;
}

/******************************************************************************

  Function Name:
    XllpUdcSWInit

  Description:
    This function initializes the handle's variables that will be used by the
    UDC driver and no hardware will be accessed it this point.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
	P_XLLP_UDC_REGISTERS_T pRegs - a pointer to the UDC's registers base

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpUdcSWInit (P_XLLP_UDC_T pUdcHandle, P_XLLP_UDC_REGISTERS_T pRegs)
{
	// XLLP_UINT32_T   endpointConfigRegValue, i = 1;

    // Set the UDC handles pointer to the UDC Registers
    // with the base address of the UDC Registers.
    pUdcHandle->pRegsBase = (P_XLLP_UDC_REGISTERS_T)pRegs;

    // Sets the UDC handles pointer to the Endpoints Configuration Table
    // with the pointer to the default Endpoints Configuration Table.
    pUdcHandle->pConfigTable = (P_XLLP_UDC_EP_CONFIG_TABLE_T)defaultUdcEndpointConfigTable;

    // Compute the values that would be loaded to the Endpoints Configuration
    // Registers UDCCRA-UDCCRX, from the data in the Endpoints Configuration Table.
//    while (pUdcHandle->pConfigTable[i].usbEndpointNum != 0)
//    {
//        XllpUdcComputeConfigRegisterValue (&pUdcHandle->pConfigTable[i],
//                                           &endpointConfigRegValue);
//        ++i;
//    }

    // Clear the Interrupt Statistics structure
    memset (&pUdcHandle->interruptStat, 0, sizeof(XLLP_UDC_INT_STATISTICS_T));

    // Clear the Endpoints transfer structures
    memset (&pUdcHandle->EpXferTable[0], 0, (sizeof(XLLP_UDC_XFER_T) * XLLP_UDC_MAX_EP_NUM));

    // Clear the Endpoint 0 control structure
    memset (&pUdcHandle->controlXfer, 0, sizeof(XLLP_UDC_USB_CTRL_XFER_T));

    // Clear the Vendor requests control structure
    memset (&pUdcHandle->vendorReq, 0, sizeof(XLLP_UDC_VENDOR_REQ_DATA_T));

    // Clear the List of Active Endpoints
    memset (&pUdcHandle->listOfActiveEndpoints, 0,
                         (sizeof(XLLP_UDC_USB_BIND_ENDPOINTS_T) * XLLP_UDC_USB_MAX_EP_NUM));

    // Clear the rest of the UDC handles variables
	pUdcHandle->setupComplete = XLLP_FALSE;
    pUdcHandle->enumerateComplete = XLLP_FALSE;
    pUdcHandle->cableAttached = XLLP_FALSE;
    pUdcHandle->enableDma = XLLP_FALSE;

    pUdcHandle->usbConfigNumActive = 0;
    pUdcHandle->usbInterfaceNumActive = 0;
    pUdcHandle->usbIntAltSettingsNumActive = 0;

    // Initialize Endpoint 0 state machine to the Idle state
    pUdcHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
}
