/******************************************************************************
**  COPYRIGHT  2007 Marvell Inernational Ltd.
**  All Rights Reserved
******************************************************************************/
/******************************************************************************
**
**  COPYRIGHT (C) 2002, 2003 Intel Corporation.
**
**  This software as well as the software described in it is furnished under
**  license and may only be used or copied in accordance with the terms of the
**  license. The information in this file is furnished for informational use
**  only, is subject to change without notice, and should not be construed as
**  a commitment by Intel Corporation. Intel Corporation assumes no
**  responsibility or liability for any errors or inaccuracies that may appear
**  in this document or any software that may be provided in association with
**  this document.
**  Except as permitted by such license, no part of this document may be
**  reproduced, stored in a retrieval system, or transmitted in any form or by
**  any means without the express written consent of Intel Corporation.
**
**  FILENAME:   Flash.c
**
**  PURPOSE:    Contain template OEM boot code flash operations
**
******************************************************************************/

#include "typedef.h"
#include "FM.h"
#include "BBM.h"
#include "errors.h"
#include "flash.h"
#include "misc.h"
#include "tim.h"


//Currently, the largest page size we'll see is 4KB (1K words)
#define MAX_PAGE_SIZE_WORDS 0x400
#define MAX_TIM_SIZE_BYTES MAX_TIM_SIZE

//allocated space for a bad block table and the partition table
static __attribute__((aligned(4)))UINT_T BBT_SPACE[MAX_PAGE_SIZE_WORDS];
//static __attribute__((aligned(4)))UINT_T MPT_SPACE[MAX_PAGE_SIZE_WORDS];
static FMProperties_T FMProperties;

P_FMProperties_T GetFMProperties()
{
    return &FMProperties;
}

void FindPT()
{
    UINT_T Retval;
}

UINT_T RelocateBlock(UINT_T Block, P_ReloTable_T pBBT, FlashBootType_T fbt)
{
    UINT16_T BlockRelocatedTo;
    UINT_T num_relo = *pBBT->pNumReloc;
    P_FMProperties_T pFMProp = GetFMProperties();
    P_FlashProperties_T pFlashProp = GetFlashProperties(fbt);

    //block 0 is guaranteed by manufacture: DO NOT RELOCATE IT
    if(Block == 0)
        return 0;

    SetBBTState(BBT_CHANGED, fbt);

    //find next valid block
    if (num_relo == 0)
        BlockRelocatedTo = (UINT16_T) (pFlashProp->NumBlocks - 1);
    else
        BlockRelocatedTo = pBBT->pRelo[num_relo - 1].To - 1; // Last Entry left off

    // Relocate it
    pBBT->pRelo[num_relo].To = BlockRelocatedTo;
    pBBT->pRelo[num_relo].From = Block;
    (*pBBT->pNumReloc)++;
#ifdef FBW_CODE
    borqs_printf2("bad block:", *pBBT->pNumReloc - 1, 2);
    borqs_printf2("from:", Block, 4);
    borqs_printf2("to:", BlockRelocatedTo, 4);
#endif
    return BlockRelocatedTo;
}

void SetBBTState(ReloState_T newState, FlashBootType_T fbt)
{
    UINT_T BlkSize, PageSize;
    P_FMProperties_T pFMProps = GetFMProperties();

    // Now set some values that will work
    PageSize = GetPageSize(fbt);
    BlkSize = GetBlockSize(fbt);

    if ( newState == BBT_ERASED )
       pFMProps->BBT_Slot = BlkSize/PageSize;

    pFMProps->BBT_State = newState;
}

UINT_T UpdateBBT(FlashBootType_T fbt)
{
    UINT_T Retval = NoError;
    UINT_T BlkSize, PageSize, NumPages, TIMpagesUsed;
    P_FMProperties_T pFMProps = GetFMProperties();

    P_ReloTable_T pBBT = &pFMProps->BBT;

    //If this flash device does not use BBM, just return
    if(GetFlashProperties(fbt)->UseBBM == 0)
        return NoError;

    //update BBT (only if there are changes)
    if(pFMProps->BBT_State == BBT_UNCHANGED)
        return NoError;

    //update BBT
    if(pFMProps->BBT_State == BBT_INVALID)
        return BBTReadError;

    if ( fbt == BOOT_FLASH )
    {
        // Now set some values that will work
        PageSize = GetPageSize(fbt);
        BlkSize = GetBlockSize(fbt);


        // ASSUMES ONLY TIM and BBT are located in block 0

        //Find and read first page after the TIM
        TIMpagesUsed = PageSize >= MAX_TIM_SIZE_BYTES ? 1 : MAX_TIM_SIZE_BYTES / PageSize;

        NumPages = BlkSize / PageSize;

        //Did we run out of BBT Slots/
        if ( TIMpagesUsed >= (pFMProps->BBT_Slot - 1))
        {
            //save data in lower portion of block 0
            Retval = ReadFlash(0, TEMP_BUFFER_AREA, MAX_TIM_SIZE_BYTES, fbt);
            //Retval = ReadFlash(0, BBT_SPACE, saved_data, fbt);
            if(Retval != NoError)
                return Retval;

            //erase all of block 0
            Retval = EraseFlash(0, BlkSize, fbt);
            if(Retval != NoError)
                return Retval;

            //restore TIM data in lower portion of block 0
            Retval = WriteFlash(0, TEMP_BUFFER_AREA, MAX_TIM_SIZE_BYTES, fbt);
            if(Retval != NoError)
                return Retval;

            pFMProps->BBT_Slot = NumPages;//BBT will go in the last page of blk 0
        }

        pFMProps->BBT_Slot--;//decrement slot number to point to open slot (not the currently used slot)
        Retval = WriteFlash(pFMProps->BBT_Slot * PageSize, (UINT_T)BBT_SPACE, PageSize, fbt);
        if(Retval != NoError)
            return Retval;
    }

    //shutdown DFC? No
    //clear out NANDProp->BBT in struct so it won't be written again
    SetBBTState(BBT_UNCHANGED, fbt);

    return Retval;
}


/* ScanBBT
 *  This function will take in a Block Number and return with the relocated
 *   block number.  If the block is not relocated, then it will simply return
 *   with the original block number
 *
 *  Inputs:
 *      BlockNum: The block number to look up in the relocation table
 *  Outputs:
 *      None
 *  Retval value: Relocated block number
 */
UINT_T ScanBBT(UINT_T BlockNum)
{
    P_ReloTable_T pBBT = &(GetFMProperties()->BBT);
    UINT_T i;

    //if no BBT was loaded, do search for a replacement
    if(GetFMProperties()->BBT_State == BBT_INVALID)
        return BlockNum;

    //block 0 is guaranteed by manufacture
    if(BlockNum == 0)
        return 0;

    // Has this block been relocated?   If so we'll need to use the relocated block
    for (i = 0; i < *(pBBT->pNumReloc); i++)
    {
        if (pBBT->pRelo[i].From == BlockNum)
        {
            BlockNum = pBBT->pRelo[i].To;
            i = 0;  // Start over in case remapped block is now bad
        }
    }
    return BlockNum;
}

/* FindBBT
 *  This function searchs block 0 for the Bad Block Table (BBT).
 *  In order to handle the multiple schemes, the search thusly:
 *      Check the page after the TIM
 *      if found, binary search upwards for latest update
 *      else NOT found, check the last page in block 0
 *          if found, binary search downwards for latest update
 *          else no BBT found
 *
 * Inputs:
 *      Flash Boot Type
 * Outputs:
 *      None
 *
 *  NOTE:  This function will update the Flash Management Properties structure
 *          BBT will be loaded into global space
 *          BBT_State will be set accordingly
 *          BBT_Slot will be found
 */
UINT_T FindBBT(FlashBootType_T fbt)
{
    UINT_T Retval = NoError;
    UINT_T Slot, StartSlot, EndSlot;
    UINT_T PageSize, BlkSize;
    P_FMProperties_T FMProp = GetFMProperties();

    //Initialize FMProperties structure:
    SetBBTState(BBT_INVALID, fbt);
    FMProp->BBT.pHeader = (USHORT*)BBT_SPACE;
    FMProp->BBT.pNumReloc = (USHORT*)((UINT_T)BBT_SPACE + 0x2);
    FMProp->BBT.pRelo = (P_ReloPair_T)((UINT_T)BBT_SPACE + 0x4);

    //If this flash device does not use BBM, just return
    if(GetFlashProperties(fbt)->UseBBM == 0)
        return NoError;

    // Now set some values that will work
    PageSize = GetPageSize(fbt);
    BlkSize = GetBlockSize(fbt);

    //Find and read first page after the TIM
    Slot = PageSize >= MAX_TIM_SIZE_BYTES ? 1 : MAX_TIM_SIZE_BYTES / PageSize;
    Retval = ReadFlash(Slot * PageSize, (UINT_T)BBT_SPACE, PageSize, fbt);

    //if we found a valid BBT, start the search from the next page and search upward
    if((Retval == NoError) && (*FMProp->BBT.pHeader == (USHORT)VALID_HEADER))
    {
        StartSlot = Slot + 1; // start with next slot
        EndSlot = (BlkSize / PageSize) - 1; // end with last page in block
    }
    else
    {//else, start the search from the last page and work our way backwards
        StartSlot = (BlkSize / PageSize) - 1;
        EndSlot = Slot + 1;
    }
    //do the binary search
    Retval = BinarySearch(StartSlot * PageSize, EndSlot * PageSize, VALID_HEADER, 0, (UINT_T)BBT_SPACE, &Slot, fbt);

    if(Retval == NoError)
    {
        FMProp->BBT_Slot = Slot;
        SetBBTState(BBT_UNCHANGED, fbt);
    } else
        Retval = BBTReadError;
    return Retval;
}

UINT_T verify_flash_data(UINT_T flash_saddr, UINT_T flash_eaddr, UINT_T buf, UINT_T bytes, UINT_T spare_bytes, UINT_T do_reloc, FlashBootType_T fbt)
{
    UINT_T Retval;
    P_FlashProperties_T pFlashP = GetFlashProperties(fbt);
    ReadFlash_F read;
    UINT_T tmp_buffer;
    UINT_T reloc = 0;
    UINT_T lastbytes;
#define page_size           (2048)
#define page_spare_size     (64)
#define block_size          (64*page_size)
#define read_start_addr     (1*block_size)
#define read_end_addr       (2048*block_size) //waste 1:30
#define tmp_buffer_addr     (0x80200000 - page_size - page_spare_size)
    int i,j,k;

    read = pFlashP->ReadFromFlash;
    tmp_buffer = tmp_buffer_addr;
    lastbytes = bytes & 0x3;
    
    for(i = flash_saddr;i <= flash_eaddr;i += block_size)
    {
        for(j = 0;j < block_size;j += page_size)
        {
            Retval = read(i + j, tmp_buffer, 2048, fbt);
            if(Retval == NoError)
            {
                if(buf == -1)
                {
                    UINT32_T *tp;
                    tp = (UINT32_T *)tmp_buffer_addr;
                    for(k = 0;k < (page_size >> 2);k++)
                    {
                        if(*tp++ != -1)
                        {
                            Retval = ReadError;
                            goto __DO_RELOC;;
                        }
                    }
                }
                else
                {
                    UINT32_T *tp,*sp;
                    UINT32_T bytes_cmp;
                    tp = (UINT32_T *)tmp_buffer_addr;
                    sp = (UINT32_T *)buf;
                    //only support 1 block verify,because i do't want to do bbt reloc function at here
                    //assume buf is 4 bytes aligned
                    if(bytes >= page_size)
                    {
                        bytes_cmp = page_size >> 2;
                        bytes -= page_size;
                    }else bytes_cmp = bytes >> 2;

                    for(k = 0;k < bytes_cmp;k++)
                    {
                        if(*tp++ != *sp++)
                        {
                            Retval = ReadError;
                            //buf -= (page_size+spare_bytes)*(j / page_size); //roll back ,when 1 block,this is not asserted
                            goto __DO_RELOC;
                        }
                    }
                    
                    if(bytes_cmp < (page_size >> 2))
                    {
                        char *ttp,*ssp;
                        ttp = (char*)tp;
                        ssp = (char*)sp;
                        for(k = 0;k < lastbytes;k++)
                        {
                            if(*ttp++ != *ssp++)
                            {
                                Retval = ReadError;
                                //buf -= (page_size+spare_bytes)*(j / page_size); //roll back ,when 1 block,this is not asserted
                                goto __DO_RELOC;
                            }
                        }
                        goto __OK_EXIT;
                    }

                    buf += (page_size+spare_bytes);
                }
            }
            else
            {
                goto __DO_RELOC;
            }
        }
__DO_RELOC:
        if(Retval != NoError)
        {
            if(do_reloc)
            {
                RelocateBlock( (i / block_size), &GetFMProperties()->BBT, fbt );
                reloc = 1;
            }
            else
            {
                break; 
            }
        }
    }
__OK_EXIT:

    return (reloc ? DORELOC:Retval);
}

UINT_T create_bbt(UINT_T start_addr, UINT_T end_addr, FlashBootType_T fbt)
{
    UINT_T Retval;
    P_FlashProperties_T pFlashP = GetFlashProperties(fbt);
    UINT_T BlkSize,BlkNum;
    UINT_T flash_addr;
#define page_size           (2048)
#define page_spare_size     (64)
#define block_size          (64*page_size)
#define tmp_buffer_addr     (0x80200000 - page_size - page_spare_size)
#define tmp_spare_buffer_addr (tmp_buffer_addr + page_size)
    UINT_T i,j,max;
    char *bbpos;
    UINT_T reloc_ok = 0;

    bbpos = (char*)(tmp_spare_buffer_addr + 0);
    BlkSize = pFlashP->BlockSize;
    BlkNum  = pFlashP->NumBlocks; 
    i = start_addr / BlkSize;
    max = end_addr / BlkSize;
    if(max >= BlkNum)max = BlkNum - 1;

    for(;i <= max;i++)
    {
        flash_addr = i * BlkSize;
        for(j  = 0;j < 2;j++)
        {
            #if NAND_CODE
            Retval = xdfc_read((UINT_T *)tmp_buffer_addr, flash_addr + j*page_size, page_size, (UINT_T *)tmp_spare_buffer_addr, GetNANDProperties());
            #else
            Retval = LoadSector(flash_addr + j*page_size, tmp_buffer_addr, page_size, fbt, (char*)tmp_spare_buffer_addr);
            #endif
            if(Retval)
            {
#ifdef FBW_CODE
                borqs_printf("data");
                borqs_printfhex((char*)tmp_buffer_addr,16);
                borqs_printf("oob");
                borqs_printfhex((char*)tmp_spare_buffer_addr,32);
#endif
                goto __create_bbt_mark;
            }
            
            if(bbpos[0] != 0xff)goto __create_bbt_mark;
            if(bbpos[1] != 0xff)goto __create_bbt_mark;
        }
        continue;

__create_bbt_mark:
        reloc_ok = 1;
#ifdef FBW_CODE
        borqs_printf("oob");
        borqs_printfhex((char*)tmp_spare_buffer_addr,32);
#endif
        RelocateBlock( i, &GetFMProperties()->BBT, fbt );
    }
    return reloc_ok;
}
/* CreateBBT
 *   If a Bad Block Table already exists, just return with no error
 *
 *   If there is not BBT, create an empty one
 *
 * Inputs:
 *      Flash Boot Type
 * Outputs:
 *      None
 *
 */
extern volatile UINT8_T block0_reset;
UINT_T CreateBBT(FlashBootType_T fbt)
{
    int i;
    UINT_T Retval = NoError;
    UINT_T PageSize, BlkSize;

    P_FMProperties_T FMProp = GetFMProperties();
    P_ReloTable_T pBBT = &(GetFMProperties()->BBT);

    //Initialize FMProperties structure:
    SetBBTState(BBT_INVALID, fbt);
    FMProp->BBT.pHeader = (USHORT*)BBT_SPACE;
    FMProp->BBT.pNumReloc = (USHORT*)((UINT_T)BBT_SPACE + 0x2);
    FMProp->BBT.pRelo = (P_ReloPair_T)((UINT_T)BBT_SPACE + 0x4);

    if ( FindBBT( fbt ) == NoError )
    {
#ifdef FBW_CODE
        borqs_printf("bbt found");
#endif
        if(!block0_reset)
            return Retval;
#ifdef FBW_CODE
        borqs_printf("but bbt will be forced reseted");
#endif
    }

    //If this flash device does not use BBM, just return
    if(GetFlashProperties(fbt)->UseBBM == 0)
        return NoError;

    // Now set some values that will work
    PageSize = GetPageSize(fbt);
    BlkSize = GetBlockSize(fbt);

    //if we found valid BBT, then just return
    if((Retval == NoError) && (*pBBT->pHeader == (USHORT)VALID_HEADER))
     if(!block0_reset)
        return Retval;

    // initialize the FMProps for the new BBT
    SetBBTState(BBT_CHANGED, fbt);
    // start at highest slot as BBT grows downward towards TIM
    FMProp->BBT_Slot = (BlkSize / PageSize);
    FMProp->NumBBTSlots = 0;        // bbt slots used
//    FMProp->pMPT = 0;  don't we need to init the partition table in some way?

    memset(BBT_SPACE, 0xff, sizeof(BBT_SPACE));
    // Create a new header
    *pBBT->pHeader = VALID_HEADER;
    *pBBT->pNumReloc = 0;

    /** Empty BBT has been created **/
    {extern void enable_i_d_cache(void);
        enable_i_d_cache();
    }
#ifdef FBW_CODE
	borqs_printf("bbt creating...");
#endif
    EraseFlash(0, GetFlashProperties(fbt)->NumBlocks*BlkSize, fbt);
    //verify_flash_data(0, (GetFlashProperties(fbt)->NumBlocks-1)*BlkSize, -1, 0, 0, 1, fbt);
    //create_bbt(0, (GetFlashProperties(fbt)->NumBlocks-1)*BlkSize, fbt);
    UpdateBBT(fbt);
    Retval = FindBBT(fbt);
#ifdef FBW_CODE
	borqs_printf("ok");
#endif
    block0_reset = 1;
    {extern void disable_i_d_cache(void);
        disable_i_d_cache();
    }
    return Retval;
}


/* Binary Search Function
 *  This function searchs within a flash block for the "last" page with a matching MagicWord at
 *  an offset into the page as specified by the MagicWordOffset input.  The page is then loaded
 *  into memory at the address specified by the buffer pointer input.  The direction
 *  of the search is determined by the values of the StartAddr and EndAddr inputs.
 *
 *      Example
 *
 * Inputs:
 *      StartAddr:  Absolute address of first possible entry location
 *      EndAddr: Absolute address of the last possible entry location
 *      MagicWord: Value to be checked when searching the block
 *      MagicWordOffset: Word offset into the page at with to check for the MagicWord
 *      fbt: Flash Boot Type that we are to search
 * Outputs:
 *      buffer: Pointer to where the page will be loaded
 *      SlotNum: Page number of the last found page
 *
 * Return Value:
 *      NoError         if page found
 *      SlotNotFound    if no valid page found
 *
 *  NOTE:  buffer MUST be allocated to a full flash Page, or else data corruption will occur
 */
UINT_T BinarySearch(UINT_T StartAddr, UINT_T EndAddr, UINT_T MagicWord, UINT_T MagicWordOffset,
                    UINT_T buffer, UINT_T* SlotNum, FlashBootType_T fbt)
{
    UINT_T Retval;
    UINT_T PageSize, BlkSize, BlkOffset, CurrentPageOffset;
    INT_T MaxSlot, MinSlot, CurrentSlot, LastFoundSlot; //page number inside the block
    UINT8_T MagicWordString[5];
    UINT8_T direction; // 0 = up, 1 = down
    UINT8_T found = 0; // used on each page check: 0 = not found, 1 = found

    // Now set some values that will work
    PageSize = GetPageSize(fbt);
    BlkSize = GetBlockSize(fbt);
    BlkOffset = StartAddr & ~(BlkSize - 1);
    //set this to an invalid number.  At the end, it will tell us if we found a good slot
    LastFoundSlot = -1;
    //put the magic number into a NULL terminated string
    *(UINT_T *)MagicWordString = MagicWord;
    MagicWordString[4] = NULL;

    if(StartAddr > EndAddr)
    {
        MaxSlot = StartAddr / PageSize;
        MinSlot = EndAddr / PageSize;
        direction = 1; //search downward
    }
    else
    {
        MaxSlot = EndAddr / PageSize;
        MinSlot = StartAddr / PageSize;
        direction = 0; //search upward
    }

    //Now we must locate the last used slot.  Do this by searching for the magic number.
    // Start in the middle of the block and do a binary search to locate the last used slot.
    // Note: found variable will be cleared on each iteration
    for(;MinSlot <= MaxSlot; found = 0)
    {
        //adjust current slot to midpoint of max and min to get next offset
        CurrentSlot = (MaxSlot + MinSlot) >> 1;
        CurrentPageOffset = ((UINT_T)CurrentSlot * PageSize) + BlkOffset;

        Retval = ReadFlash(CurrentPageOffset, buffer, PageSize, fbt);

        // Did we find a valid entry?
        if((Retval == NoError) && (((USHORT*)buffer)[MagicWordOffset] == (USHORT)MagicWord))
        {
            found = 1;
            //save the slot number
            LastFoundSlot = CurrentSlot;
        }

        //if: searching upward and DO NOT find a slot -OR- searching downward and find a slot
        //      lower the max slot
        //else: searching upward and find a slot -OR- downward and DO NOT find a slot
        //      raise minimum slot
        if (found == direction)
        //Set new Max slot to the next page below
            MaxSlot = CurrentSlot - 1;
        else
        //Set new Min slot to the next page above
            MinSlot = CurrentSlot + 1;

    } // End while

    //Check to make sure we even found a good slot
    if(LastFoundSlot == -1)
        Retval = NotFoundError;
    else
    {
        *SlotNum = (UINT_T)LastFoundSlot;
        //The last valid slot may get overwritten on last iteration
        //If the current slot is not the same as the last found slot
        // then we need to re-read the last valid entry in the buffer
        if((UINT_T)CurrentSlot != LastFoundSlot)
        {
            CurrentPageOffset = ((UINT_T)LastFoundSlot * PageSize) + BlkOffset;
            Retval = ReadFlash(CurrentPageOffset, buffer, PageSize, fbt);
        }
    }
    return Retval;
}

