/******************************************************************************
**	(C) Copyright 2007 Marvell International Ltd.
**	All Rights Reserved
******************************************************************************/

/******************************************************************************
**
**  FILENAME:
**
**  PURPOSE:
**      ChipIdea 1136-a UBSB 2.0 OTG Client Device Controller transfer routines
**
**
**  LAST MODIFIED:
**
******************************************************************************/

#include "xllp_defs.h"
#include "misc.h"      // memcpy
#include "xllp_ciu2d.h"
#include "protocolmanager.h"
#include "interrupts.h"
#include "mvUsbDevApi.h"
#include "mvUsbDevPrv.h"


extern void EnableIrqInterrupts();
extern void DisableIrqInterrupts();


extern const  XLLP_U2D_EP_T  usbTransmitEp = U2D_ENDPOINT_A;
extern const  XLLP_U2D_EP_T  usbReceiveEp  = U2D_ENDPOINT_B;
extern USB_DEV_STATE_STRUCT_PTR gCI2Handle;

//----------------------------------------------------------------------------
//  CI2Transmit
//
//
//
//----------------------------------------------------------------------------
XLLP_UINT32_T CI2Transmit( XLLP_UINT32_T TxDataLength,
                           XLLP_UINT8_T  *pTxBuff,
                           XLLP_UINT8_T  SyncTransfer      )
{
    UINT8_T Retval;
    XLLP_INT32_T actualRxCount;

    Retval = xllpCI2Xfer( usbTransmitEp,          // Endpoint
                            USB_IN,
                            pTxBuff,                 // Memory buffer
                            TxDataLength,           // Buffer Size
                            &actualRxCount          // Not used on transmit...
                          );

    return (Retval);
}

//----------------------------------------------------------------------------
//  CI2Receive
//
//
//
//----------------------------------------------------------------------------
XLLP_UINT32_T CI2Receive( XLLP_UINT32_T RxDataLength, XLLP_UINT8_T  *pRxBuff, XLLP_INT32_T  *byteCount )
{
    UINT8_T Retval;

    Retval = xllpCI2Xfer( usbReceiveEp,        // Endpoint
                            USB_OUT,
                            pRxBuff,             // Memory buffer
                            RxDataLength,       // Buffer Size
                            byteCount           // # bytes transferred by U2D
                          );


    return (Retval);
}

//--------------------------------------------------------------------------
// CI2GetMsg -> Get bytes/packet etc. from U2D
//              This executes in Interrupt execution context chain
//
// This code has changed from the U2D code, since we don't get
// "short packet" interrupts, therefore more work to keep in line with
// protocol state.
//
// 
//typedef void (*USB_SERVICE_FUNC)(void* handle, uint_8, boolean, uint_8, uint_8_ptr, uint_32, uint_8);
//--------------------------------------------------------------------------

void CI2GetMsg( void *handle, UINT8 ep_num, UINT32_T setup, UINT8 direction,
                        UINT8 *buff_start_address, UINT32 actual_transfer_length, UINT8 errors)
{
  pProtocolCmd pCmd;
  pProtocolISR pISR;
  XLLP_INT32_T byteCount = 0, packetLength = 0;
  UINT8_T Buff[PREAMBLE_SIZE], *buffPtr;

  buffPtr = Buff;
  pISR    = getProtocolISR();
  pCmd    = getProtocolCmd();
    
  //reject calls from Tx callback
  //defensive coding
  //-----------------------------
  if( direction == MV_USB_SEND )
    return;

  // if theres an error
  // reprime the EP and hope
  // the WTPTP host re-sends
  //------------------------
  if( errors )
    goto PRIME_RX;
  // if downloads are disabled then reprime so we do not crash the host 
  // dump any data received in the bit bucket
  if( !(getProtocolISR()->DownloadsEnabled) )
    goto PRIME_RX;
        
  packetLength = actual_transfer_length;

  // Is it the Preamble ??
  //----------------------
  if(packetLength == PREAMBLE_SIZE )
  {
      //Clear the stack buffer
      //----------------------
      memset( buffPtr, 0, sizeof(Buff));

      //Get the bytes from the U2D
      //--------------------------
      CI2Receive( PREAMBLE_SIZE, buffPtr, &byteCount);

      // See if it's the Preamble string
      //--------------------------------
      if( ProtocolVerifyPreamble( buffPtr ) != 0 )
      {
#if !VERBOSE_DEBUG  //leave this on for dumping messages
         ShutdownPort(FFUART_D);
#endif
         ShutdownPort(DIFF_USB_D);
         pISR->PreambleReceived = TRUE;
         pISR->InterruptPort    = CI2_USB_D;

		 ClearPortInterruptFlag();

         // Preamble checks out, so respond
         //--------------------------------
         SendResponse(Buff, PREAMBLE_SIZE);
         goto PRIME_RX;
      }
  }

  //-------------------------------------------------------------------
  // ok, process command stream
  //-------------------------------------------------------------------

  //------------------------------------------
  // Check for bogus transmission, if so then
  // zero the command structure, and let the
  // protocol layer handle it...
  //------------------------------------------

  if( packetLength < 8 )
  {
     CI2Receive( packetLength, (XLLP_UINT8_T *)&(pCmd->Command), &byteCount);
     memset( (XLLP_UINT8_T *)&(pCmd->Command),0,sizeof(ProtocolCmd) );
     return;
  }

  // packet size seems ok ...
  // read constant part of command packet
  //-------------------------------------
  CI2Receive( 8, (XLLP_UINT8_T *)&(pCmd->Command), &byteCount);
  packetLength -= byteCount;

  // update Interrupt Status Flags
  //-----------------------------
  pISR->CurrentCommand  = pCmd->Command;
  pISR->DataLeft        = pCmd->Length;
  pISR->CommandReceived = TRUE;


  switch( pISR->CurrentCommand )
  {
	case MessageCmd:
		 //try and send a message
		 HandleMessageCmd();
		 break;

	case DownloadDataCmd:

         // Next we get the rest of the image that follows
         // the command stucture and put it in the proper
         // image area
         //
         // NOTE: this should not have to loop like this
         //       to complete the DATA transfer... this
         //       is a work-around until I can figure out
         //       if my DMA code is BAD or there's a
         //       Monahans B0 or "P" silicon bug...
         //-----------------------------------------------
         CI2Receive( pISR->DataLeft, (XLLP_UINT8_T *)(pISR->CurrentDownloadAddress), &byteCount);
         pISR->CurrentDownloadAddress += byteCount;
         pISR->DataLeft -= byteCount;
         break;

	case DisconnectCmd:
		 pISR->PreambleReceived = TRUE;
        
    default:

         // rest of messages: should not have over 4 bytes of data
         // this will need to change if protocol changes
         //-------------------------------------------------------
         if(packetLength > 4)
           packetLength = 4;

         if(packetLength)
           CI2Receive( packetLength, (XLLP_UINT8_T *)(&(pCmd->Data[0])), &byteCount);

         break;
  }
  
PRIME_RX:
  // Reset the buffer index for the next Rx
  // and prime the endpoint
  //---------------------------------------
  CI2_ResetBufPosition( usbReceiveEp );
}


//-----------------------------------------
//  ImageU2DInterruptHandler
//
//  Top Level Application interrupt handler
//  for U2D
//
//  We pass through the initial interrupt
//  to the lower-level handler first,
//  then take care of the application next
//
//-----------------------------------------
void ImageCI2InterruptHandler(void)
{

  // This central handler will chain off
  // to 
  //-------------------------------------
  CI2InterruptHandler();

  //allow enumeration.  check to make sure downloads are allowed
  if( ( xllpIsEnumCompleted()) == XLLP_TRUE )
  {
      if( !(getProtocolISR()->DownloadsEnabled) )
	  {
         ClearPortInterruptFlag();
         xllpCI2InterruptDisable();
      }
  }
}

