/******************************************************************************
**	(C) Copyright 2007 Marvell International Ltd. 
**	All Rights Reserved
******************************************************************************/

/******************************************************************************
**
**  COPYRIGHT (C) 2006 Intel Corporation.
**
**  This software as well as the software described in it is furnished under
**  license and may only be used or copied in accordance with the terms of the
**  license. The information in this file is furnished for informational use
**  only, is subject to change without notice, and should not be construed as
**  a commitment by Intel Corporation. Intel Corporation assumes no
**  responsibility or liability for any errors or inaccuracies that may appear
**  in this document or any software that may be provided in association with
**  this document.
**  Except as permitted by such license, no part of this document may be
**  reproduced, stored in a retrieval system, or transmitted in any form or by
**  any means without the express written consent of Intel Corporation.
**
**  FILENAME:       xllp_u2d.c
**
**  PURPOSE:   		This files contains the driver for USB 2.0 Device controller
**                  (U2D).
**
**
******************************************************************************/

/*
*******************************************************************************
*   HEADER FILES
*******************************************************************************
*/


#include "xllp_defs.h"
#include "xllp_usbdefs.h"
#include "xllp_u2d.h"
#include "xllp_u2d_gpio.h"
#include "interrupts.h"
#include "misc.h"
#include "usb_descriptors.h"
														  
#define TAVOR_CHIPTYPE_MASK (0x00080000)

extern OPT_USB_DESCRIPTOR_LOADS USBDescriptors[NUM_USB_DESCRIPTORS];
extern UINT_T	USBStringIndexTable[MAX_USB_STRINGS];
extern const XLLP_U2D_EP_CONFIG_TABLE_T defaultU2DEndpointConfigTable[];
extern UINT_T FindVendorRequestInTIM (XLLP_UINT8_T bRequest, XLLP_UINT8_T bmRequestType, XLLP_UINT32_T *pTxBuffEp0, XLLP_INT32_T *ppacketLength);

//-------------------------------------------------------------------------------
// Private function prototypes
//-------------------------------------------------------------------------------
void XllpU2DWriteDwordFifo(P_XLLP_VUINT32_T pReg, void *src, XLLP_INT32_T len);
void XllpU2DReadDwordFifo (P_XLLP_VUINT32_T pReg, void *buf, XLLP_INT32_T len);
void XllpU2DBuildListOfActiveEndpoints (P_XLLP_U2D_T pUdcHandle,
                                        XLLP_UINT32_T configuration,
                                        XLLP_UINT32_T interface,
                                        XLLP_UINT32_T settings,
                                        P_XLLP_UINT32_T pNumActiveEndpoints);
void XllpU2DControlProcessIdle (P_XLLP_U2D_T pUdcHandle,
                                P_XLLP_UINT32_T pControlXferStatus);
void XllpU2DControlProcessOutData (P_XLLP_U2D_T pUdcHandle,
                                   P_XLLP_UINT32_T pControlXferStatus);
void XllpU2DControlProcessInData (P_XLLP_U2D_T pUdcHandle,
                                  P_XLLP_UINT32_T pControlXferStatus);
void XllpU2DControlProcessEndXfer (P_XLLP_U2D_T pUdcHandle,
                                   P_XLLP_UINT32_T pControlXferStatus);


//------------------------------------------------------------------------------
//  U2DTxEndPtZeroFifo()
//
//       Sets the appropriate bits in the Endpoint 0 control/status register
//       to Transmit the data to the USB host for an IN control packet after
//       having already loaded the EP 0 FIFO
//
//------------------------------------------------------------------------------
void U2DTxEndPtZeroFifo( P_XLLP_U2D_T pU2DHandle, XLLP_UINT8_T packetLength )
{
  P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

  // Set U2DCSR0.IPR = 1 regardless of packet size, you must also
  // simultaneously set IPA if the "trailing" byte count == 3
  //------------------------------------------------------------------------
  if( (packetLength & 0x03) == 3)
  {
     pRegs->U2DCSR_X[U2D_ENDPOINT_0] = (XLLP_U2DCSR0_IPR | XLLP_U2DCSR0_IPA);
  }
  else
     pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_IPR;
}


//-------------------------------------------------------------------------------
// XllpU2DComputeRegMask()
//         This is used to get around the complier complaining about shifts
//         on config values stored in the struct that are less than 32 bits...
//-------------------------------------------------------------------------------
XLLP_UINT32_T XllpU2DComputeRegMask( XLLP_UINT32_T val, XLLP_UINT32_T leftShift )
{
  return (val << leftShift);
}

/******************************************************************************

  Function Name:
    XllpU2DComputeConfigRegisterValue

  Description:
    This function computes the configuration values that will be loaded to
    the Endpoints Configuration Registers:

      U2DCR_X,(Fifo config reg) and
      U2DNE_X (info reg)

    based on the entries in the Endpoints Configuration table. It stores the
    configuration values in the Endpoints Configuration table. It also returns
    the configuration register's value.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_U2D_EP_CONFIG_TABLE_T pEpCfgTbl - a pointer to the Endpoints
                                             Configuration Table.
  Output Arguments:
    P_XLLP_UINT32_T pU2DEpInfoRegVal
                    pU2DEpConfigRegVal
  Return Value:
    None

*******************************************************************************/
void XllpU2DComputeConfigRegisterValue (P_XLLP_U2D_EP_CONFIG_TABLE_T pEpCfgTbl,
                                        P_XLLP_UINT32_T              pU2DEpInfoRegVal,
                                        P_XLLP_UINT32_T              pU2DEpConfigRegVal)
{
  XLLP_UINT32_T tmpMask, numBytes;

  // Compute the Info Register value (U2DENA->U2DENG)
  //-------------------------------------------------
  // Usb Configuration Number: (1-3)
  tmpMask = XllpU2DComputeRegMask( pEpCfgTbl->usbConfigNum, XLLP_U2DNEX_CN_SHIFT);
  // Usb Interface Number: (1-7)
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->usbInterfaceNum, XLLP_U2DNEX_IN_SHIFT);
  // Usb Interface Alternate Settings Number: (1-7)
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->usbIntAltSettingsNum, XLLP_U2DNEX_AISN_SHIFT);
  // Usb Endpoint Number: (1-15)
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->usbEndpointNum, XLLP_U2DNEX_EN_SHIFT);
  // Usb Endpoint type: Bulk, Iso, Interrupt
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->usbEndpointType, XLLP_U2DNEX_ET_SHIFT);
  // Usb Endpoint direction: IN, OUT
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->usbEndpointDirection, XLLP_U2DNEX_ED_SHIFT);
  // Max. Packet Size: (1-1023)
  tmpMask |= XllpU2DComputeRegMask( pEpCfgTbl->maxPacketSize, XLLP_U2DNEX_MPS_SHIFT);

  *pU2DEpInfoRegVal = tmpMask;

  // Compute the Configuration Register Value (U2DCRA->U2DCRG)
  // NOTE: the fifo size written to the register should be:
  // (desired fifo size)/8;  Fifo sizes must be 8 byte aligned,
  // and include another 8 bytes for each packet delimiter, we
  // check for correlation of the buffer size and max packet size
  // and add the required packet delimiters appropriately...
  // we also check that the fifo size MUST be at least as large
  // as the max packet size...
  //----------------------------------------------------------
  if( (pEpCfgTbl->dmaFifoSize < pEpCfgTbl->maxPacketSize) ||
      (pEpCfgTbl->dmaFifoSize == 0)                           )
  {
    U2DLogError( U2D_ERR_FIFO_SIZE_TOO_SMALL );
    *pU2DEpConfigRegVal = 0;
    return;
  }

  // Compute the number of 8 byte delimiters needed,
  // add the fifo size, and then up the value to the
  // next 8 byte alignment; each packet in the FIFO
  // requires an 8 byte delimiter
  //------------------------------------------------
  numBytes =   pEpCfgTbl->dmaFifoSize / pEpCfgTbl->maxPacketSize;
  numBytes += (pEpCfgTbl->dmaFifoSize % pEpCfgTbl->maxPacketSize) ? 1:0;
  numBytes <<= 3;
  numBytes += pEpCfgTbl->dmaFifoSize;

  // Set to next 8 byte alignment if required
  //-----------------------------------------
  if( numBytes & 0x07 )
    numBytes += (8 - (numBytes & 0x07));

  // Register val is stored as FifoSize/8
  //-------------------------------------
  numBytes >>= 3;

  tmpMask  = XllpU2DComputeRegMask( numBytes, XLLP_U2DCRX_BS_SHIFT);
  tmpMask |= XLLP_U2DCRX_EE;

  *pU2DEpConfigRegVal = tmpMask;
}

/******************************************************************************

  Function Name:
    XllpU2DConfigureEndpoints

  Description:
    This function configures and enables all of the Endpoints that are
    used for each of the USB Configurations and Interfaces.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    None

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T  XllpU2DConfigureEndpoints(P_XLLP_U2D_T pU2DHandle )
{
    P_XLLP_U2D_REGISTERS_T pRegs              = pU2DHandle->pRegsBase;
    P_XLLP_U2D_EP_CONFIG_TABLE_T pConfigTable = pU2DHandle->pConfigTable;

    XLLP_UINT32_T j, i;
    XLLP_STATUS_T error = 0;
    XLLP_UINT32_T endpointConfigRegValue, endpointInfoRegValue;
	XLLP_UINT32_T values[16];

    // Program the Endpoint Configuration Registers:
    // Disable U2D controller by writing zero to the UDE bit (U2DCR.UDE = 0),
    // that enables the Endpoint Configuration Registers for write operation.
    // Do not clear the U2DCR.CC bit here
    //-----------------------------------------------------------------------
    pRegs->U2DCR = 0;

    // Endpoint zero is not configurable, so start with next EP in table.
    // End of table is then signified by finding a "0" USB endpoint
    // Note that we examine the USB endpoint in the record and not udcEndpoint
    //------------------------------------------------------------------------
    i=1;
    while( (pConfigTable[i].usbEndpointNum) != 0)
    {
       // Compute the values that would be loaded to the Endpoints Configuration
       // Registers U2DCRA-U2DCRX, and Info Registers U2DNE_X
       // from the data in the Endpoints Configuration Table.
       //-----------------------------------------------------------------------------
       XllpU2DComputeConfigRegisterValue (&pU2DHandle->pConfigTable[i],
                                          &endpointInfoRegValue,
                                          &endpointConfigRegValue );

       // Write the configuration value to the Endpoints
       // Configuration Registers U2DCRA-U2DCRX
       // write the Config register first before setting
       // the info reg. per Monahans U2D spec
       //-----------------------------------------------
       j = pConfigTable[i].udcEndpointNum;
       pRegs->U2DCR_X[j] = endpointConfigRegValue;  // this reg can only be written while UDE == 0
	   values[i] = endpointInfoRegValue;
       ++i;
    }

    // Enable U2D controller by writing one to the bit UDE (U2DCR.UDE = 1).
    // This disables the Endpoint Configuration Registers for write operations
    // and enables the Endpoint Info Registers for write.
    //------------------------------------------------------------------------
    pRegs->U2DCR |= XLLP_U2DCR_SMAC;
    Delay(1000);                     // 1000 uS

    pRegs->U2DCR = (XLLP_U2DCR_SMAC | XLLP_U2DCR_UDE);
    Delay(2500);  // 2500us -> Cypress UTMI needs 1.9ms for reset; Unknown how long it may take EMCE to be set...

    // Check if the Endpoint Memory Configuration is valid by reading the status of
    // the Endpoint Memory Configuration Error U2DCR.EMCE bit.
    //-----------------------------------------------------------------------------
    if((pRegs->U2DCR & XLLP_U2DCR_EMCE) || ((pRegs->U2DCR & XLLP_U2DCR_UDE) != XLLP_U2DCR_UDE))
    {
       // This bit is set if the endpoint memory configuration has an error.
       // The driver must clear U2DCR.EMCE bit by writing one to it before again enabling the U2D.
       //-----------------------------------------------------------------------------------------
       pRegs->U2DCR |= XLLP_U2DCR_EMCE;
       U2DLogError( U2D_ERR_ENDPOINT_CONFIG_ERROR );
       error = XLLP_TRUE;
    }
    else
    {

      // U2D bug handling, TIBET bug # 2000803
      // https://tibet.intel.com/WCCGPCGProcessors/Default.asp?Focus=Bug&ID=2000803
      //
      // Set all INFO register CONFIGURATION # fields to a NON-ZERO value so that
      // U2D internal logic does not confuse endpoints ... make sure to use an un-used
      // config number and endpoint for your application... I use 0x0F here for
      // the CONFIG number and set the endpoint to a non-zero value (loop counter) since
      // we can't have duplicate endpoints in the same config...
      //---------------------------------------------------------------------------------
      for( i = U2D_ENDPOINT_A; i < U2D_ENDPOINT_BAD; i++ )
        pRegs->U2DNE_X[i] = (0x0F << XLLP_U2DNEX_CN_SHIFT) | (0x0F - i);

      // Endpoint zero is not configurable, so start with next EP in table
      //------------------------------------------------------------------
      i = 1;
      while (pConfigTable[i].usbEndpointNum != 0)
      {
         // Write the Info value to the Endpoints
         // Info Registers U2DNE_X can only be written
         // when UDE == 1
         //-----------------------------------------------
         j = pConfigTable[i].udcEndpointNum;
         pRegs->U2DNE_X[j] = values[i];
         ++i;
      }

      // Setting UDR brings the UTMI OUT of SUSPEND
      //---------------------------------------------
      pRegs->U2DCR |= XLLP_U2DCR_UDR ;
    }

    return error;
}

/******************************************************************************

  Function Name:
    XllpU2DEnableInterrupt

  Description:
    This function enables the Endpoints or Events interrupt.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle
    XLLP_U2D_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType - SP, PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpU2DEnableInterrupt( P_XLLP_U2D_T pU2DHandle,
                             XLLP_U2D_INTERRUPTS_T udcEndpointNum,
                             XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T mask;
    volatile P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

    // Set the corresponding bit in the U2D's interrupt control register
    // U2DICR to enable the endpoint's and event's interrupt requests.
    //-------------------------------------------------------------------------
    if (udcEndpointNum <= U2D_INT_ENDPOINT_G)
    {
        mask = (0x1U << interruptType);
        mask = mask << (3*udcEndpointNum);
   	    pRegs->U2DICR |= mask;
 	}
    else if (udcEndpointNum >= U2D_INT_DATA_PACKET_ERROR)
    {
        mask = (0x1U << udcEndpointNum);
   	    pRegs->U2DICR |= mask;
    }
}

/******************************************************************************

  Function Name:
    XllpU2DDisableInterrupt

  Description:
    This function disables the Endpoints or Events interrupts.

  Global Register Modified:
    None

   Input Arguments:
     P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle
     XLLP_U2D_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType - SP, PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/
void XllpU2DDisableInterrupt( P_XLLP_U2D_T                 pU2DHandle,
                              XLLP_U2D_INTERRUPTS_T        udcEndpointNum,
                              XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T mask;
    volatile P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

    // Clear the corresponding bit in the U2D's interrupt control register
    //  U2DICR to disable the endpoint's and event's interrupt requests.
    //--------------------------------------------------------------------
    if (udcEndpointNum <= U2D_INT_ENDPOINT_G)
    {
        mask = (0x1U << interruptType);
        mask = mask << (3*udcEndpointNum);
        pRegs->U2DICR &= ~mask;
 	}
    else if (udcEndpointNum >= U2D_INT_DATA_PACKET_ERROR )
    {
        mask = (0x1U << udcEndpointNum);
        pRegs->U2DICR &= ~mask;
    }
}

/******************************************************************************

  Function Name:
    XllpU2DClearInterrupt

  Description:
    This function clears the Endpoints or Events interrupts.
    It also clears the Endpoint FIFO Error bit in the Endpoint's U2DCSR register.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle
    XLLP_U2D_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpU2DClearInterrupt (P_XLLP_U2D_T                 pU2DHandle,
                            XLLP_U2D_INTERRUPTS_T        udcEndpointNum,
                            XLLP_U2D_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T mask;
    volatile P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

    // Write one to the corresponding bit of the U2D interrupt status register
    // U2DISR to clear the endpoint's and event's interrupt. writing 0 to any
    // bit has no effect on the register
    //------------------------------------------------------------------------
    if (udcEndpointNum <= U2D_INT_ENDPOINT_G)
    {
        mask = (0x1U << interruptType);
        mask = mask << (3*udcEndpointNum);
        pRegs->U2DISR = mask;
 	}
    else if (udcEndpointNum >= U2D_INT_DATA_PACKET_ERROR)
    {
        mask = (0x1U << udcEndpointNum);
        pRegs->U2DISR = mask;
    }
}



/*------------------------------------------------------------------------------

  Function Name:
    XllpU2DWriteDwordFifo

  Description:
    This function writes the specified number of bytes to the Endpoint 0's FIFO.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_VUINT32_T pReg   - a pointer to the Endpoints FIFO
    void *src               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to write

  Output Arguments:
    None

  Return Value:
    None

------------------------------------------------------------------------------*/
void XllpU2DWriteDwordFifo(P_XLLP_VUINT32_T pReg, void *src, XLLP_INT32_T len)
{
	XLLP_INT32_T    i;
	XLLP_UINT32_T	temp[18];   // MPS for U2D Ep0 = 64 bytes, so this should suffice...
	P_XLLP_UINT32_T p1;

	if (len == 0)
		return;

	if( (len > sizeof(temp)) && ((XLLP_UINT32_T)src & 0x3) )
	{
		// ERROR alignment, return
		return;
	}

	if ((XLLP_UINT32_T) src & 0x3)
	{
		memcpy (temp, src, len);
		p1 = temp;
	}
	else
	{
		p1 = src;
	}

	if (((XLLP_INT32_T) len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++)
		{
			*pReg = *p1++;
	    }
	}
	else
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*pReg = *p1++;
	    }

        // For any "trailing" bytes with the U2D you
        // are allowed only 1 write operation, otherwise
        // a FIFO/Transmit error will occur.
        // Remember to set IPA in the U2DSCR0 for
        // trailing writes of "3" bytes
        //-----------------------------------------
        switch( len & 0x3 )
        {
            case 1:
              *(P_XLLP_VUINT8_T)pReg=*(P_XLLP_VUINT8_T)p1;
              break;

            case 2:
              *(P_XLLP_VUINT16_T)pReg=*(P_XLLP_VUINT16_T)p1;
              break;

            case 3:
              *(P_XLLP_VUINT32_T)pReg=*(P_XLLP_VUINT32_T)p1;
              break;

            default:
              ;
        }
#if 0
        if ( (len & 0x3) == 2)
        {
            *(P_XLLP_VUINT16_T)pReg=*(P_XLLP_VUINT16_T)p1;
        }
        if ( (len & 0x3) == 1)
        {
            *(P_XLLP_VUINT8_T)pReg=*(P_XLLP_VUINT8_T)p1;
        }
        if ( (len & 0x3) == 3)
        {
            *(P_XLLP_VUINT32_T)pReg=*(P_XLLP_VUINT32_T)p1;
        }
#endif
	}
}


/******************************************************************************

  Function Name:
    XllpU2DReadDwordFifo

  Description:
    This function reads the specified number of bytes to the Endpoint's FIFO.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_VUINT32_T pReg   - a pointer to the Endpoints FIFO
    void *buf               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to read

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/
void XllpU2DReadDwordFifo( P_XLLP_VUINT32_T pReg, void *buf, XLLP_INT32_T len )
{
	XLLP_INT32_T    i;
    XLLP_UINT32_T   temp[18], rdat;  // MPS for U2D Ep0 = 64 bytes, so this should suffice...
	P_XLLP_UINT32_T p1;
    XLLP_BOOL_T     copyback;
    P_XLLP_VUINT8_T pBuff, pTemp;

	if ( ((XLLP_UINT32_T) buf & 0x3)  && (len > sizeof(temp)) )
	{
		// ERROR alignment, return
		return;
	}

	if( (XLLP_UINT32_T)buf & 0x3 )
	{
		p1 = temp;
		copyback = XLLP_TRUE;
	}
	else
	{
		p1 = buf;
		copyback = XLLP_FALSE;
	}


	if (((XLLP_INT32_T)len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++)
		{
			*p1++ = *pReg;
	    }
	}
	else
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*p1++ = *pReg;
	    }

	    rdat = *pReg;
        pBuff = (P_XLLP_VUINT8_T)p1;
        pTemp = (P_XLLP_VUINT8_T)&rdat;

	    for (i=0; i<(len&0x3); i++)
	    {
           *pBuff++ = *pTemp++;
	    }
	}

	if (copyback)
	{
		memcpy (buf, temp, len);
	}


}


/******************************************************************************

  Function Name:
    XllpU2DBuildListOfActiveEndpoints

  Description:
    This function fills the list of the endpoints that are used in the active
    configuration, active interface, and active alternate interface. This list
    will be processed from the main U2D's interrupt handler to accelerate the
    determination which of the Endpoints A-G generated the interrupt.

  Global Register Modified:
    None

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle
    XLLP_UINT32_T configuration
    XLLP_UINT32_T interface
    XLLP_UINT32_T settings

  Output Arguments:
    P_XLLP_UINT32_T pNumActiveEndpoints - the total number of the endpoints in
                                          the list of the active endpoints will
                                          be returned.
  Return Value:
    None

*******************************************************************************/

void XllpU2DBuildListOfActiveEndpoints (P_XLLP_U2D_T    pU2DHandle,
                                        XLLP_UINT32_T   configuration,
                                        XLLP_UINT32_T   interface,
                                        XLLP_UINT32_T   settings,
                                        P_XLLP_UINT32_T pNumActiveEndpoints)
{
  XLLP_UINT32_T i, j;
  P_XLLP_U2D_EP_CONFIG_TABLE_T pConfigTable = pU2DHandle->pConfigTable;

  // Find all of the Endpoints that supports the active configuration and interface
  // and build the table of active endpoints. Endpoint 0 is not configurable so do not
  // touch it -> start with next endpoint descriptor in table...
  //----------------------------------------------------------------------------------

  i=1; j=0;
  while( pConfigTable[i].usbConfigNum > 0 )
  {
    if (pConfigTable[i].usbConfigNum == configuration)
    {
       if (pConfigTable[i].usbInterfaceNum == interface)
       {
          if (pConfigTable[i].usbIntAltSettingsNum == settings)
          {
             pU2DHandle->listOfActiveEndpoints[j].udcEndpointNum =
                                                           pConfigTable[i].udcEndpointNum;
             pU2DHandle->listOfActiveEndpoints[j].usbEndpointNum =
                                                           pConfigTable[i].usbEndpointNum;

             // If it's an USB HOST OUT endpoint, enable interrupts...
             //-------------------------------------------------------
             if( XllpU2DGetEndPointTransferDirection(pConfigTable[i].udcEndpointNum) == 0)
             {
               // Enable endpoint's packet complete interrupt
               XllpU2DEnableInterrupt( pU2DHandle,
                                        (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                         U2D_PACKET_COMPL_INT );

               // Enable endpoint's FIFO error interrupt
               XllpU2DEnableInterrupt( pU2DHandle,
                                        (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                        U2D_FIFO_ERROR_INT );

               // DCB Enable only for USB HOST OUT endpoints!!!
               XllpU2DEnableInterrupt( pU2DHandle,
                                        (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                         U2D_SHORT_PACKET_INT );
             }
             else
             {
                // Disable endpoint's interrupts

                XllpU2DDisableInterrupt( pU2DHandle,
                                         (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                          U2D_PACKET_COMPL_INT );

                XllpU2DDisableInterrupt( pU2DHandle,
                                          (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                           U2D_FIFO_ERROR_INT );

                XllpU2DDisableInterrupt( pU2DHandle,
                                          (XLLP_U2D_INTERRUPTS_T)pConfigTable[i].udcEndpointNum,
                                           U2D_SHORT_PACKET_INT );
             }

             ++j;
          }
       }
    }

    ++i;
  }

  *pNumActiveEndpoints = j;
}

//-------------------------------------------------------------
// SetDescrSize()
//
//    Removes repeated code sequence
//
//-------------------------------------------------------------
XLLP_UINT32_T SetDescrSize( XLLP_UINT32_T first, XLLP_UINT32_T second )
{
   return (first > second) ? second : first;
}

/******************************************************************************

  Function Name:
    XllpU2DControlProcessIdle

  Description:
    This function is an entry point of Endpoint 0 state machine.
    It processes the setup stage of the setup transaction.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - this argument is not used

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

void XllpU2DControlProcessIdle (P_XLLP_U2D_T    pU2DHandle,
                                P_XLLP_UINT32_T pControlXferStatus)
{
    XLLP_USB_SETUP_DATA_T reqData;
    P_XLLP_UINT32_T rxBuffP = (P_XLLP_UINT32_T)&reqData;
    P_XLLP_UINT32_T pTxBuffEp0;
    XLLP_UINT32_T   stringIndex;
    XLLP_UINT32_T   stringLength;
    XLLP_UINT32_T   tempReqType;
    XLLP_INT32_T    packetLength;
	XLLP_UINT32_T   PlatformType = 0;  // Used to determine if Monahans or Tavor
    P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

    // Determine if OPC and SA both are set.
    // It means that OUT packet in the Endpoint 0 buffer
    // identifying SETUP transaction
    //---------------------------------------------------------------------
    if( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
        (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)                                         )
    {

      // Read data from Endpoint 0 data buffer
      // The data packet of the setup stage is always 8 bytes
      //-----------------------------------------------------
      XllpU2DReadDwordFifo ( &(pRegs->U2DDR0), rxBuffP, 8);

      // Clear U2DCS0.OPC and U2DCS0.SA
      //--------------------------------
      pRegs->U2DCSR_X[U2D_ENDPOINT_0] = (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA);

      // Parse the data that is in the buffer
      // Identify RequestType
      //-------------------------------------
      tempReqType = (reqData.bmRequestType & XLLP_U2D_REQ_TYPE_MASK) >> 5;

      if (tempReqType == STANDARD_REQ)
      {
          PlatformType = *CHIPBOOTTYPE;
		  if ((PlatformType & TAVOR_CHIPTYPE_MASK) == TAVOR_CHIPTYPE_MASK)
		  {
			PlatformType = 1;  // U2D not present on TAVOR ...
		  }
		  else
		  {
			PlatformType = 0;  // Monahans
		  }

		    // Identify Standard Request type
          switch (reqData.bRequest)
          {
              case GET_DESCRIPTOR:

                  // Get the number of data bytes in the data stage that follows
                  // This is the maximum number, and the device can return this
                  // number of bytes or fewer.
                  //-------------------------------------------------------------
                  pU2DHandle->controlXfer.dataLengthEp0 = reqData.wLength;

                  // Identify type of descriptor
                  //-----------------------------
                  switch (reqData.wValue >> 8)
                  {
                      case DEVICE_DESCRIPTOR:

                         if (PlatformType == 0)  // Monahans platform
						 {
                           // Read the length of the device descriptor
                           // and compare with number of bytes in the data stage
                           //---------------------------------------------------
                           pU2DHandle->controlXfer.dataLengthEp0 =
                                 SetDescrSize( pU2DHandle->controlXfer.dataLengthEp0, USBDescriptors[USB_DEVICE_DESCRIPTOR & 0xFF].DesSize);

                           // Initialize the Tx buffer pointer with the device descriptor's address
                           pU2DHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_DEVICE_DESCRIPTOR & 0xFF].pDesPayload;

                          }

                          break;

                      case CONFIG_DESCRIPTOR:

                         pU2DHandle->controlXfer.dataLengthEp0 =
                                SetDescrSize( pU2DHandle->controlXfer.dataLengthEp0, USBDescriptors[USB_CONFIG_DESCRIPTOR & 0xFF].DesSize);

                         // Initialize the Tx buffer pointer with the config. descriptor's address
                         //-----------------------------------------------------------------------
                         pU2DHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_CONFIG_DESCRIPTOR & 0xFF].pDesPayload;

                         break;

                      case STRING_DESCRIPTOR:

						  stringIndex  = ((reqData.wValue) & 0xFF);
                          stringLength = ((reqData.wLength) & 0xFFFF);


							// Bounds protection
							if (stringIndex >= MAX_USB_STRINGS)
							{
								pU2DHandle->controlXfer.dataLengthEp0 =
                                    SetDescrSize( stringLength, USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		    // Initialize the Tx buffer pointer with the config. descriptor's address
                          		pU2DHandle->controlXfer.pTxBuffEp0 =
                                    (P_XLLP_UINT32_T)USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].pDesPayload;
                            }

						    switch (USBStringIndexTable[stringIndex])
							{
								case USB_LANGUAGE_STRING_DESCRIPTOR:

                                    pU2DHandle->controlXfer.dataLengthEp0 =
                                        SetDescrSize( stringLength, USBDescriptors[USB_LANGUAGE_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		        // Initialize the Tx buffer pointer with the config. descriptor's address
                                    //-----------------------------------------------------------------------
                          		    pU2DHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_LANGUAGE_STRING_DESCRIPTOR & 0xFF].pDesPayload;

                          		break;

								case USB_MANUFACTURER_STRING_DESCRIPTOR:

									// EXPERIMENT SENDING ZERO BYTES TO THE HOST
									// TO TEST FOR THE BEHAVIOR IF STRINGLENGTH IS 0
									// MUST DO!!!!!!!!

                                  pU2DHandle->controlXfer.dataLengthEp0 =
                                      SetDescrSize( stringLength, USBDescriptors[USB_MANUFACTURER_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		      // Initialize the Tx buffer pointer with the config. descriptor's address
                          		  pU2DHandle->controlXfer.pTxBuffEp0 = (P_XLLP_UINT32_T)USBDescriptors[USB_MANUFACTURER_STRING_DESCRIPTOR & 0xFF].pDesPayload;

                          		break;

								case USB_PRODUCT_STRING_DESCRIPTOR:

              		               pU2DHandle->controlXfer.dataLengthEp0 =
                                       SetDescrSize( stringLength, USBDescriptors[USB_PRODUCT_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		       // Initialize the Tx buffer pointer with the config. descriptor's address
                          		   pU2DHandle->controlXfer.pTxBuffEp0 =
                                    (P_XLLP_UINT32_T)USBDescriptors[USB_PRODUCT_STRING_DESCRIPTOR & 0xFF].pDesPayload;

								break;

								case USB_SERIAL_STRING_DESCRIPTOR:

                                  pU2DHandle->controlXfer.dataLengthEp0 =
                                      SetDescrSize( stringLength, USBDescriptors[USB_SERIAL_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		      // Initialize the Tx buffer pointer with the config. descriptor's address
                          		  pU2DHandle->controlXfer.pTxBuffEp0 =
                                    (P_XLLP_UINT32_T)USBDescriptors[USB_SERIAL_STRING_DESCRIPTOR & 0xFF].pDesPayload;

                          		break;

								case USB_INTERFACE_STRING_DESCRIPTOR:

                                  pU2DHandle->controlXfer.dataLengthEp0 =
                                      SetDescrSize( stringLength, USBDescriptors[USB_INTERFACE_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		       // Initialize the Tx buffer pointer with the config. descriptor's address
                          	       pU2DHandle->controlXfer.pTxBuffEp0 =
                                    (P_XLLP_UINT32_T)USBDescriptors[USB_INTERFACE_STRING_DESCRIPTOR & 0xFF].pDesPayload;

								break;

								default:

                                  pU2DHandle->controlXfer.dataLengthEp0 =
                                      SetDescrSize( stringLength, USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].DesSize);

                      		      // Initialize the Tx buffer pointer with the config. descriptor's address
                          		  pU2DHandle->controlXfer.pTxBuffEp0 =
                                    (P_XLLP_UINT32_T)USBDescriptors[USB_DEFAULT_STRING_DESCRIPTOR & 0xFF].pDesPayload;

	                         	break;
							}

						break;
                  }

                  //---------------------------------------------------------------
                  // Preload EP0 FIFO to get ready for IN transaction that follows.
                  // Write up to sixteen bytes to the Endpoint 0 data buffer
                  //---------------------------------------------------------------
                  packetLength = pU2DHandle->controlXfer.dataLengthEp0;
                  pTxBuffEp0   = pU2DHandle->controlXfer.pTxBuffEp0;

                  // Write to the Endpoint 0 FIFO
                  //-----------------------------
                  if (packetLength >= U2D_MPS_CNTRL_64)
                  {
                      XllpU2DWriteDwordFifo((&pRegs->U2DDR0), pTxBuffEp0, U2D_MPS_CNTRL_64);
                      pTxBuffEp0 = (P_XLLP_UINT32_T)((XLLP_UINT32_T)pTxBuffEp0 + U2D_MPS_CNTRL_64);
                      pU2DHandle->controlXfer.dataLengthEp0 -= U2D_MPS_CNTRL_64;

                      // Update pointer to the transmit buffer
                      //--------------------------------------
                      pU2DHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;
                  }
                  else
                  {
                      XllpU2DWriteDwordFifo ((&pRegs->U2DDR0), pTxBuffEp0, packetLength);
                      pU2DHandle->controlXfer.dataLengthEp0 -= packetLength;
                  }

                  if (packetLength < U2D_MPS_CNTRL_64)
                  {
                      // Set the status of the Endpoint 0 state machine to XFER data
                      //------------------------------------------------------------
                      pU2DHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
                  }
                  else
                  {
                      // Set the status of the Endpoint 0 state machine to IN data
                      //----------------------------------------------------------
                      pU2DHandle->controlXfer.statusEp0 = EP0_IN_DATA_STATE;
                  }

                  // Send the control packet on its way
                  //-----------------------------------
                  U2DTxEndPtZeroFifo( pU2DHandle, packetLength );

                  break;

               case SET_DESCRIPTOR:

                  // Save descriptor index (lower byte of wValue)
                  pU2DHandle->controlXfer.descIndexEp0 = reqData.wValue & 0xff;
                  // Save descriptor type (upper byte of wValue)
                  pU2DHandle->controlXfer.descTypeEp0 = reqData.wValue >> 8;
                  // Get the number of data bytes in the data stage that follows
                  pU2DHandle->controlXfer.dataLengthEp0 = reqData.wLength;
                  // Initialize Rx buffer pointer with address where the descriptor will be stored
                  pU2DHandle->controlXfer.pRxBuffEp0 = pU2DHandle->controlXfer.outDataEp0;
                  // Set U2DCSR0.IPR in case of premature STATUS IN stage
                  pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_IPR;
                  // Set the status of the Endpoint 0 state machine to OUT data
                  pU2DHandle->controlXfer.statusEp0 = EP0_OUT_DATA_STATE;
          }
      }

      // Process Vendor Request
      else if (tempReqType == VENDOR_REQ)
      {
			// Set this flag to indicate processing a vendor request
            pU2DHandle->vendorReq.vendorReqComplete = XLLP_FALSE;

            // Get ready to process the data that would be sent in the data stage
            pU2DHandle->vendorReq.recipient = reqData.bmRequestType;
            pU2DHandle->vendorReq.vendorReqType = reqData.bRequest;
            pU2DHandle->vendorReq.vendorReqEndpointNum = reqData.wValue;

            // Get the number of data bytes in the data stage that follows
            pU2DHandle->controlXfer.dataLengthEp0 = reqData.wLength;

            // 0: Host to device
            if ( !(reqData.bmRequestType & 0x80) )
            {
				// Set U2DCSR0.IPR in case of premature STATUS IN stage
        		//-----------------------------------------------------
			 	pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_IPR;

				// Get ready to input from the host
                pU2DHandle->controlXfer.statusEp0 = EP0_OUT_DATA_STATE;
            	pU2DHandle->controlXfer.pRxBuffEp0 = pU2DHandle->controlXfer.outDataEp0;

				// Don't allow a buffer overflow
				if (reqData.wLength > 64*4)
					pU2DHandle->controlXfer.dataLengthEp0 = 64*4;
            }
            // 1: Device To Host	Transfer Type
            else
                XllpU2DProcessVendorRequest (pU2DHandle);
      }
      else if (tempReqType == CLASS_REQ)
      {}
    }

    *pControlXferStatus = 0;
}

/******************************************************************************

  Function Name:
    XllpU2DProcessVendorRequest

  Description:
    Vendor Request format:
        bmRequestType: bits 6 and 5 = 10 (vendor defined)
                       bits 4 through 0 = 00010 (endpoint)
        bRequest:      SETUP IN EP   = 0x01
                       SETUP OUT EP  = 0x02
                       SETUP INT EP  = 0x03
                       SETUP LOOPBACK  = 0x04
                       bRequest is passed in requestType

        wValue:        LSB - Rx(OUT) Endpoint number
                       MSB - Tx(IN) Endpoint number
                       wValue is used to pass IN and/or OUT Endpoints
        wIndex:        Not Used
        wLength:       4 bytes will be sent in the data stage to the UDC
                       these bytes can be used to control various functions

    To setup loopback: 1. Send SETUP OUT EPx request
                       2. Send SETUP LOOPBACK on EPx and EPy

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None.

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/
XLLP_STATUS_T XllpU2DProcessVendorRequest (P_XLLP_U2D_T pU2DHandle)
{
	XLLP_UINT32_T Retval;
    P_XLLP_UINT32_T pTxBuffEp0;
	XLLP_INT32_T packetLength;
    P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;

	// Are we in a host to device request?
	if (!(pU2DHandle->vendorReq.recipient & 0x80))
	{
    	// No more data left, force U2D to send zero lenght packet.
       	pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_IPR;

        // Set Endpoint 0 state machine to the End of transfer state
        pU2DHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
	}
	else
	{
		// Is this request registered in the TIM?
		Retval = FindVendorRequestInTIM (pU2DHandle->vendorReq.vendorReqType, pU2DHandle->vendorReq.recipient,
						(XLLP_UINT32_T *) &pTxBuffEp0, &packetLength);

		// Preload EP0 FIFO to get ready for IN transaction that follows.
    	// Write up to sixteen bytes to the Endpoint 0 data buffer
    	pU2DHandle->controlXfer.dataLengthEp0 = packetLength;
        pU2DHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;

        if (packetLength >= U2D_MPS_CNTRL_64)
        {
          XllpU2DWriteDwordFifo((&pRegs->U2DDR0), pTxBuffEp0, U2D_MPS_CNTRL_64);
          pTxBuffEp0 = (P_XLLP_UINT32_T)((XLLP_UINT32_T)pTxBuffEp0 + U2D_MPS_CNTRL_64);
          pU2DHandle->controlXfer.dataLengthEp0 -= U2D_MPS_CNTRL_64;

          // Update pointer to the transmit buffer
          //--------------------------------------
          pU2DHandle->controlXfer.pTxBuffEp0 = pTxBuffEp0;
        }
        else
        {
          XllpU2DWriteDwordFifo ((&pRegs->U2DDR0), pTxBuffEp0, packetLength);
          pU2DHandle->controlXfer.dataLengthEp0 -= packetLength;
        }

        if (packetLength < U2D_MPS_CNTRL_64)
        {
          // Set the status of the Endpoint 0 state machine to XFER data
          //------------------------------------------------------------
          pU2DHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
        }
        else
        {
          // Set the status of the Endpoint 0 state machine to IN data
          //----------------------------------------------------------
          pU2DHandle->controlXfer.statusEp0 = EP0_IN_DATA_STATE;
        }
		// Send the control packet on its way
        //-----------------------------------
       U2DTxEndPtZeroFifo( pU2DHandle, packetLength );
	}
    pU2DHandle->vendorReq.vendorReqComplete = XLLP_TRUE;

	return 0;
}


/******************************************************************************

  Function Name:
    XllpU2DControlProcessOutData

  Description:
    This function services the data OUT state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpU2DControlProcessOutData (P_XLLP_U2D_T pU2DHandle,
                                   P_XLLP_UINT32_T pControlXferStatus)
{
    P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;
    XLLP_INT32_T packetLength;

    // Check OPC and SA
    //-----------------
    if ( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
         (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)
       )
    {
        // U2DCSR0.OPC and U2DCSR0.SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
    }

    else if( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) == 0 )
    {

        // U2DCSR0.OPC = 0 and U2DCSR0.SA = 0 this is premature STATUS stage of
        // the SETUP transaction.
        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
    }

    else if( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA))
               == XLLP_U2DCSR0_OPC
           )
    {
        // Read data from the Endpoint 0 FIFO
        if ((pRegs->U2DCSR_X[U2D_ENDPOINT_0]) & XLLP_U2DCSR0_RNE )
        {
            packetLength = (pRegs->U2DBCR_X[U2D_ENDPOINT_0]) & XLLP_U2DBCRX_BC_MASK;
            XllpU2DReadDwordFifo (&pRegs->U2DDR0, pU2DHandle->controlXfer.pRxBuffEp0, packetLength);
            pU2DHandle->controlXfer.dataLengthEp0 -= packetLength;
            *pControlXferStatus = 0;
        }

        // Clear OPC
        pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_OPC;

        // Set IPR in case of premature STATUS IN stage
        pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_IPR;
    }


}

/******************************************************************************

  Function Name:
    XllpU2DControlProcessInData

  Description:
    This function services the data IN state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2DC_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpU2DControlProcessInData (P_XLLP_U2D_T pU2DHandle,
                                  P_XLLP_UINT32_T pControlXferStatus)
{
    P_XLLP_U2D_REGISTERS_T pRegs = (P_XLLP_U2D_REGISTERS_T)pU2DHandle->pRegsBase;
    XLLP_INT32_T packetLength;

    // Check OPC and SA
    if ( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
         (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)
       )
    {
        // OPC and SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
    }
    else if( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
              XLLP_U2DCSR0_OPC
           )
    {
        // OPC = 1 and SA = 0 this is premature STATUS stage of
        // the SETUP transaction. Clear U2DCSR0.OPC and empty transmit FIFO
        //-----------------------------------------------------------------
        pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_FTF;

        // Reset all state machine variables
        pU2DHandle->controlXfer.dataLengthEp0 = 0;
        pU2DHandle->controlXfer.pTxBuffEp0 = NULL;
        pU2DHandle->controlXfer.pRxBuffEp0 = NULL;

        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
    }

    else if ((pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) == 0)
    {
        packetLength = pU2DHandle->controlXfer.dataLengthEp0;

        if( packetLength > U2D_MPS_CNTRL_64 )
          packetLength = U2D_MPS_CNTRL_64;

        XllpU2DWriteDwordFifo (&pRegs->U2DDR0, pU2DHandle->controlXfer.pTxBuffEp0, packetLength);
        pU2DHandle->controlXfer.dataLengthEp0 -= packetLength;

        if( packetLength != U2D_MPS_CNTRL_64 )
        {
            // short packet or ZLP

            // No more data left, force U2D to send zero length packet.
            // Set Endpoint 0 state machine to the End of transfer state

            pU2DHandle->controlXfer.statusEp0 = EP0_END_XFER_STATE;
        }
        else
        {
            pU2DHandle->controlXfer.pTxBuffEp0 += (U2D_MPS_CNTRL_64 >> 2);
            // It needs more data to transmit,
            // so on the next interrupt go back to the same state.
        }

        // Send the control packet on its way
        //-----------------------------------
        U2DTxEndPtZeroFifo( pU2DHandle, packetLength );

        *pControlXferStatus = 0;
    }
}

/******************************************************************************

  Function Name:
    XllpU2DControlProcessEndXfer

  Description:
    This function services the Status state of Endpoint 0 state machine.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    P_XLLP_UINT32_T pControlXferStatus - returns the value that tells the Endpoint 0
                                         state machine to switch to the Idle state.

  Return Value:
    None.

*******************************************************************************/

void XllpU2DControlProcessEndXfer (P_XLLP_U2D_T pU2DHandle,
                                   P_XLLP_UINT32_T pControlXferStatus)
{
    P_XLLP_U2D_REGISTERS_T pRegs = (P_XLLP_U2D_REGISTERS_T)pU2DHandle->pRegsBase;

    // Check OPC and SA
    if ( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
         (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)
       )
    {
        // OPC and SA both are set,
        // there is another SETUP transaction has been sent by the host
        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
        *pControlXferStatus = EP0_JUMP_TO_IDLE;
    }
    else if( (pRegs->U2DCSR_X[U2D_ENDPOINT_0] & (XLLP_U2DCSR0_OPC | XLLP_U2DCSR0_SA)) ==
             XLLP_U2DCSR0_OPC
           )
    {
        // OPC = 1 and SA = 0 this is STATUS stage of
        // the IN transaction. Clear OPC.
        pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_OPC;

        // Initialize Endpoint 0 state machine to the Idle state
        pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;
    }

    *pControlXferStatus = 0;
}


/******************************************************************************

  Function Name:
    XllpU2DEp0InterruptHandler

  Description:
    This function is an interrupt handler for Endpoint 0 of U2D.
    It dispatches the functions of the Endpoint 0 in according to the status
    of the Endpoint 0 state machine.
    It handles the Get_Descriptor and Set_Descriptor standard requests
    and the Vendor specific requests from the USB Host.

    The Set_Configuration and Set_Interface Requests are serviced upon
    the invocation of the Configuration interrupt.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/
void XllpU2DEp0InterruptHandler (P_XLLP_U2D_T pU2DHandle)
{
  XLLP_UINT32_T controlXferStatus = 0, statusEp0;

  // Determine the status of the Endpoint 0 state machine
  // and call the apropriate function
  do
  {
     statusEp0 = pU2DHandle->controlXfer.statusEp0;
     if (statusEp0 == EP0_IDLE_STATE)
     {
         XllpU2DControlProcessIdle (pU2DHandle, &controlXferStatus);
     }
     else if (statusEp0 == EP0_OUT_DATA_STATE)
     {
        XllpU2DControlProcessOutData (pU2DHandle, &controlXferStatus);
        if((pU2DHandle->vendorReq.vendorReqComplete == XLLP_FALSE) && (pU2DHandle->controlXfer.dataLengthEp0 == 0 ))
        {
            XllpU2DProcessVendorRequest (pU2DHandle);
        }
     }
     else if (statusEp0 == EP0_IN_DATA_STATE)
     {
        XllpU2DControlProcessInData (pU2DHandle, &controlXferStatus);
     }
     else if (statusEp0 == EP0_END_XFER_STATE)
     {
        XllpU2DControlProcessEndXfer (pU2DHandle, &controlXferStatus);
     }

  } while (controlXferStatus == EP0_JUMP_TO_IDLE);
}


/******************************************************************************

  Function Name:
    XllpU2DConfigInterruptHandler

  Description:
    This is the U2D's Configuration interrupt handler. This function is called
    when the USB Host issues the Set_Config or Set_Interface requests on the USB bus.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/
void XllpU2DConfigInterruptHandler (P_XLLP_U2D_T pU2DHandle)
{
   P_XLLP_U2D_REGISTERS_T pRegs = (P_XLLP_U2D_REGISTERS_T)pU2DHandle->pRegsBase;
   XLLP_UINT32_T numActiveEndpoints;
   XLLP_UINT32_T configuration, interface, settings;

   // Get Active U2D Configuration number
   configuration = (pRegs->U2DCR & XLLP_U2DCR_ACN_MASK) >> XLLP_U2DCR_ACN_SHIFT;

   // Get Active U2D Interface Number
   interface = (pRegs->U2DCR & XLLP_U2DCR_AIN_MASK) >> XLLP_U2DCR_AIN_SHIFT;

   // Get Active U2D Alternative Interface Setting Number
   settings = (pRegs->U2DCR & XLLP_U2DCR_AAISN_MASK) >> XLLP_U2DCR_AAISN_SHIFT;

   if( (pU2DHandle->usbConfigNumActive         != configuration) ||
       (pU2DHandle->usbInterfaceNumActive      != interface    ) ||
       (pU2DHandle->usbIntAltSettingsNumActive != settings     )
     )
   {
       // Build the list of the endpoints in active configuration
       XllpU2DBuildListOfActiveEndpoints( pU2DHandle, configuration,
                                          interface, settings,
                                          &numActiveEndpoints        );

       // Update active configuration, interface and alt. settings in U2D handler structure
       //-----------------------------------------------------------------------------------
       pU2DHandle->usbConfigNumActive         = configuration;
       pU2DHandle->usbInterfaceNumActive      = interface;
       pU2DHandle->usbIntAltSettingsNumActive = settings;

       // Switch Endpoint memory to active configuration
       // Do this LAST since it indicates to HOST that U2D
       // is completely reconfigured, otherwise we'll lose an interrupt
       // for the next OUT packet ...
       //---------------------------------------------------------------
       pRegs->U2DCR |= XLLP_U2DCR_SMAC;
   }

}

/******************************************************************************

  Function Name:
    XllpU2DHWSetup

  Description:
    This function initilizes U2D Controller's hardware.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    None


  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/
XLLP_STATUS_T XllpU2DHWSetup(P_XLLP_U2D_T pU2DHandle)
{
   P_XLLP_U2D_REGISTERS_T pRegs = pU2DHandle->pRegsBase;
   XLLP_UINT32_T tempReg, temp;
   XLLP_STATUS_T error = 0;

   // Ensure U2D off ...
   //---------------------------
   pRegs->U2DCR  = 0;
   pRegs->U2DICR = 0;
   tempReg = pRegs->U2DCR;

   // Clear endpoints A->G configuration registers now
   // Ensure DME bit also off
   //-------------------------------------------------
   for( tempReg = U2D_ENDPOINT_A; tempReg <= U2D_ENDPOINT_G; tempReg++)
   {
     pRegs->U2DCR_X[tempReg]  = 0;
     pRegs->U2DCSR_X[tempReg] = 0;
     temp=pRegs->U2DCSR_X[tempReg];
     pRegs->U2DCSR_X[tempReg] = XLLP_U2DCSRX_FEF;
   }

    // Connect the UTMI (transciever) to the U2D Controller
    //--------------------------------------------------------------------------------
    //MonsU2DPlatformSetup();

    // Clear any spurious interrupts
    //------------------------------
    pRegs->U2DISR  = 0xfeffffff;

    // Enables Reset, Resume, and Configuration change interrupts by
    // setting the corresponding bits to one in the U2D Interrupt Control Register U2DICR,
    // and disables Start of the frame  and suspend interrupts.
    //------------------------------------------------------------------------------------
    tempReg  =  (XLLP_U2DICR_IERS | XLLP_U2DICR_IERU | XLLP_U2DICR_IECC);
    tempReg &= ~(XLLP_U2DICR_IESOF | XLLP_U2DICR_IESU);
    pRegs->U2DICR = tempReg;

    // Enable the Endpoint 0 interrupts by setting IE0[0] and IE0[1] & IE[2]in the U2D
    // Interrupt Control Register U2DICR, so it will respond to the USB Host requests.
    //--------------------------------------------------------------------------------
    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_0,
                            U2D_PACKET_COMPL_INT);

    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_0,
                            U2D_FIFO_ERROR_INT);

    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_0,
                            U2D_SHORT_PACKET_INT);

    // Clear endpoint 0 FIFO
    //--------------------------------
    pRegs->U2DCSR_X[U2D_ENDPOINT_0] = XLLP_U2DCSR0_FTF;

    // Enable the Endpoint B interrupts in the U2D
    // Interrupt Control Register U2DICR, so it will respond to the USB Host OUT packets
    //----------------------------------------------------------------------------------
    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_B,
                            U2D_PACKET_COMPL_INT);

    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_B,
                            U2D_FIFO_ERROR_INT);

    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_B,
                            U2D_SHORT_PACKET_INT);

    // Allow Endpoint A FIFO error interrupt...
    //-----------------------------------------
    XllpU2DEnableInterrupt (pU2DHandle,
                            U2D_INT_ENDPOINT_A,
                            U2D_FIFO_ERROR_INT);


   // Program the Endpoint Configuration Registers.
   //----------------------------------------------
   error = XllpU2DConfigureEndpoints(pU2DHandle );

   if( !error )
   {
      // Enable the first level U2Dinterrupt.
      // This is function wraper, OS specific or none
      XllpU2DEnableFirstLevelInterrupt();

      // Set the setup complete flag in the U2Ds handler structure to signal
      // that hardware setup was successful.
      pU2DHandle->setupComplete = XLLP_TRUE;
   }
   else
   {
      // Disconnect the UTMI (transciever) from the U2D
      //----------------------------------------------------------------------------------
      //MonsU2DPlatformDefault( );

      // This also resets endpoint info registers
      // and clears the config registers
      //-----------------------------------------
      pRegs->U2DCR = 0;
      XllpU2DDisableFirstLevelInterrupt();

      // Clear the configuration registers
      //----------------------------------
      for( tempReg=1;tempReg<8;tempReg++)
        pRegs->U2DCR_X[tempReg] = 0;
   }

   return error;
}

/******************************************************************************

  Function Name:
    XllpUdcHWShutdown

  Description:
    This function will shutdown the U2D Controller's hardware and clears the U2D
    handle's variables.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle

  Output Arguments:
    None

  Return Value:
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T XllpU2DHWShutdown (P_XLLP_U2D_T pU2DHandle)
{
  P_XLLP_U2D_REGISTERS_T pRegs = (P_XLLP_U2D_REGISTERS_T)pU2DHandle->pRegsBase;
  XLLP_UINT32_T tempReg;
  XLLP_STATUS_T error = 0;

  //-------------------------------------------------------
  // read-back reg values to ensure in-order execution path
  //-------------------------------------------------------

  // Disable the interrupt requests from endpoints 0 -> G,
  // and all Event interrupts
  //------------------------------------------------------
  pRegs->U2DICR = 0;
  tempReg = pRegs->U2DICR;

  // Disable first level U2D interrupt
  error = XllpU2DDisableFirstLevelInterrupt();

  // Disconnect the UTMI (transciever) from the U2D
  //-----------------------------------------------
  //MonsU2DPlatformDefault();

  // This also resets endpoint info registers
  // and clears the config registers
  //-----------------------------------------
  pRegs->U2DCR = 0;
  tempReg = pRegs->U2DCR;
  for( tempReg=1;tempReg<8;tempReg++)
  {
    pRegs->U2DCR_X[tempReg] = 0;
    pRegs->U2DCSR_X[tempReg] = 0;
  }

  // Clear the rest of the U2D handles variables
  //---------------------------------------------
  pU2DHandle->setupComplete =     XLLP_FALSE;
  pU2DHandle->enumerateComplete = XLLP_FALSE;
  pU2DHandle->cableAttached =     XLLP_FALSE;
  pU2DHandle->enableDma =         XLLP_FALSE;

  pU2DHandle->usbConfigNumActive         = 0;
  pU2DHandle->usbInterfaceNumActive      = 0;
  pU2DHandle->usbIntAltSettingsNumActive = 0;

  // Initialize Endpoint 0 state machine to the Idle state
  //------------------------------------------------------
  pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;

  // Clear any spurious interrupts
  //------------------------------
  pRegs->U2DISR  = 0xfeffffff;
  tempReg = pRegs->U2DISR;

  return error;
}

/******************************************************************************

  Function Name:
    XllpU2DSWInit

  Description:
    This function initializes the handle's variables that will be used by the
    U2D driver and no hardware will be accessed it this point.

  Global Register Modified:
    None.

  Input Arguments:
    P_XLLP_U2D_T pU2DHandle - a pointer to the U2D's handle
	P_XLLP_U2D_REGISTERS_T pRegs - a pointer to the U2D's registers base

  Output Arguments:
    None

  Return Value:
    None

*******************************************************************************/

void XllpU2DSWInit (P_XLLP_U2D_T pU2DHandle, P_XLLP_U2D_REGISTERS_T pRegs)
{
    // Set the U2D handles pointer to the U2D Registers
    // with the base address of the U2D Registers.
    //---------------------------------------------------
    pU2DHandle->pRegsBase = (P_XLLP_U2D_REGISTERS_T)pRegs;

    // Sets the U2D handles pointer to the Endpoints Configuration Table
    // with the pointer to the default Endpoints Configuration Table.
    //--------------------------------------------------------------------
    pU2DHandle->pConfigTable = (P_XLLP_U2D_EP_CONFIG_TABLE_T)defaultU2DEndpointConfigTable;

    // Clear the rest of the U2D handles variables
	pU2DHandle->setupComplete     = XLLP_FALSE;
    pU2DHandle->enumerateComplete = XLLP_FALSE;
    pU2DHandle->cableAttached     = XLLP_FALSE;
    pU2DHandle->enableDma         = XLLP_FALSE;

    pU2DHandle->usbConfigNumActive         = 0;
    pU2DHandle->usbInterfaceNumActive      = 0;
    pU2DHandle->usbIntAltSettingsNumActive = 0;

    // Initialize Endpoint 0 state machine to the Idle state
    //------------------------------------------------------
    pU2DHandle->controlXfer.statusEp0 = EP0_IDLE_STATE;

    // Get the U2D DMA initialized
    //----------------------------
    XllpU2DInit(pU2DHandle);
}
