/******************************************************************************
**  COPYRIGHT Jan 22, 2007 Marvell Inernational Ltd.
**  All Rights Reserved
**
**  COPYRIGHT (C) 2006, Intel Corporation.
**
**  This software as well as the software described in it is furnished under
**  license and may only be used or copied in accordance with the terms of the
**  license. The information in this file is furnished for informational use
**  only, is subject to change without notice, and should not be construed as
**  a commitment by Intel Corporation. Intel Corporation assumes no
**  responsibility or liability for any errors or inaccuracies that may appear
**  in this document or any software that may be provided in association with
**  this document.
**  Except as permitted by such license, no part of this document may be
**  reproduced, stored in a retrieval system, or transmitted in any form or by
**  any means without the express written consent of Intel Corporation.
**
**  FILENAME:	MSYS_INTERFACE.c
**
**  PURPOSE: 	Interface layer between bootrom/DKB and m-sys devices
**
******************************************************************************/
#include "typedef.h"
#include "errors.h"
#include "msys.h"
#include "misc.h"
#include "typedef.h"

//====================================================================
// mDOC_GetTIM
//
//     Retrieves the TIM data block from the mDOC device
//
//
//====================================================================

UINT_T mDOC_GetTIM( UINT8_T *pAddress, UINT_T timByteCount )
{
  UINT_T  sectorCount;
  #define MDOC_TIM_SECTOR_OFFSET 0   // For now, we hard code these values
  #define MDOC_TIM_PARTITION 2

  // Read the TIM from fixed location (partition 2, sector 0)
  //-------------------------------------------------------------------
  if( mDOC_ReadSectors( MDOC_TIM_SECTOR_OFFSET, timByteCount, pAddress, MDOC_TIM_PARTITION ) == M_SYS_TRUE )
    return NoError;
  else
    return ReadError;

}


//=======================================================================================
// mDOC_GetImage
//          Retrieve an image from the mDOC device
//
//=======================================================================================
UINT_T mDOC_GetImage( UINT8_T *pTargetAddress,
                        UINT_T sectorOffset,
                           UINT_T byteCount,
                              UINT8_T partition )
{
  if( mDOC_ReadSectors( sectorOffset, byteCount, pTargetAddress, partition ) == M_SYS_TRUE )
    return NoError;
  else
    return ReadError;
}

//=======================================================================================
// mDOC_WriteImage
//          Write an image to the mDOC device
//
//=======================================================================================
UINT_T mDOC_WriteImage( UINT8_T *pSourceAddress,
                         UINT_T sectorOffset,
                          UINT_T byteCount,
                           UINT8_T partition )
{
  if( mDOC_WriteSectors( sectorOffset, byteCount, pSourceAddress, partition ) == M_SYS_TRUE )
    return NoError;
  else
    return WriteError;
}

//=======================================================================================
// mDOC_ReadOtp
//          Read from the OTP area of the mDOC device.
//          The OTP area is treated as a "logical byte addressable" region
//          for consistency between differing devices.
//
//          otpAddress is logical zero-based address in the OTP region of the device
//
//          We assume that this is a sufficiently small region so as not to overflow
//          a 32 bit address...
//
//=======================================================================================
#define MDOC_OTP_PARTITION 1
#define MDOC_MAX_OTP_SECTOR 29 /* 15k area of 512 byte sectors -> sectors 0->29 */
UINT_T mDOC_ReadOtp( UINT8_T *pTargetBuffer,
                       UINT_T otpAddress,
                         UINT_T byteCount )
{
  UINT_T retVal;
  UINT_T sectorOffset;
  
  // This is the offset within the shim buffer "equivalent" to the OTP address
  //--------------------------------------------------------------------------
  UINT_T sBuffLogicalOffset;
  
  UINT_T bytesRead;
  unsigned char shimBuffer[DOCH_SECTOR_SIZE];	// limit to single sector
   
  while( byteCount )
  {
    sectorOffset = otpAddress / DOCH_SECTOR_SIZE;

	if( sectorOffset > MDOC_MAX_OTP_SECTOR )
	  return ReadError;

    retVal = mDOC_ReadSectors( sectorOffset, sizeof(shimBuffer), shimBuffer, MDOC_OTP_PARTITION );
    
    if( retVal != M_SYS_TRUE )
      return ReadError;
     
    sBuffLogicalOffset = (otpAddress -(sectorOffset * DOCH_SECTOR_SIZE));
     
    // Copy from shim buffer to target buffer
    // Calculate "logical" address extents
    //---------------------------------------
    {
      bytesRead = sizeof(shimBuffer) - sBuffLogicalOffset;
	  bytesRead = (bytesRead >= byteCount) ? byteCount : bytesRead;
      memcpy( pTargetBuffer, shimBuffer + sBuffLogicalOffset, bytesRead);
      pTargetBuffer += bytesRead;
      otpAddress    += bytesRead;
      byteCount     -= bytesRead;
    }
  } 
   
  return NoError;
}
 
 
 
//==============================================================
// DISABLED CODE - UNIT TEST AREA
//============================================================== 
#if 0
#define MDOC_OTP_SECTOR_COUNT 30
#define MDOC_BUF_SIZE (1*1024)
//=============================================================
//	mDOC_UnitTest()
//
//  Test functionality of mDOC within DKB
//
//  NOTE:  You may need to change some parameters based on how
//         the mDOC device partitions are formatted...
//
//=============================================================
void mDOC_UnitTest( void )
{
   UINT32 i, j, k, diff, rc;
   UINT8 *pBuf;
   UINT8 clrBuffer[MDOC_BUF_SIZE];
   UINT8 testBuffer[MDOC_BUF_SIZE];

   UINT8 testString[] = "This is a test of the mDOC OTP  ";
   
   if( mDOC_Initialize())
   {
     memset( testBuffer, 0xff, sizeof(testBuffer));
     memset (clrBuffer, 0xff, sizeof(clrBuffer));
     
     //First read from the TIM location, read only first 100 bytes
     //to test that no overwrite occurs
     //-----------------------------------------------------------
     rc = mDOC_ReadSectors( 0, 100, testBuffer, 2 );  // Tested, works - no overwrite
     // Examine memory window
   
     memset( testBuffer, 0xff, sizeof(testBuffer));
     rc = mDOC_ReadSectors( 0, sizeof(testBuffer), testBuffer, 2 ); //Tested, works - no overwrite
     
     memset(clrBuffer, 0xff, sizeof(clrBuffer));
     rc = mDOC_WriteSectors( 0, sizeof(clrBuffer), clrBuffer, 2);
     
     memset(clrBuffer, 0x00, sizeof(clrBuffer));
     rc = mDOC_ReadSectors( 0, sizeof(clrBuffer), clrBuffer, 2 ); // Tested, works
     
     //Now write the original data back to the mDOC
     //--------------------------------------------
     rc = mDOC_WriteSectors(0, sizeof(testBuffer), testBuffer, 2);  // Tested, works
     
     //Now check if the write went well, and compare the two buffers
     //-------------------------------------------------------------
     rc = mDOC_ReadSectors(0, sizeof(clrBuffer), clrBuffer, 2);
     diff = 0;
     for( i=0; i < sizeof(testBuffer); i++ )     
       if( testBuffer[i] != clrBuffer[i] )
       {
          diff = 1;
          break;
       }				// Tested, works; diff == 0
      
     //------------------------------------------------------------
     // Testing of OTP area with write-locking enabled in SDK code
	 //------------------------------------------------------------
	 pBuf = testBuffer;
	 while( (pBuf + sizeof(testString)) < (testBuffer + sizeof(testBuffer) -1) )
	 {
	   memcpy( pBuf, testString, sizeof(testString));
	   pBuf += sizeof(testString);
	 }
	 
	 // rc should == WriteError, if OTP write-locking enabled - tested, works
	 //----------------------------------------------------------------------
	 rc = mDOC_WriteSectors( 0,                  /*unsigned int sectorOffset,*/
                             sizeof(testBuffer), /*unsigned int numBytesToWrite,*/
                             testBuffer,         /*unsigned char *pSourceBuffer,*/
                             MDOC_OTP_PARTITION  /*unsigned char partitionNum*/ );

     // Now clear the testBuffer again and read via bootrom func
	 //--------------------------------------------------------
	 memset( testBuffer, 0, sizeof(testBuffer));
     rc = mDOC_ReadOtp( testBuffer, 0, sizeof(testBuffer));		// tested - works

     memset( testBuffer, 0xff, sizeof(testBuffer));		   
	 rc = mDOC_ReadOtp( testBuffer, 0, sizeof(testBuffer));		// tested - works

     // Now do various reads to test shim bufering, use logical addresses
	 // based at beginning of OTP area.  Use XDB memory window to verify
	 // correct content and read size, the OTP data was previously written
	 // using the DKB with lock after write
	 //-----------------------------------------------------------------
	 rc = mDOC_ReadOtp( clrBuffer, 0, sizeof(clrBuffer));		       // should see hex: 00000001000200030004... tested-works
	 rc = mDOC_ReadOtp( clrBuffer, 5122, sizeof(clrBuffer));		   // should see hex: 0A010A020A030A040A05... tested-works
	 rc = mDOC_ReadOtp( clrBuffer, 68, sizeof(clrBuffer));             // should see hex: 00220023002400250026... tested-works
     rc = mDOC_ReadOtp( clrBuffer, 512+33, sizeof(clrBuffer));  	   // should see hex: 10011101120113011401... tested-works
     rc = mDOC_ReadOtp( clrBuffer, 512+170, sizeof(clrBuffer));        // should see hex: 01550156015701580159... tested-works
	 rc = mDOC_ReadOtp( clrBuffer, (512*29)+170, sizeof(clrBuffer)/4); // should see hex: 1D551D561D571D581D59... tested-works

     //Check to see all of otp data via XDB memory window - tested all works
     //---------------------------------------------------------------------
     j = 0;   // otp area logical address
     for( i=0; i<MDOC_OTP_SECTOR_COUNT; i++ )
	 {
	   memset( clrBuffer, 0xff, sizeof(clrBuffer));
	   mDOC_ReadOtp( clrBuffer, j, sizeof(clrBuffer)/2);
	   j+= DOCH_SECTOR_SIZE;
	 }
  }
}                               
#endif